#open "unix";;
#open "lexing";;
#open "parsing";;
#open "syntax";;
#open "scanner";;
#open "grammar";;
#open "lexgen";;
#open "output";;

try
    let source_name = command_line.(1) in
    let dest_name =
      let pre4 = sub_string source_name 0 (string_length source_name - 4)
      and post4 = sub_string source_name (string_length source_name - 4) 4 in
        (if post4 = ".mll" then pre4 else source_name) ^ ".ml" in
    ic := open_in source_name;
    oc := open_out dest_name;
    let lexbuf =
      create_lexer_channel !ic in
    let (Lexdef(header,_) as def) =
      try
        lexer_definition main lexbuf
      with Parse_error x ->
        print_string "Syntax error around char ";
        print_int (get_lexeme_start lexbuf);
        print_newline();
        exit 2 in
    let ((init, states, acts) as dfa) = make_dfa def in
    output_lexdef header dfa;
    close_in !ic;
    close_out !oc
with x ->
  let retcode =
    match x with
      failure s ->
        print_string "Evaluation failed: ";
        print_string s;
        3
    | End_of_file ->
        print_string "End of file encountered during I/O";
        3
    | OS_error n ->
        print_string "System call failed: ";
        print_string(os_error_message n);
        3
    | x ->
        print_string "Uncaught exception";
        4
  in
    print_newline();
    exit retcode
;;
