#open "exc";;
#open "eq";;
#open "int";;
#open "fvect";;
#open "io";;
#open "obj";;
#open "unix";;

type qualid = {qual:string; id:string}
;;

let exc_handler fct arg =
  try
    fct arg
  with x ->
    flush std_out;
    begin match x with
      Out_of_memory ->
        prerr_string "Out of memory"
    | match_failure(file, first_char, last_char) ->
        prerr_string "Pattern matching failed, file ";
        prerr_string file;
        prerr_string " chars "; prerr_int first_char;
        prerr_string "-"; prerr_int last_char
    | failure s ->
        prerr_string "Evaluation failed : "; prerr_string s
    | Invalid_argument s ->
        prerr_string "Invalid argument : "; prerr_string s
    | OS_error n ->
        prerr_string "System call failed : ";
        prerr_string(os_error_message n)
    | x ->
        let tag = obj_tag (Repr x) in
          prerr_string "Uncaught exception ";
          prerr_string (string_of_int tag);
          begin try
            let ic = open_in command_line.(0) in
            while input_char ic != `\n` do () done;
            let size_code = input_int ic in
            let size_data = input_int ic in
            let size_symb = input_int ic in
            let size_debug = input_int ic in
              seek_in ic (pos_in ic + size_code + size_data);
              input_value ic;
              input_value ic;
              let tag_exn_table = (input_value ic : qualid vect) in
                if tag >= vect_length tag_exn_table then
                  prerr_string " (never compiled)"
                else begin
                  prerr_string " (";
                  prerr_string tag_exn_table.(tag).qual;
                  prerr_string "__";
                  prerr_string tag_exn_table.(tag).id;
                  prerr_string ")"
                end;
            close_in ic
          with _ ->
            ()
          end
    end;
    prerr_char `\n`;
    flush std_err;
    exit 2
;;
