(* Les entrees-sorties *)

#open "eq";;
#open "exc";;
#open "int";;
#open "unix";;
#open "fstring";;

type in_channel
 and out_channel
;;

let std_in = open_descriptor_in 0
and std_out = open_descriptor_out 1
and std_err = open_descriptor_out 2
;;

let exit n =
  flush std_out;
  flush std_err;
  unix__exit n
;;

let open_in_gen mode rights filename =
  let fd = open filename mode rights in
    if fd == -1
    then raise (OS_error (os_errno()))
    else open_descriptor_in fd
;;

let open_in = open_in_gen mode_rdonly 0
;;

let open_out_gen mode rights filename =
  let fd = open filename mode rights in
    if fd == -1
    then raise (OS_error (os_errno()))
    else open_descriptor_out fd
;;

let open_out =
  open_out_gen (mode_rdwr + mode_trunc + mode_creat)
               (rights_everybody * (rights_write + rights_read))
;;

let output_string channel s =
  output channel s 0 (string_length s)
;;

let print_char =
  output_char std_out
;;
let print_string =
  output_string std_out
;;
let print_int i =
  print_string (int__string_of_int i)
;;
let print_float f =
  print_string (float__string_of_float f)
;;
let print_endline s =
  print_string s; print_string "\n"
;;
let print_newline () =
  print_string "\n"; flush std_out
;;


let prerr_char =
  output_char std_err
;;
let prerr_string =
  output_string std_err
;;
let prerr_int i =
  prerr_string (int__string_of_int i)
;;
let prerr_float f =
  prerr_string (float__string_of_float f)
;;
let prerr_endline s =
  output_string std_err s; output_string std_err "\n"; flush std_err
;;
