(* Les listes *)

#open "int";;
#open "exc";;
#open "eq";;

let list_length l = length_l 0 l
  where rec length_l n = function
     []  -> n
  | _::l -> length_l (succ n) l
;;

let prefix @ l1 l2 =
  append l1 where rec append = function
     []  -> l2
  | a::l -> a :: append l
;;

let rec rev_append = fun
      []   l' -> l'
  | (a::l) l' -> rev_append l (a::l')
;;

let rev l = rev_append l []
;;

let do_list f = do_list_f
 where rec do_list_f = function
     [] -> () | x::l -> f x; do_list_f l
;;

let map f = function
    [] -> []
  | [a] -> [f a]
  | [a1; a2] -> [f a1; f a2]
  | l -> map_f l
      where rec map_f = function
          [] -> [] | a::l -> f a::map_f l
;;

let it_list f = it_list_f
 where rec it_list_f a = function
     [] -> a | b::l -> it_list_f (f a b) l
;;

let list_it f l b = list_it_f l
 where rec list_it_f = function
     [] -> b | a::l -> f a (list_it_f l)
;;

let flat_map f = flat_map_f
 where rec flat_map_f = function
     [] -> [] | x::l -> f x @ flat_map_f l
;;

let for_all p = for_all_p
 where rec for_all_p = function
     [] -> true | a::l -> p a & for_all_p l
;;

let exists p = exists_p
 where rec exists_p = function
     [] -> false | a::l -> p a or exists_p l
;;

(* Listes avec egalite *)

let mem x = mem_x
 where rec mem_x = function
     [] -> false | y::l -> x = y or mem_x l
;;

let except e = except_e
 where rec except_e = function
     [] -> []
   | elem::l -> if e = elem then l else elem::except_e l
;;

let subtract = fun
    f [] -> f
  | f e  -> subtract_e f
     where rec subtract_e = function
         [] -> []
       | elem::l -> if mem elem e then subtract_e l else elem :: subtract_e l
;;

let index a =
  index_rec 0 where rec index_rec i = function
     []  -> failwith "index"
  | b::L -> if a = b then i else index_rec (succ i) L
;;

(* Listes d'association *)

let assoc name = ass where rec ass =
  function [] -> failwith "find" | (x,y)::l -> if name = x then y else ass l
;;

let mem_assoc name = ass where rec ass =
  function [] -> false | (x,y)::l -> name = x  or  ass l
;;
