(* System functions for interactive use *)

exception Break;;
        (* Raised when the CAML process receives the interrupt signal. *)

value quit : 'a -> 'b
        (* Exits the toplevel loop. *)
  and include : string -> unit
        (* Reads, compiles and executes source phrases from the given file.
           Definitions are still entered in module "top".
           The ".ml" extension is automatically added to the file name,
           if not present. *)
  and load : string -> unit
        (* Loads in core the source for a module.
           Reads, compiles and executes source phrases from the given file.
           Definitions are entered in a new module with the given name.
           The ".ml" extension is automatically added to the file name,
           if not present. *)
  and load_object : string -> unit
        (* Loads in core a compiled bytecode file.
           The ".zo" extension is automatically added, if not present. *)
  and trace : string -> unit
        (* After trace "foo", all calls to the global function named "foo"
           will be traced.
           That is, the argument and the result are displayed for each call,
           as well as exceptions escaping out of "foo". *)
  and untrace : string -> unit
        (* untrace "foo" stops tracing function "foo". *)
;;
