.TH CAMLLEX 1 _DATE _FOOTER _HEADER

.SH NAME
camllex \- a lexical analyzer generator for Caml Light

.SH SYNOPSIS
.B camllex
.I lexer.xxx

.SH DESCRIPTION

.B camllex
produces a lexical analyzer from a set of regular expressions with
attached semantic actions, in the style of lex(1). Assuming the input
file is
.I lexer.xxx,
camllex produces the file
.I lexer.ml.
The output file must be compiled by Caml Light, and linked with the
standard library named ``lexing''.

The generated module defines one lexing function per entry points in
the lexer definition. These functions have the same names as the entry
points. Lexing functions take as argument a lexer buffer, and return
the semantic attribute of the corresponding entry point. Lexer buffers
are an abstract data type implemented in the standard library
``lexing''. The function ``lexing__create_lexer'' creates a lexer
buffer that reads from an input channel. See the documentation for
library ``lexing''. 

When used in conjunction with a parser generated by
camlyacc(1), the semantic actions compute a value belonging to the
type ``token'' defined by the generated parsing module.

.SH LEXER FORMAT

The syntax for lexer specifications is as follows:

{
.br
.I header
.br
}
.br
rule
.I entry_point
=
.br
  parse
.I regexp
{
.I action
}
.br
      | ...
.br
      |
.I regexp
{
.I action
}
.br
and
.I entry_point
=
.br
  parse ...
.br
and ...
.br
;;

The header section is arbitrary Caml Light text. It can be omitted. It
is copied as is at the beginning of the output file. Typically, it
contains the `#open' directives required by the actions, and possibly
some auxiliary functions used in the actions.

The entry points are any valid Caml Light identifier.

.SS Regular expressions

The regular expressions are in the style of lex(1), with a more
camlish syntax:
.TP
.I `c`
A character constant, with the same syntax as in Caml Light. Matches
the denoted character.
.TP
_
Matches any character.
.TP
eof
Matches the end of the lexer input.
.TP
.I "string"
A string constant, with the same syntax as in Caml Light. Matches the
corresponding sequence of characters.
.TP
.I [ character set ]
Matches any single character belonging to the given character set. Valid
character sets are: single character constants
.I `c`;
ranges of characters
.I `c1`-`c2`
(all characters between
.I c1
and
.I c2,
inclusive); and concatenation of two character sets.
.TP
.I [ ^ character set ]
Matches any single character not belonging to the given character set.
.TP
.I regexp *
(Repetition.) Matches the concatenation of zero, one, or several
strings that all match
.I regexp.
.TP
.I regexp +
(Strict repetition.) Matches the concatenation of one or several
strings that all match
.I regexp.
.TP
.I regexp ?
(Option.) Matches either the empty string, or a string matching
.I regexp.
.TP
.I regexp1 | regexp2
(Alternative.) Matches any string that match either
.I regexp1
or
.I regexp2
.TP
.I regexp1 regexp2
(Concatenation.) Matches the concatenation of two strings, the first
matching
.I regexp1,
the second matching
.I regexp2.
.TP
.I ( regexp )
Matches the same strings as
.I regexp.

.SS Actions

The actions are arbitrary Caml Light expression. They are evaluated in
a context where identifier ``lexbuf'' is bound to the current lexer
buffer. Some typical use for ``lexbuf'' are listed below:

.TP
lexing__get_lexeme lexbuf
Returns the matched string.
.TP
lexing__get_lexeme_char lexbuf n
Returns the 
.I n
th character in the matched string. The first character corresponds
to
.I n
= 0.
.TP
lexing__get_lexeme_start lexbuf
.TP
lexing__get_lexeme_end lexbuf
Return the absolute position in the input stream of the beginning
(resp. the end) of the matched string.
.TP
.IR entry_point \ lexbuf
Recursively call the lexer on the given entry point. Useful for lexing
e.g. nested comments.

.SH EXAMPLE

An example of lexer definition appears in the man page for
camlyacc(1).

.SH SEE ALSO
camlyacc(1)
.br
.I
LEX - Lexical Analyzer Generator,
.P
by M. E. Lesk and E. Schmidt.
.br
.I
Compilers - Principles, Techniques, and Tools
.P
(chapter 3), by A. V. Aho, R. Sethi, and J. D. Ullman.

.SH BUGS

There are no ways to name parts of the matched string and refer to
these parts in the actions (as with the
.I \\\(regexp\\\)
construct in ed(1)).

Type checking is performed when compiling the .ml output file, not
when compiling the lexer definition. This does not break type safety, since the
.ml output file is well-typed iff thedefinition is well-typed.
However, this results in type error messages that are somehow hard to
relate to the definition.
