.TH CAMLLIGHT 1 _DATE _FOOTER _HEADER

#ifdef msdos
#define CAMLLIGHT caml
#else
#define CAMLLIGHT camllight
#endif

.SH NAME
CAMLLIGHT \- The CAML Light toplevel system


.SH SYNOPSIS
.B CAMLLIGHT
[
.B \-I
.I lib-dir
]
[
.B \-O
.I module-set
]


.SH DESCRIPTION

.B CAMLLIGHT
is the toplevel system for CAML Light, an implementation of the ML
language. The
.B CAMLLIGHT
command permits interactive use of CAML Light, through a Lisp-like
toplevel loop. It repeatedly reads phrases on standard input and
prints inferred types and result values on standard output.
End-of-file on standard input terminates CAMLLIGHT. (For
batch-oriented use of CAML Light, see camlc(1).)

.SH OPTIONS

The following options are interpreted by CAMLLIGHT.

.TP
.BI \-I \ lib-dir
Adds lib-dir to the path of directories searched for compiled
interface files `.zi' and compiled object files `.zo'. The default is
to search in the CAML Light library first (usually
/usr/local/lib/caml-light), then in the current directory. A
directory added by -I is searched first. This option takes effect for
the compilation of subsequent `.ml' and `.mli' files, as well as for
the final linking.
.TP
.BI -O \ module-set
Specifies which set of standard library modules are to be
implicitly "opened" at the beginning of a compilation, and linked with
the user's object code. (See below the section on modules for an
explanation of "opened" modules). There are three module sets
currently available: "cautious" and "fast", and "none".
"Cautious" and "fast" both provide the same standard operations on
integers, floating-point numbers, characters and character
strings, polymorphic lists and vectors, ..., as well as basic
input/output, exception handling, ... Modules from the "cautious"
set perform range and bound checking on string and vector
operations, and various sanity checks on their arguments as well,
while those from the "fast" set don't check anything. The default
is "cautious". The "-O none" option suppresses all default module
opening; compilation starts in an (almost) empty environment, and
no library code is linked. It is not of general use.

.SH THE STANDARD LIBRARY

A number of primitives and predefined functions are available to the
user. They are documented in the .mli files in the library directory.
Some of these libraries are implicitly "opened" at the beginning of a
compilation, and implicitly linked with the user's object code.
Without the "-O" option, or with "-O cautious", the default modules are:

    exc  bool  cstring  cchar  cvect  list
    pair  ref  float  int  eq  io

With "-O fast", the default modules are:

    exc  bool  fstring  fchar  fvect  list
    pair  ref  float  int  eq  io

.SH TOPLEVEL CONTROL FUNCTIONS

Some system functions are provided at toplevel, to control its
behavior, load files in-core, and trace program execution.
.TP
quit()
Exit the toplevel loop and terminate the CAMLLIGHT process.
.TP
.RI include \ filename
Read, compile, and execute source phrases contained in file
.I filename. (Extension `.ml' is automatically added to the file name,
if not present.) This is textual inclusion: phrases are processed just
as if they were entered on standard input. In particular, global
identifiers defined by these phrases are entered in the module named
`top', not in a new module.
.TP
.RI load \ filename
Same as
.RI include \ filename,
except that a new module is created, with name
.I filename
minus the .ml extension. Global identifiers defined in
.I filename
are entered in this module. For instance, assuming file foo.ml
contains the single phrase `let bar = 1;;', loading foo.ml defines
identifier foo__bar with value 1. Caveat: the defined module is not
automatically opened: identifier bar does not automatically completes
to foo__bar. To achieve this, the user must execute `#open "foo";;' by hand.
.TP
.RI load_object \ filename
Load in-core a compiled bytecode file. The `.zo' extension is added,
if not present. Caveat: the defined module is not
automatically opened. Assuming file foo.zo defines identifier bar,
after performing `load_object "foo";;', identifier foo__bar is
defined, but identifier bar does not automatically complete to
foo__bar. To achieve this, the user must execute `#open "foo";;' by
hand.
.TP
.RI trace \ function-name
After evaluation of `trace "foo";;', all calls to the global function
named `foo' will be traced. That is, the argument and the result
are displayed for each call,  as well as exceptions escaping out of "foo".
.TP
.RI untrace \ function-name
Stop tracing a function.

.SH FILES

.ta 3.5i
#ifdef msdos
C:\\CAML\\LIB	standard library
.br
C:\\CAML\\CAMLRUN.EXE	the bytecode interpreter
.br
C:\\CAML\\CAMLC.BAT	the compiler
.br
C:\\CAML\\CAML.BAT	the toplevel
.br
#else
#ifdef macintosh
"Macintosh HD:Caml Light:Lib:"	standard library
.br
CamlRun	the bytecode interpreter
.br
Camlc	the compiler
.br
Caml	the toplevel
.br
#else
/usr/local/lib/caml-light	standard library
.br
/usr/local/lib/caml-light/run	the bytecode interpreter
.br
/usr/local/bin/camlc	the compiler
.br
/usr/local/bin/camllight	the toplevel
.br
#endif
#endif

.SH BUGS

There are very few debugging tools.

Polymorphic references, and more generally mutable concrete types, are
not safe: it is possible to create polymorphic references through a
functional encoding; and pattern-matching on mutable structures is not
compiled correctly.

Structured input/output is not type-safe; nothing prevents the user
from writing data of one type and reading it with another type.

Various size limitations are not checked by the compiler: the size of
a local environment should not exceed 256; the number of constructors
in a concrete type should not exceed 253; the number of global values
should be less than 65536; etc.

The module system is not foolproof, since it relies on the (base) names
of the files. Anything can happen if the user monkeys with the interface
path, or renames compiled interfaces.
