/* config.h */

#ifndef _config_
#define _config_


/*********** Processor dependencies **********/

/* There is nothing to do if the processor is one of the following:
        Motorola 680x0, Sparc, Mips R2000, Vax, NS32000.

   If this is your case, go to the "Operating system dependencies" below.

   Otherwise:

      - Define BIG_ENDIAN if the processor is big endian (MSB comes first).
        Leave it undefined if the processor is little endian.

      - Leave ALIGNMENT and INTERP_ALIGN undefined if the processor can access
        16-bit and 32-bit words at any address, without alignment constraints.
        This is the case for most CISCs.
        Define ALIGNMENT and INTERP_ALIGN otherwise.
        This is the case for most RISCs.                                */

#ifdef mc68020
#define BIG_ENDIAN		/* Motorola 68020, 68030, 68040 */
#undef ALIGNMENT
#undef INTERP_ALIGN
#else
#ifdef mc68000
#define BIG_ENDIAN		/* Motorola 68000 */
#define ALIGNMENT
#define INTERP_ALIGN
#endif
#endif

#ifdef sparc
#define BIG_ENDIAN		/* Sparc */
#define ALIGNMENT
#define INTERP_ALIGN
#endif

#ifdef mips
#ifdef MIPSEL			/* Mips R2000, R3000 */
#undef BIG_ENDIAN
#else
#define BIG_ENDIAN
#endif
#define ALIGNMENT
#define INTERP_ALIGN
#endif

#ifdef vax
#undef BIG_ENDIAN		/* Dec Vax */
#undef ALIGNMENT
#undef INTERP_ALIGN
#endif

#ifdef ns32000
#undef BIG_ENDIAN		/* National Semiconductor 32000 */
#undef ALIGNMENT
#undef INTERP_ALIGN
#endif

/************* Operating system dependencies **********/

/* Memory model.
        Define UNIX_MEMORY if you have brk() and sbrk().
        This enables dynamic resizing of the heap.
        Almost all Unix variants have this.
        Notable exception: the NeXT computer. */

#define UNIX_MEMORY

/* System calls.
        Define UNIX if you have the following system calls:
        truncate, stat, fstat, link, pipe, fork, execv, times.
        All Unix variants have them, except truncate, maybe (?) */

#define UNIX

/* Signals.
        Define BSD_SIGNALS if signal handlers are not reset to default
        behavior when the signal is sent. This is the case in BSD derivatives.
        In System V, just comment out the following line.
*/

#define BSD_SIGNALS

/* _longjmp/_setjmp.
        BSD provides _setjmp and _longjmp as faster variants of setjmp and
        longjmp, that do not save and restore the signal mask. If you
        don't have _setjmp and _longjmp, just uncomment the following lines */
/*
#define _setjmp setjmp
#define _longjmp longjmp
*/

/* bcopy.
  You need a memory copy routine guaranteed to work with overlapping areas.
  If you have bcopy and it works, do not #define anything.
  If you have memcpy and it works, use the first #define.
  If you have ANSI libraries, use the second #define.
  If you do not have any of the above, use the third and fourth #defines; then
   (if you are paranoiac) read the file "mem_misc.c" and make sure the memmov
   function found in that file will work on your machine.
*/
/*
#define bcopy(src, dst, len) memcpy (dst, src, len)
#define bcopy(src, dst, len) memmove (dst, src, len)
*/
#define bcopy(src, dst, len) memmov (dst, src, len)
#define USING_MEMMOV


/**************** Compiler dependencies ****************/

/* Type schar is the type of signed characters.
   On most old C compilers: this is type "char".
   On ANSI C compilers: this is type "signed char".
   On compilers in between: put "signed char" if the compiler accepts it,
     or put "char" and be sure that "char" is a signed type,
     if necessary by setting the "-signed" flag on the compiler command line.
*/
/* volatile is defined by ANSI, but you might need it on non-ANSI, optimizing
   compilers for the concurrent version.
*/

#ifdef ANSI
typedef signed char schar;
#else
typedef char schar;
#define volatile
#endif


/**************** Memory model parameters **************/

/* The following values should work on all machines.
   You can twiddle them for optimal performance of the memory allocator. */

/* The size of a page for memory management (in bytes).  Must be a multiple of
   sizeof (long).  Had better be a power of 2. */
#define Page_size 4096

/* Initial sizes of stacks (bytes). */
#define Arg_stack_size 16384
#define Ret_stack_size 16384

/* Minimal free size of stacks (words); below that, they are reallocated. */
#define Arg_stack_threshold 256 /* 1 K */
#define Ret_stack_threshold 256 /* 1 K */

/* Default initial sizes for the zones (bytes).  [Heap_size] must be a
   multiple of [Page_size]. */
#define Generation_size 16384
#define Heap_size (62 * Page_size) /* 248k */

/* Maximal size of a block allocated in the young generation (words) */
#define Max_young_size (Generation_size / 8 / sizeof (value))

/* Size increment when growing the heap (bytes).  Depending on your [malloc]
   library function, it might be a good idea to use a power of two minus
   twice the page size.  This is also the amount of memory the major GC will
   sweep before relinquishing access to the master free-list. */
#define Heap_grow_increment (62 * Page_size) /* 248k */

/* Default threshold below which the major GC wakes up.
   (Percent of free memory per running process.) */
#define Memory_low_percent 10

/* Size of the cache for gray objects (major gc) (words).  Should be the
   length of your longest list. */
#define Gray_stack_size 1024

/* Size of a chunk of free-list taken by a thread for its allocations (bytes).
*/
#define Sub_fl_size 8192

/* Tolerance margin when cutting a chunk of the free-list (bytes).  Must be
   at least 12 (3 words). */
#define Sub_fl_margin 1024

/* This should be an upper bound of the set of valid pointer values. */
#define Max_pointer ((unsigned char *) ((1 << (sizeof (char *) * 8 - 1)) - 1))

#define Num_small_fl 30


#endif /* _config_ */
