/* debugger.c : tout pour mettre au point */

#include <stdio.h>
#include <signal.h>
#ifdef PROFIL
#include <sys/time_stamp.h>
#endif
#include "control.h"
#include "debugger.h"
#include "instruct.h"
#include "memory.h"
#include "mlvalues.h"
#include "opnames.h"
#include "private.h"
#include "stacks.h"
#include "unalignd.h"

int trace_flag;

void break_here () {}

void itoh (buff, arg)
     char *buff;
     unsigned long arg;
{
  long i;

  for (i = 7; i >= 0; i--){
    buff [i] = arg % 0x10;
    buff [i] += (buff [i] < 10) ? '0' : 'a' - 10;
    arg /= 0x10;
  }
}

static unsigned long time = 0;
#define Hexlen (2 * sizeof (long))
#define Declen (3 * sizeof (long) + 2)
char dprint_buff [Dprint_buffer_size];
char *dprint_ptr = dprint_buff;

void dprint (msg)
     char *msg;
{
  unsigned long len = strlen (msg);

  if (dprint_ptr == dprint_buff){
    itoh (dprint_ptr, time++); dprint_ptr += Hexlen;
    *dprint_ptr++ = ' ';
  }
  if (dprint_ptr + len < dprint_buff + Dprint_buffer_size){
    strcpy (dprint_ptr, msg); dprint_ptr += len;
  }
}

void dprinth (msg, arg)
     char *msg;
     unsigned long arg;
{
  unsigned long len = strlen (msg);

  if (dprint_ptr == dprint_buff){
    itoh (dprint_ptr, time++); dprint_ptr += Hexlen;
    *dprint_ptr++ = ' ';
  }
  if (dprint_ptr + len < dprint_buff + Dprint_buffer_size){
    strcpy (dprint_ptr, msg); dprint_ptr += len;
  }
  if (dprint_ptr + Hexlen < dprint_buff + Dprint_buffer_size){
    itoh (dprint_ptr, arg); dprint_ptr += Hexlen;
  }
}

void dprintd (msg, arg)
     char *msg;
     long arg;
{
  unsigned long len = strlen (msg);
  long i;

  if (dprint_ptr == dprint_buff){
    itoh (dprint_ptr, time++); dprint_ptr += Hexlen;
    *dprint_ptr++ = ' ';
  }
  if (dprint_ptr + len < dprint_buff + Dprint_buffer_size){
    strcpy (dprint_ptr, msg); dprint_ptr += len;
  }
  for (i = 0; i < Declen; i++) dprint_ptr [i] = ' ';
  if (dprint_ptr + Declen < dprint_buff + Dprint_buffer_size){
    itoa (dprint_ptr, arg); dprint_ptr += Declen - 1;
  }
}

void dline ()
{
  if (dprint_ptr < dprint_buff + Dprint_buffer_size) *dprint_ptr++ = '\n';
  write (2, dprint_buff, dprint_ptr - dprint_buff);
  dprint_ptr = dprint_buff;
}

/* Affichage d'un objet structure */

long max_print = 100;
long max_print_depth = 10;

long print_cnt;

static void print_val(v, d)
	value v;
	long d;
{
  long n;
  value * p;
  char c [2];

  if (d <= 0) {
    dprint(".");
    return;
  }
  print_cnt--;
  if (print_cnt <= 0) {
    if (print_cnt == 0) dprint("...");
    return;
  }
  if (Is_long(v))
    dprintd("", Long_val(v));
  else switch(Tag_val(v)) {
    case String_tag:
      dprint("\"");
      dprint (String_val(v));
      dprint ("\"");
      break;
    case Double_tag:
      dprint("<float>");
      break;
    default:
      n = Tag_val(v);
      if (n < 26){
	c [0] = n + 'A';
	c [1] = '\0';
        dprint(c);
      }else{
        dprintd(" tag", n);
      }
      n = Wosize_val(v);
      if (n > 0) {
        dprint("(");
        p = &Field(v, 0);
        while (n > 1) {
          print_val(*p, d-1);
          dprint(", ");
          p++;
          n--;
        }
        print_val(*p, d-1);
        dprint(")");
      }
      break;
  }
}

void print_value(v)
	value v;
{
  print_cnt = max_print;
  print_val(v, max_print_depth);
  dline ();
}

extern code_t start_code;

void print_pc(p)
     code_t p;
{
  dprinth ("", cthread_self());
  dprintd (" ", p - start_code);
  dprint (" ");
}

/* Desassemblage d'une instruction */

code_t disasm_instr(p, rsp, cache_size)
     code_t p;
     value * rsp;
     mlsize_t cache_size;
{
  int i;

  print_pc(p);
  i = *p++;
  if (i < 0 || i >= sizeof(names_of_instructions) / sizeof(char *)) {
    dprint("???");
    dline ();
    return p;
  }
  dprint(names_of_instructions[i]);
  switch(i) {
/* instructions avec parametre numerique sur 1 octet */
    case ACCESS: case DUMMY: case UPDATE: case ENDLET: case CONSTBYTE:
    case ATOM: case GETFIELD: case SETFIELD:
    case MAKEBLOCK1: case MAKEBLOCK2: case MAKEBLOCK3: case MAKEBLOCK4: 
    case MAKESHAREDBLOCK1: case MAKESHAREDBLOCK2:
    case C_CALL1: case C_CALL2: case C_CALL3: case C_CALL4: case C_CALL5:
    case C_CALL_ALLOC1: case C_CALL_ALLOC2: case C_CALL_ALLOC3:
    case C_CALL_ALLOC4: case C_CALL_ALLOC5:
      dprintd(" ", *p++);
      break;
/* instruction avec parametre numerique sur 2 octets, non signe */
    case GETGLOBAL: case SETGLOBAL:
    case PUSH_GETGLOBAL_APPLY: case PUSH_GETGLOBAL_APPTERM:
      dprintd(" ", u16(p)); p += 2; break;
/* instruction avec parametre numerique sur 2 octets, signe */
    case CONSTSHORT: 
      dprintd(" ", s16(p)); p += 2; break;
/* instruction avec parametre numerique sur 4 octets */
    case CONSTLONG: case MAKEBLOCK: case MAKESHAREDBLOCK:
      dprintd(" ", s32(p)); p += 4; break;
/* instructions avec deplacement */
    case CUR: case LETREC1: case PUSHTRAP:
    case BRANCH: case BRANCHIF: case BRANCHIFNOT:
    case BRANCHIFEQ: case BRANCHIFNEQ: case BRANCHIFLT:
    case BRANCHIFGT: case BRANCHIFLE: case BRANCHIFGE:
    depl:
      { int orig = p - start_code;
	dprintd(" ", orig + s16(p));
	p += 2;
        break;
      }
/* instructions avec tag + deplacement */
    case BRANCHIFEQTAG: case BRANCHIFNEQTAG:
      dprintd(" tag ", (unsigned) *p++);
      dprint (", ");
      goto depl;
/* instructions diverses */
    case FLOATOP:
      dprint(names_of_float_instructions[*p++]);
      break;
    case SWITCH:
#define Switch_default_size 8
      { code_t orig = p;
	int i;
	for (i = 0; i < Switch_default_size - 1; i++) {
	  dprintd("", orig + s16(p) - start_code);
	  dprint (", ");
          p += 2;
	}
        dprintd("", orig + s16(p) - start_code);
	dline ();
        return orig;
      }
    case BRANCHIFNEQIMM:
      dprintd("", s32(p)); p += 4; break;
    case BRANCHIFNEQIMMSTRING:
      { int orig = p - start_code;
	int c;
	orig = orig + s16(p);
        p += 2;
	while(c = *p++) putchar(c);
	dprintd(", ", orig);
	break;
      }
    case STARTFUN:
      p += 2;
      break;
    }
  if (trace_flag >= 2 && rsp != 0) {
    mlsize_t i;
    value env;
    dprint("\t");
    for (i = cache_size; i > 0; i--, rsp++) {
      dprinth("", *rsp); dprint (" ");
    }
    dprint("/ ");
    env = ((struct return_frame *) rsp)->env;
    for (i = 0; i < Wosize_val(env); i++) {
      dprinth("", Field(env, i)); dprint (" ");
    }
  }
  dline ();
  return p;
}

void post_mortem()
{
  code_t * p;
  int n;

  p = log_ptr;
  for (n = Log_buffer_size; n > 0; n--) {
    disasm_instr(*p, 0, 0);
    p++;
    if (p >= log_buffer + Log_buffer_size) p = log_buffer;
  }
}

#ifdef TRACE
struct trace *tr;
volatile long index = 0;
#endif

mutex_t dying;

void something_went_wrong ()
{
  long i;

  mutex_lock (dying);
#ifdef TRACE
  { FILE *outfile = fdopen (2, "a");
    char buf [65536];

    setbuffer (outfile, buf, 65536);
    for (i = (index - 1) & (Trace_buffer_size - 1); i != index;
	 i = (i - 1) & (Trace_buffer_size - 1)){
      fprintf (outfile, "%010u %s %x %x\n",
	       tr [i].date,
	       tr [i].msg, tr [i].arg1, tr [i].arg2);
      ++time;
    }
    fflush (outfile);
  }
#endif
  post_mortem();
  exit(100);
}

#ifdef TRACE

volatile int pseudo_lock;

void debug_trace (msg, arg1, arg2)
     char *msg;
     unsigned long arg1, arg2;
{
  long sav_index;
#ifdef PROFIL
  struct tsval tv;
#endif

  while (pseudo_lock)
    ;
  pseudo_lock = 1;
  sav_index = index;
  index = (index + 1) & (Trace_buffer_size - 1);
  pseudo_lock = 0;

  sav_index &= (Trace_buffer_size - 1);
#ifdef PROFIL
  kern_timestamp (&tv);
  tr [sav_index].date = tv.low_val;
#endif;
  tr [sav_index].msg = msg;
  tr [sav_index].arg1 = arg1;
  tr [sav_index].arg2 = arg2;
}

void failed_assert (expr, file, line)
     char *expr, *file;
     unsigned long line;
{
  dprint ("Assertion failed: ");
  dprint (expr);
  dprint ("; file: ");
  dprint (file);
  dprintd ("; line: ", line);
  dline ();
  something_went_wrong ();
}

#else /* TRACE */

void failed_assert (expr, file, line)
     char *expr, *file;
     unsigned long line;
{
  dprint ("Assertion failed: ");
  dprint (expr);
  dprint ("; file: ");
  dprint (file);
  dprintd ("; line: ", line);
  dline ();
  exit(100);
}

#endif /* TRACE */

void init_debugger()
{
  dying = mutex_alloc ();

  signal(SIGBUS, something_went_wrong);
  signal(SIGSEGV, something_went_wrong);
  signal(SIGILL, something_went_wrong);
}
