/* debugger.h */

#ifndef _debugger_
#define _debugger_


#ifdef DEBUG

#include "mlvalues.h"

#define Log_buffer_size 100
#define Dprint_buffer_size 4096
extern int trace_flag;

#define Debug(x) x
#ifdef TRACE
#define Trace(msg, x, y) \
  debug_trace (msg, (unsigned long) x, (unsigned long) y)
#else
#define Trace(x, y, z)
#endif

#define nTrace(x, y, z)

#ifdef ANSI
#define Assert(x) \
  { if (!(x)) failed_assert ( #x , __FILE__, (unsigned long) __LINE__); }
#define Dprintx(x) \
  { dprinth ("expression " #x " ", (unsigned long) x); dline (); }
#else
#ifdef __LINE__
#define Assert(x) \
  { if (!(x)) failed_assert ("(?)" , __FILE__, (unsigned long) __LINE__); }
#define Dprintx(x) { dprinth ("expression ", (unsigned long) x); dline (); }
#else
#define Assert(x) \
  { if (!(x)) failed_assert ("(?)" , "(?)", 0l); }
#define Dprintx(x) { dprinth ("expression ", (unsigned long) x); dline (); }
#endif /* __LINE__ */
#endif /* ANSI */

#ifdef ANSI
extern void break_here (void);
extern void debug_trace (char *, unsigned long, unsigned long);
extern void failed_assert (char *, char *, unsigned long);
extern void print_value(value);
extern code_t disasm_instr(code_t, value *, mlsize_t);
extern void post_mortem(void);
extern void init_debugger (void);
#else
void break_here ();
void debug_trace ();
void failed_assert ();
void print_value();
code_t disasm_instr();
void post_mortem();
void init_debugger ();
#endif /* ANSI */

#ifdef TRACE

#include "control.h"

struct trace {
  unsigned date;
  char *msg;
  unsigned long arg1, arg2;
};
extern struct trace *tr;
#define Trace_buffer_size 0x20000  /* Must be a power of 2. */
#endif

#else /* DEBUG */

#define Debug(x)
#define Dprintx(x)
#define Trace(msg, x, y)
#define nTrace(msg, x, y)
#define Assert(x)

#endif /* DEBUG */


#endif /* _debugger_ */
