#include "config.h"
#include "control.h"
#include "debugger.h"
#include "freelist.h"
#include "gc.h"
#include "mem_misc.h"
#include "mlvalues.h"
#include "private.h"
#include "profil.h"

uchar *malloc ();

typedef struct {
  uchar *next_block_bp;
} block;

small_fl_t small_fl [4];

free_list_t fl_new ()
{
  int i;
  free_list_t result;

  result = (free_list_t) malloc (4 * sizeof (char *));
  for (i = 0; i < 4; i++) result [i] = NULL;
  return result;
}

void purge (fl, i)
     free_list_t fl;
     int i;
{
  uchar *cur;
  header_t *prev;

  cur = fl [i];
  while (cur != NULL){
    prev = (header_t *) Hp_bp (cur);
    cur = ((block *) cur)->next_block_bp;
    Assert (Color_hd (*prev) == Blue);
    *prev = Whitehd_hd (*prev);
  }
  fl [i] = NULL;
}

void fl_free (fl)
     free_list_t fl;
{
  int i;

  for (i = 0; i < 4; i++) purge (fl, i);
  free ((char *) fl);
}

value list_allocate (l, wosz)
     uchar **l;
     mlsize_t wosz;
{
  uchar *cur;
  uchar *new_hp;

  cur = *l;
  while (cur != NULL){
    if (Wosize_op (cur) >= wosz){
      if (Wosize_op (cur) >= Whsize_wosize (wosz) + 1){
	/* Allocate a chunk from the end of the block. */
	new_hp = cur + Bosize_op (cur) - Bhsize_wosize (wosz);
	/* Blue tells the GC not to collect this block. */
	Hd_hp (new_hp) = Make_header (wosz, 0, Blue);
nTrace ("list_allocate_cut", new_hp, wosz);
	Hd_op (cur) = Make_header(Wosize_op(cur)-Whsize_wosize(wosz), 0, Blue);
	return Val_hp (new_hp);
      }else{
	/* Detach the block from the free-list. */
	*l = ((block *) cur)->next_block_bp;
	if (Wosize_op (cur) == Whsize_wosize (wosz)){
	  /* Leave an extra empty block. */
	  Hd_hp (cur + Bsize_wsize (wosz)) = Make_header (0, 0, White);
	  Hd_op (cur) = Make_header (wosz, 0, Blue);
	  return Val_bp (cur);
	}else{
	  Assert (Wosize_op (cur) == wosz);
	  /* Allocate the whole block. */
	  return Val_bp (cur);
	}
      }
    }
    l = &(((block *) cur)->next_block_bp);
    cur = *l;
  }
  return NULL;
}

void replenish (fl, sz)
     free_list_t fl;
     mlsize_t sz;
{
  int i;

nTrace ("enter_replenish", cthread_self (), sz);
  condition_signal (memory_low);                   /* ### Hmmmm ! */
  if (sz == 0){
    mutex_lock (master_fl_mutex);{
      uchar *cur;
      uchar **prev;
      mlsize_t wosz = Wsize_bsize (Sub_fl_size);
      mlsize_t cumul_sz = 0;
      mlsize_t cur_sz;

      fl [0] = master_fl;
      prev = &(fl [0]);
      cur = *prev;
      Assert (Wsize_bsize (Sub_fl_margin) * sizeof (value) >= sizeof (char *));
      while (1){
	if (cur == NULL){               /* Take the whole master free-list. */
	  master_fl = NULL;
	  break;
	}
	if (cumul_sz >= wosz - Wsize_bsize (Sub_fl_margin)){
                                             /* We have enough.  Stop here. */
	  *prev = NULL;
	  master_fl = cur;
	  break;
	}
	cur_sz = Wosize_bp (cur);
	if (cumul_sz + cur_sz > wosz + Wsize_bsize (Sub_fl_margin)){
                                                /* Cut this block and stop. */
	  uchar *hp;
	  Assert (sizeof (uchar *) <= sizeof (value));
	  Assert (Wsize_bsize (Sub_fl_margin) > Whsize_wosize (2));
	  Assert (Whsize_wosize (wosz - cumul_sz) <= cur_sz - 1);
	  hp = cur + Bsize_wsize (cur_sz) - Bhsize_wosize (wosz - cumul_sz);
	  Hd_hp (hp) = Make_header (wosz - cumul_sz, 0, Blue);
	  Hd_bp (cur) = Make_header (cur_sz - Whsize_wosize (wosz - cumul_sz),
				     0, Blue);
	  ((block *) Bp_hp (hp))->next_block_bp = NULL;
	  *prev = Bp_hp (hp);
	  master_fl = cur;
nTrace ("replenish_cut_block", hp, Wosize_hp (hp));
	  break;
	}
                                                   /* Go to the next block. */
	cumul_sz += cur_sz;
	prev = &(((block *) cur)->next_block_bp);
	cur = *prev;
      }
    }mutex_unlock (master_fl_mutex);
  }else{
    Assert (sz >= 1 && sz <= 3);
    mutex_lock (small_fl [sz].mu);{
      for (i = 0; i < Num_small_fl; i++){
	small_fl [sz].tail = (small_fl [sz].tail + 1) % Num_small_fl;
	if (small_fl [sz].lists [small_fl [sz].tail] != NULL){
	  fl [sz] = small_fl [sz].lists [small_fl [sz].tail];
nTrace ("replenish_get_small", cthread_self (), fl [sz]);
	  small_fl [sz].lists [small_fl [sz].tail] = NULL;
	  goto found;
	}
      }
      small_fl [sz].avail = 0;
    found: ;
    }mutex_unlock (small_fl [sz].mu);
  }
nTrace ("exit_replenish", cthread_self (), sz);
}

value fl_allocate (fl, wosz)
     free_list_t fl;
     mlsize_t wosz;
{
  value result;
  uchar *new_hp;

  Assert (wosz > 0);
  if (wosz <= 3){
    if (fl [wosz] == NULL && small_fl [wosz].avail) replenish (fl, wosz);
    if (fl [wosz] != NULL){
      result = Val_bp (fl [wosz]);
      fl [wosz] = ((block *) (fl [wosz]))->next_block_bp;
      return result;
    }
  }

  if (fl [0] == NULL) replenish (fl, 0);

  result = list_allocate (&(fl [0]), wosz);
  if (result != NULL){
    return result;
  }
  /* No suitable block was found. */
  if (wosz <= 10) purge (fl, 0);
  return NULL;
}
