/* i_interp.c */

#include "config.h"
#include "control.h"
#include "freelist.h"
#include "major_gc.h"
#include "memory.h"
#include "mem_misc.h"
#include "private.h"

void init_interp (cycle)
     gc_cycle_t cycle;
{
  private_data * pdata;

  pdata = (private_data *) stat_alloc (sizeof (private_data));
  Set_private_data (pdata);
  young_start = stat_alloc (young_size);
  young_end = young_start + young_size;
  young_ptr = young_start;
  private_fl = fl_new ();
  arg_stack_low = (value *) stat_alloc(arg_stack_size);
  arg_stack_high = arg_stack_low + arg_stack_size / sizeof (value);
  arg_stack_threshold = arg_stack_low + Arg_stack_threshold;
  extern_asp = arg_stack_high;
  ret_stack_low = (value *) stat_alloc(ret_stack_size);
  ret_stack_high = ret_stack_low + ret_stack_size / sizeof (value);
  ret_stack_threshold = ret_stack_low + Ret_stack_threshold;
  extern_rsp = ret_stack_high;
  gtp = (struct trap_frame *) ret_stack_high;
  c_roots_head = NULL;
  per_thread_data = Atom(0);
  am_dead = 0;
  sem_init (&new_gc_cycle);
  force_minor_gc = 0;
  mutex_init (&running);
  mutex_lock (&running);
  local_current_cycle = cycle;
  mutex_lock (&cycle->lock);{
    pdata->next = cycle->head;
    cycle->head = pdata;
  }mutex_unlock (&cycle->lock);
  gc_message_str ("^");
#ifdef PROFIL
  p_modify = profil_new ("Modify");
  p_minor_gc = profil_new ("Minor_gc");
  p_survivors = profil_new ("Minor_gc_survivors:");
#endif
}

void final_interp()
{
  gc_message_str ("~");
    /* 0. Give back what remains in the private free-list. */
  fl_free (private_fl);
    /* 1. Remove all roots. */
  extern_asp = arg_stack_high;
  extern_rsp = ret_stack_high;
  c_roots_head = NULL;
  per_thread_data = Atom(0);
    /* 2. Free the memory (except the private_data structure itself). */
  fl_free (private_fl);
  stat_free ((uchar *) ret_stack_low);
  stat_free ((uchar *) arg_stack_low);
  stat_free (young_start);
    /* 3. Tell the major GC not to wait for this thread. */
  mutex_unlock (&running);
  sem_V (&new_gc_cycle);
    /* 4. Die. */
  am_dead = 1;
}

/* This function is called by the GC thread. */
void clear_private_data (pdata)
     struct private_data *pdata;
{
  mutex_clear (&pdata->pd_running);
  sem_clear (&pdata->pd_new_gc_cycle);
  stat_free ((uchar *) pdata);
}
