
/* interp.c  :  l'interpreteur de bytecode, version C rapide */


#include <stdio.h>
#include <math.h>
#include "control.h"
#include "cprim.h"
#include "debugger.h"
#include "fail.h"
#include "instruct.h"
#include "memory.h"
#include "mem_misc.h"
#include "mlvalues.h"
#include "private.h"
#include "signals.h"
#include "stacks.h"
#include "strings.h"
#include "unalignd.h"
#ifdef macintosh
#include <CursorCtl.h>
static value spin_count = 0;
#endif


/* Les registres de la machine abstraite */

/*	pc  	le pointeur de programme
	asp  	le pointeur de pile d'arguments (croit vers le bas)
	rsp  	le pointeur de pile de retour (croit vers le bas)
	gtp  	le pointeur vers le trap_frame actif
	env  	la partie remanente (dans le tas) de l'environnement
     cache_size le nombre de valeurs volatiles dans l'environnement
	accu  	l'accumulateur

"asp" et "rsp" sont des contreparties locales de "extern_asp" et "extern_rsp".

*/

extern value global_data;
extern code_t start_code;

/******************** Les macros de l'interprete ***********************/

/* D'autres point de vue sur rsp */

#define retsp  ((struct return_frame *) rsp)
#define trapsp ((struct trap_frame   *) rsp)

#define push_ret_frame() \
  (rsp = (value *) ((char *) rsp - sizeof(struct return_frame)))
#define pop_ret_frame() \
  (rsp = (value *) ((char *) rsp + sizeof(struct return_frame)))
#define push_trap_frame() \
  (rsp = (value *) ((char *) rsp - sizeof(struct trap_frame)))
#define pop_trap_frame() \
  (rsp = (value *) ((char *) rsp + sizeof(struct trap_frame)))

/* D'autres points de vue sur le pc (pour lire les operandes immediats). */

#define Short  (sizeof(short))
#define Long   (sizeof(long))
#define Double (sizeof(double))

#define s8pc  (* (schar *) pc)
#ifdef INTERP_ALIGN
#define s16pc s16(pc)
#define u16pc u16(pc)
#define s32pc s32(pc)
#define u32pc u32(pc)
#else
#define s16pc (* (short *) (pc))
#define u16pc (* (unsigned short *) (pc))
#define s32pc (* (long *) (pc))
#define u32pc (* (unsigned long *) (pc))
#endif

/* L'environnement vide */

#define null_env Atom(0)

/* Le code "raise Bottom" */

unsigned char code_for_dummy[] = { ATOM0 + Bottom_exn, RAISE };

/* pour sauver les registres avant d'appeler le gc. */

#undef Setup_for_gc
#undef Restore_after_gc

#define Setup_for_gc							      \
  { push_ret_frame();							      \
    retsp->env = env;							      \
    retsp->cache_size = cache_size;					      \
    *--asp = accu;							      \
    extern_asp = asp; extern_rsp = rsp;					      \
  }

#define Restore_after_gc						      \
  { accu = *asp++;							      \
    env = retsp->env;							      \
    pop_ret_frame ();							      \
  }

/* Pareil, mais sans sauver l'accu.
   A faire avant l'appel d'une primitive C qui peut allouer. */

#define Setup_for_c_call                                                      \
  { push_ret_frame();                                                         \
    retsp->env = env;                                                         \
    retsp->cache_size = cache_size;                                           \
    extern_asp = asp;                                                         \
    extern_rsp = rsp;                                                         \
  }
#define Restore_after_c_call                                                  \
  { asp = extern_asp;                                                         \
    rsp = extern_rsp;                                                         \
    env = retsp->env;                                                         \
    pop_ret_frame();                                                          \
  }

/* Pour mettre l'environnement tout entier dans le tas */

#define heapify_env()							      \
{									      \
  int env_size = Wosize_val(env);					      \
  int new_size = env_size + cache_size;					      \
  value * from, * to;							      \
									      \
  Alloc_small_unshr (tmp, new_size, 0);			 		      \
  for(to = Op_val(tmp); cache_size > 0; cache_size--) *to++ = *rsp++;	      \
  for(from = Op_val(env); env_size > 0; env_size--  ) *to++ = *from++;	      \
  env = tmp;								      \
}

/* L'interprete en personne */

value interprete(prog, start)          /* ML */
     code_t prog;
     value start;
{
/* Les declarations des registres de la machine abstraite.
   Donnes par frequence d'utilisation decroissante.
   Suivant les architectures et les compilateurs, on peut avoir a enlever
   certains "register", ou a imposer des registres (GCC).
   En particulier, la plupart des compilateurs optimiseurs sous-estiment
   l'importance de "pc", qui DOIT etre mis en registre. */

#if defined(__GNUC__) && defined(mc68000)
  register code_t pc asm("a5");     /* Le pointeur de code */
  register value accu;              /* L'accumulateur */
  register value * asp;		    /* Contrepartie locale de extern_asp*/
  register value * rsp;		    /* ... de extern_rsp */
#else
#if defined(__GNUC__) && defined(sparc)
  register code_t pc asm("%l0");    /* Le pointeur de code */
  register value accu asm("%l1");   /* L'accumulateur */
  register value * asp asm("%l2");  /* Contrepartie locale de extern_asp*/
  register value * rsp asm("%l3");  /* ... de extern_rsp */
#else
#if defined(__GNUC__) && defined(i386)
  register code_t pc asm("si");     /* Le pointeur de code */
  register value accu asm("ax");    /* L'accumulateur */
  register value * asp;             /* Contrepartie locale de extern_asp*/
  register value * rsp;             /* ... de extern_rsp */
#else
#if defined(__GNUC__) && defined(ns32000)
#ifndef PROFILING
  register code_t pc asm("r6");     /* Le pointeur de code */
#endif
  register value accu;              /* L'accumulateur */
  register value * asp asm("r5");   /* Contrepartie locale de extern_asp*/
  register value * rsp asm("r4");   /* ... de extern_rsp */
#else
#ifndef PROFILING
  register code_t pc;               /* Le pointeur de code */
#endif
  register value accu;              /* L'accumulateur */
  register value * asp;		    /* Contrepartie locale de extern_asp*/
  register value * rsp;		    /* ... de extern_rsp */
#endif
#endif
#endif
#endif
  mlsize_t cache_size;          /* Nombre d'entrees dans l'env. volatile */
  value env;                    /* L'environnement persistant */
  value tmp;		        /* Auxiliaire */
  jmp_buf * initial_external_raise;
  				/* "external_raise" au debut de l'interprete */
  int initial_rsp_offset;       /* "rsp" au demarrage de l'interprete */
  value * initial_c_roots_head; /* "C_roots_head" au demarrage de l'interp. */
  jmp_buf raise_buf;            /* Pour traiter les d'exceptions externes */

  asp = extern_asp;
  rsp = extern_rsp;
  pc = prog + Long_val(start);
  env = null_env;
  cache_size = 0;
  accu = Val_long(0);
  initial_c_roots_head = c_roots_head;
#ifdef DEBUG
  log_ptr = log_buffer;
#endif

  if (setjmp(raise_buf)) {
    c_roots_head = initial_c_roots_head;
    accu = exn_bucket;
    goto raise;
  }
  initial_external_raise = external_raise;
  external_raise = (jmp_buf *) raise_buf;
  initial_rsp_offset = (char *) ret_stack_high - (char *) rsp;

  while (1) {

#ifdef DEBUG
    *log_ptr++ = pc;
    if (log_ptr >= log_buffer + Log_buffer_size)
      log_ptr = log_buffer;
    if (trace_flag)
      disasm_instr(pc, rsp, cache_size);
    Assert(asp >= arg_stack_low);
    Assert(asp <= arg_stack_high);
    Assert(rsp >= ret_stack_low);
    Assert(rsp <= ret_stack_high);
#endif /* DEBUG */

    switch (*pc++) {

    case STOP:
      extern_asp = asp;
      extern_rsp = rsp;
      external_raise = initial_external_raise;
      return accu;
      
/********** Construction et application de valeurs fonctionnelles **********/
      
    case CUR:
      if (cache_size) heapify_env();
      Alloc_small_unshr(accu, Closure_wosize, Closure_tag);
      Env_val(accu) = env;
      Code_val(accu) = pc + s16pc;
      pc += Short;
      break;

    case APPLY:
    apply:
      push_ret_frame();
      retsp->retaddr = pc;
      retsp->env = env;
      retsp->cache_size = cache_size;
      *--rsp = *asp++;		/* Place le premier arg. dans l'env. */
      cache_size = 1;
      pc = Code_val(accu);
      env = Env_val(accu);
      break;
      
    case STARTFUN:
      if (asp < arg_stack_threshold || rsp < ret_stack_threshold) {
        Setup_for_gc;
        realloc_stacks();
        rsp = extern_rsp;
        asp = extern_asp;
        Restore_after_gc;
      }
      if (signal_is_pending) goto raise_signal;
      if (force_minor_gc){
	force_minor_gc = 0;
	if (current_gc_cycle != local_current_cycle){
	  Setup_for_gc;
	  minor_collection ();
	  Restore_after_gc;
	}
      }
#ifdef macintosh
      /* This is the right time to be Multifinder-friendly. */
      if ((spin_count++ & 15) == 0) SpinCursor ((short) 1);
#endif
      break;

    case RETURN:
      if (*asp == Mark) {
	rsp += cache_size;	/* Vide le cache */
	asp++;			/* Detruit la marque */
	pc = retsp->retaddr;		/* Et retourne a l'appelant */
	env = retsp->env;
	cache_size = retsp->cache_size;
	pop_ret_frame();
	break;
      }
      /* fall through APPTERM */
      
    case APPTERM:
    appterm:
      rsp += cache_size;	/* Vide le cache */
      *--rsp = *asp++;		/* Place le premier arg. dans l'env. */
      cache_size = 1;
      pc = Code_val(accu);
      env = Env_val(accu);
      break;

    case PUSH_GETGLOBAL_APPLY:
      *--asp = accu;
      accu = Field(global_data, u16pc);
      pc += Short;
      goto apply;

    case PUSH_GETGLOBAL_APPTERM:
      *--asp = accu;
      accu = Field(global_data, u16pc);
      pc += Short;
      goto appterm;

/******* Manipulations de l'environnement (acces, ajout, destruction) *****/
      
    case GRAB:
      if (*asp != Mark) {
	*--rsp = *asp++;
	cache_size++;
      } else {
	if (cache_size) heapify_env();
	Alloc_small_unshr(accu, Closure_wosize, Closure_tag);
	Code_val(accu) = pc;
	Env_val(accu) = env;
	asp++;			/* Detruit la marque */
	pc = retsp->retaddr;
	env = retsp->env;
	cache_size = retsp->cache_size;
	pop_ret_frame();
      }
      break;
      
#define access(n) (cache_size > n ? rsp[n] : Field(env, n - cache_size))
#define access0() (cache_size > 0 ? rsp[0] : Field(env,0))
      /* Un peu plus efficace que access(0)... */

    case ACC0:
      accu = access0(); break;
    case ACC1:
      accu = access(1); break;
    case ACC2:
      accu = access(2); break;
    case ACC3:
      accu = access(3); break;
    case ACC4:
      accu = access(4); break;
    case ACC5:
      accu = access(5); break;
    case ACCESS:
      { int n = *pc++;
	accu = access(n);
	break;
      }
      
    case LET:
      *--rsp = accu;
      cache_size++;
      break;
      
    case DUMMY:
      { int n = *pc++;
	int i = n;
	Assert (n > 0);
	tmp = alloc_shr (n);
	while (i--){
	  Field (tmp, i) = Val_long (0);
	}
	end_alloc (tmp, n, 0);
	*--rsp = tmp;
	cache_size++;
	break;
      }

    case UPDATE:
      { int n = *pc++;
	tmp = access(n);
	Tag_val (tmp) = Tag_val (accu);
                                /**** The previous line assumes:
                                   1. That [Tag_val (tmp)] is a single byte.
                                   2. That writing a byte is atomic.
                                */
	for (n = 0; n < Wosize_val (accu); n++){
	  Modify (&Field (tmp, n), Field (accu, n));
	}
	break;
      }

    case LETREC1:		/* Remplace Dummy 2; Cur lbl; Update 0 */
      Alloc_small_unshr (accu, Closure_wosize, Closure_tag);
      Code_val(accu) = pc + s16pc;
      Env_val(accu) = Atom(0);
      *--rsp = accu;
      cache_size++;
      heapify_env();
      accu = Field(env,0);
      if (Is_in_heap (accu)){
	Modify (&Env_val (accu), env);
      }else{
	Env_val (accu) = env;
      }
      pc += Short;
      break;

    case ENDLET1:
      if (cache_size != 0) {
	cache_size--; rsp++;
      } else {
	int i;                  /* On recopie dans le cache l'ancien env. */
	value * from;           /* prive de sa premiere valeur */
	i = Wosize_val(env);
        from = &Field(env, i);
        cache_size = i - 1;
        for (i = cache_size; i > 0; i--) *--rsp = *--from;
	env = null_env;
      }
      break;

    case ENDLET:
      { int n = *pc++;
	if (cache_size >= n) {
	  cache_size -= n;
	  rsp += n;
	} else {
	  int i;
	  value * from;
	  n -= cache_size;	/* Ce qu'il faut oter dans l'env du tas */
	  rsp += cache_size;	/* On vide entierement le cache */
	  			/* On recopie ds le cache l'ancien env */
	  			/* prive de ses n premieres valeurs */
          i = Wosize_val(env);
	  cache_size = i - n;
          from = &Field(env, i);
	  for (i = cache_size; i > 0; i--) *--rsp = *--from;
	  env = null_env;
	}
	break;
      }
      
/**************** Gestion des exceptions ******************/
      
    case PUSHTRAP:
      { value * src = rsp + cache_size;
	int i = cache_size;
	
	push_trap_frame();		/* On sauve l'etat de la machine */
	trapsp->retaddr = pc + s16pc;
	pc += Short;
	trapsp->env = env;
	trapsp->cache_size = cache_size + 2;  /* +2 : astuce pour le GC */
	trapsp->asp = asp;
	trapsp->tp = gtp;
	gtp = trapsp;
	while(i--) *--rsp = *--src;	/* On duplique le cache */
	*--asp = Mark;		/* On stoppe la consommation d'arguments */
	break;
      }

    raise_signal:               /* La detection d'un signal amene ici */
      accu = pending_signal;
      signal_is_pending = 0;
      pending_signal = Val_long(1);

    raise:			/* un raise externe amene ici */

    case RAISE:
      if (signal_is_pending) goto raise_signal;
      rsp = (value *) gtp;	/* On arase la pile R */
      if (rsp >= (value *)((char *) ret_stack_high - initial_rsp_offset)) {
                                /* On sort de l'interprete? */
        exn_bucket = accu;
        external_raise = initial_external_raise;
        longjmp(*external_raise, 1);
      }
      pc = trapsp->retaddr;	/* On restaure l'etat de la machine */
      env = trapsp->env;
      cache_size = trapsp->cache_size - 2;   /* -2 : astuce pour le GC */
      asp = trapsp->asp;
      gtp = trapsp->tp;
      pop_trap_frame();		/* On depile le trap frame */
      *--rsp = accu;		/* On met le bucket en tete de l'env. */
      cache_size++;
      break;
      
    case POPTRAP:
      if (signal_is_pending) goto raise_signal;
      rsp = (value *) gtp;
      env = trapsp->env;
      cache_size = trapsp->cache_size - 2;
      asp = trapsp->asp;
      gtp = trapsp->tp;
      pop_trap_frame();
      break;

/***************** Chargement de constantes dans l'accu *******************/
      
    case CONSTBYTE:		/* Pour les petits entiers */
      accu = s8pc;  pc += 1; break;
    case CONSTSHORT:		/* Pour les moins petits entiers */
      accu = s16pc; pc += Short; break;
    case CONSTLONG:		/* Pour une valeur hors du tas quelconque */
      accu = s32pc; pc += Long; break;
    case ATOM0:
      accu = Atom(0); break;
    case ATOM1:
      accu = Atom(1); break;
    case ATOM2:
      accu = Atom(2); break;
    case ATOM3:
      accu = Atom(3); break;
    case ATOM4:
      accu = Atom(4); break;
    case ATOM5:
      accu = Atom(5); break;
    case ATOM6:
      accu = Atom(6); break;
    case ATOM7:
      accu = Atom(7); break;
    case ATOM8:
      accu = Atom(8); break;
    case ATOM9:
      accu = Atom(9); break;
    case ATOM:
      accu = Atom(*pc++); break;
      
/********** Acces aux valeurs globales *****************/
      
    case GETGLOBAL:
      accu = Field(global_data, u16pc);
      pc += Short;
      break;
    case SETGLOBAL:
      Modify(&Field(global_data, u16pc), accu);
      pc += Short;
      break;
      
/************ Les classique manipulations de pile (d'arguments) **********/
      
    case PUSH:
      *--asp = accu; break;
    case POP:
      accu = *asp++; break;
    case PUSHMARK:
      *--asp = Mark;
      break;
      
/************ Branchements, branchements conditionnels et tutti quanti *******/
      
    case NOP1:			/* Generes pour assurer l'alignement */
      break;			/* si necessaire */
    case NOP2:
      pc++; break;
    case NOP3:
      pc += 2; break;
      
#define branch() pc += s16pc
#define cond_branch(condition) if (condition) branch(); else pc += 2
      
    case BRANCH:
      branch(); break;
    case BRANCHIF:
      cond_branch(Tag_val(accu) != 0); break;
    case BRANCHIFNOT:
      cond_branch(Tag_val(accu) == 0); break;
    case BRANCHIFEQTAG:
      if (Tag_val(accu) == *pc++) branch(); else pc += Short;
      break;
    case BRANCHIFNEQTAG:
      if (Tag_val(accu) != *pc++) branch(); else pc += Short;
      break;
    case SWITCH:
      pc += s16(pc + accu - 1);
      break;
    case BOOLNOT:
      accu = Atom(Tag_val(accu) == 0); break;
      
/******************* Acces aux champs d'un bloc **********************/
      
    case GETFIELD0:
      tmp = accu;
      accu = Field(accu,0); break;
    case GETFIELD1:
      tmp = accu;
      accu = Field(accu,1); break;
    case GETFIELD2:
      tmp = accu;
      accu = Field(accu,2); break;
    case GETFIELD3:
      tmp = accu;
      accu = Field(accu,3); break;
    case GETFIELD:
      tmp = accu;
      accu = Field(accu,*pc++); break;
      
    case SETFIELD0:
      tmp = accu; accu = *asp++; Modify(&Field(tmp,0), accu); break;
    case SETFIELD1:
      tmp = accu; accu = *asp++; Modify(&Field(tmp,1), accu); break;
    case SETFIELD2:
      tmp = accu; accu = *asp++; Modify(&Field(tmp,2), accu); break;
    case SETFIELD3:
      tmp = accu; accu = *asp++; Modify(&Field(tmp,3), accu); break;
    case SETFIELD:
      { value * ptr;
        ptr = &Field(accu,*pc++);
        accu = *asp++;
        Modify(ptr, accu);
        break;
      }
      
/******************* Construction d'un bloc **************************/
      
    case MAKEBLOCK1:
      Alloc_small_unshr (tmp, 1, *pc);
      pc++;
      Field(tmp,0) = accu;
      accu = tmp;
      break;
    case MAKEBLOCK2:
      Alloc_small_unshr (tmp, 2, *pc);
      pc++;
      Field(tmp,0) = accu;
      Field(tmp,1) = *asp++;
      accu = tmp;
      break;
    case MAKEBLOCK3:
      Alloc_small_unshr (tmp, 3, *pc);
      pc++;
      Field(tmp,0) = accu;
      Field(tmp,1) = *asp++;
      Field(tmp,2) = *asp++;
      accu = tmp;
      break;
    case MAKEBLOCK4:
      Alloc_small_unshr (tmp, 4, *pc);
      pc++;
      Field(tmp,0) = accu;
      Field(tmp,1) = *asp++;
      Field(tmp,2) = *asp++;
      Field(tmp,3) = *asp++;
      accu = tmp;
      break;
    case MAKEBLOCK:
      { header_t hdr;
	mlsize_t size;
	tag_t tag;
	value * to;
	mlsize_t i;

	hdr = u32pc;
	pc += Long;
	size = Wosize_hd (hdr);
	tag = Tag_hd (hdr);
        if (size < Max_young_size) {
          Alloc_small_unshr(tmp, size, tag);
          to = &Field(tmp, 0);
          *to++ = accu;
          for (i = size - 1; i > 0; i--) *to++ = *asp++;
          accu = tmp;
        } else {
          tmp = alloc_shr(size);
          to = &Field(tmp, 0);
          Initialize (to, accu);
	  to++;
          for (i = size - 1; i > 0; i--, to++, asp++) Initialize (to, *asp);
	  end_alloc (tmp, size, tag);
          accu = tmp;
        }
	break;
      }

    case MAKESHAREDBLOCK1:
      tmp = alloc_shr(1);
      Initialize(&Field(tmp,0), accu);
      end_alloc(tmp, 1, *pc);
      pc++;
      accu = tmp;
      break;
    case MAKESHAREDBLOCK2:
      tmp = alloc_shr(2);
      Initialize(&Field(tmp,0), accu);
      Initialize(&Field(tmp,1), *asp);
      asp++;
      end_alloc(tmp, 2, *pc);
      pc++;
      accu = tmp;
      break;
    case MAKESHAREDBLOCK:
      { header_t hdr;
	mlsize_t size, i;
	tag_t tag;
	value * to;
	
	hdr = u32pc;
	pc += Long;
	size = Wosize_hd (hdr);
	tag = Tag_hd (hdr);
	tmp = alloc_shr(size);
	to = &Field(tmp, 0);
	Initialize (to, accu);
	to++;
	for (i = size - 1; i > 0; i--, to++, asp++) Initialize (to, *asp);
	end_alloc (tmp, size, tag);
	accu = tmp;
	break;
      }

    case TAGOF:
      accu = Val_long(Tag_val(accu));
      break;

/******************* Appel de fonctions C ***********************/

    case C_CALL1:
      accu = (cprim[*pc++])(accu);
      break;
    case C_CALL2:
      accu = (cprim[*pc++])(accu, asp[0]);
      asp += 1;
      break;
    case C_CALL3:
      accu = (cprim[*pc++])(accu, asp[0], asp[1]);
      asp += 2;
      break;
    case C_CALL4:
      accu = (cprim[*pc++])(accu, asp[0], asp[1], asp[2]);
      asp += 3;
      break;
    case C_CALL5:
      accu = (cprim[*pc++])(accu, asp[0], asp[1], asp[2], asp[3]);
      asp += 4;
      break;
    case C_CALLN:
      { int n = *pc++;
        *--asp = accu;
        accu = (cprim[*pc++])(asp, n);
        asp += n;
        break; }

    case C_CALL_ALLOC1:
      Setup_for_c_call;
      accu = (cprim[*pc++])(accu);
      Restore_after_c_call;
      break;
    case C_CALL_ALLOC2:
      Setup_for_c_call;
      accu = (cprim[*pc++])(accu, asp[0]);
      Restore_after_c_call;
      asp += 1;
      break;
    case C_CALL_ALLOC3:
      Setup_for_c_call;
      accu = (cprim[*pc++])(accu, asp[0], asp[1]);
      Restore_after_c_call;
      asp += 2;

      break;
    case C_CALL_ALLOC4:
      Setup_for_c_call;
      accu = (cprim[*pc++])(accu, asp[0], asp[1], asp[2]);
      Restore_after_c_call;
      asp += 3;
      break;
    case C_CALL_ALLOC5:
      Setup_for_c_call;
      accu = (cprim[*pc++])(accu, asp[0], asp[1], asp[2], asp[3]);
      Restore_after_c_call;
      asp += 4;
      break;
    case C_CALL_ALLOCN:
      { int n = *pc++;
        *--asp = accu;
        Setup_for_c_call;
        accu = (cprim[*pc++])(asp, n);
        Restore_after_c_call;
        asp += n;
        break; }
      
/******************** Primitives sur les entiers **********************/
      
    case NEGINT:
      accu = 2 - accu; break;
    case SUCCINT:
      accu += 2; break;
    case PREDINT:
      accu -= 2; break;
    case ADDINT:
      accu += *asp++ - 1; break;
    case SUBINT:
      accu -= *asp++ - 1; break;
    case MULINT:
      accu = 1 + (accu >> 1) * (*asp++ - 1); break;
    case DIVINT:
      tmp = *asp++ - 1;
      if (tmp == 0) {
        accu = Atom(Zero_divide_exn);
        goto raise;
      }
      accu = Val_long((accu - 1) / tmp);
      break;
    case MODINT:
      tmp = *asp++ - 1;
      if (tmp == 0) {
        accu = Atom(Zero_divide_exn);
        goto raise;
      }
      accu = 1 + (accu - 1) % tmp;
      break;
    case ANDINT:
      accu &= *asp++; break;
    case ORINT:
      accu |= *asp++; break;
    case XORINT:
      accu = 1 + (accu ^ *asp++); break;
    case SHIFTLEFTINT:
      accu = 1 + ((accu - 1) << Long_val(*asp++)); break;
    case SHIFTRIGHTINT:
      accu = 1 | ((accu - 1) >> Long_val(*asp++)); break;
      
#define inttest(name1,name2,tst)					      \
    case name1:								      \
      accu = Atom(accu tst *asp++);					      \
      break;								      \
    case name2:								      \
      cond_branch(accu tst *asp++);					      \
      break;
      
      inttest(EQ, BRANCHIFEQ, ==);
      inttest(NEQ, BRANCHIFNEQ, !=);
      inttest(LTINT, BRANCHIFLT, <);
      inttest(GTINT, BRANCHIFGT, >);
      inttest(LEINT, BRANCHIFLE, <=);
      inttest(GEINT, BRANCHIFGE, >=);
      
    case BRANCHIFNEQIMM:
      if (accu != s32pc) { pc += Long; branch(); } else pc += Long+Short;
      break;
    case INCR:
      accu = (Field(accu, 0) += 2); break;
    case DECR:
      accu = (Field(accu, 0) -= 2); break;

/****************** Primitives sur les flottants *************************/
      
    case FLOATOP:
      { Alloc_small_unshr(tmp, Double_size, Double_tag);
	switch(*pc++) {
	case FLOATOFINT: Double_val(tmp) = Long_val(accu); break;
	case NEGFLOAT:   Double_val(tmp) = -Double_val(accu); break;
	case ADDFLOAT:
	  Double_val(tmp) = Double_val(accu) + Double_val(*asp++); break;
	case SUBFLOAT:
	  Double_val(tmp) = Double_val(accu) - Double_val(*asp++); break;
	case MULFLOAT:
	  Double_val(tmp) = Double_val(accu) * Double_val(*asp++); break;
	case DIVFLOAT:
	  Double_val(tmp) = Double_val(accu) / Double_val(*asp++); break;
	}
	accu = tmp;
	break;
      }
      
    case INTOFFLOAT:
      accu = Val_long((long)Double_val(accu)); break;
      
#define floattest(name, tst)    					      \
    case name:								      \
      accu = Atom(Double_val(accu) tst Double_val(*asp++));		      \
      break;
      
      floattest(EQFLOAT, ==);
      floattest(NEQFLOAT, !=);
      floattest(LTFLOAT, <);
      floattest(GTFLOAT, >);
      floattest(LEFLOAT, <=);
      floattest(GEFLOAT, >=);
      
    case BRANCHIFNEQIMMFLOAT:
      if (Double_val(accu) != *((double *)pc))
        { pc += Double; branch(); }
      else
        pc += Double+Short;
      break;

/********************* Primitives sur les chaines ***********************/
      
    case MAKESTRING:
      { mlsize_t size = Long_val(accu);
        Alloc_string(accu, size);
        break; }
    case STRINGLENGTH:
      { accu = Val_long (string_length (accu));
        break; }
    case GETSTRINGCHAR:
      accu = Val_long(Byte(accu, Long_val(*asp++)));
      break;
    case SETSTRINGCHAR:
      Byte(accu, Long_val(asp[0])) = Long_val(asp[1]);
      accu = asp[1];
      asp += 2;
      break;

    case BRANCHIFNEQIMMSTRING:
      { code_t newpc = pc + s16pc;
        unsigned char * p = (unsigned char *) String_val(accu);
        pc += Short;
        do {
          if (*p++ != *pc) { pc = newpc; break; }
        } while (*pc++);
        break;
      }
      
#define stringtest(name, tst)                                                 \
    case name:                                                                \
      accu = Atom(compare_strings(accu, *asp++) tst Val_long(0));             \
      break;
      
      stringtest(EQSTRING, ==);
      stringtest(NEQSTRING, !=);
      stringtest(LTSTRING, <);
      stringtest(GTSTRING, >);
      stringtest(LESTRING, <=);
      stringtest(GESTRING, >=);
      
/********************* Primitives sur les vecteurs *********************/
      
    case MAKEVECTOR:
      { mlsize_t size = Long_val(accu);
        if (size == 0){
          accu = Atom(0);
        }else{
	  value *to;
	  mlsize_t i;
	  accu = alloc_shr (size);
	  tmp = oldify (*asp);
	  for (i = size, to = &Field (accu, 0); i > 0; i--, to++) *to = tmp;
	  end_alloc (accu, size, 0);
	}
        asp++;
	break;
      }
    case VECTLENGTH:
      accu = Val_long(Wosize_val(accu));
      break;
    case GETVECTITEM:
      accu = Field(accu, Long_val(*asp++));
      break;
    case SETVECTITEM:
      tmp = asp[1];
      Modify(& Field(accu, Long_val(asp[0])), tmp);
      asp += 2;
      accu = tmp;
      break;
#ifdef DEBUG
    default:
      asm (".globl foobargee");
      asm ("foobargee:");
      post_mortem();
      fatal_error("post mortem");
#endif
    }
  }
}
