/* intext.c : entrees/sorties structurees */

#include "control.h"
#include "debugger.h"
#include "fail.h"
#include "gc.h"
#include "intext.h"
#include "io.h"
#include "memory.h"
#include "mlvalues.h"
#include "private.h"

#define Hash(v) (((unsigned long) (v) >> 2) % extern_table_size)

static void resize_result()
{
  extern_size = 2 * extern_size;
  extern_block =
    (unsigned long *) stat_resize((uchar *) extern_block,
				  extern_size * sizeof(unsigned long));
}

static void alloc_table()
{
  unsigned long i;
  extern_table = (struct extern_obj *)
                   stat_alloc(extern_table_size * sizeof(struct extern_obj));
  for (i = 0; i < extern_table_size; i++)
    extern_table[i].obj = 0;
}
 
static void resize_table()
{
  unsigned long oldsize;
  struct extern_obj * oldtable;
  unsigned long i, h;

  oldsize = extern_table_size;
  oldtable = extern_table;
  extern_table_size = 2 * extern_table_size;
  alloc_table();
  for (i = 0; i < oldsize; i++) {
    h = Hash(oldtable[i].obj);
    while (extern_table[h].obj != 0) {
      h++;
      if (h >= extern_table_size) h = 0;
    }
    extern_table[h].obj = oldtable[i].obj;
    extern_table[h].ofs = oldtable[i].ofs;
  }
  stat_free((uchar *) oldtable);
}

static unsigned long emit(v)
     value v;
{
  mlsize_t size;
  unsigned long h;
  unsigned long res;
  value * p;
  unsigned long * q;
  mlsize_t n;
  unsigned long end_pos;

  if (Is_long(v)) return (long) v;
  size = Wosize_val(v);
  if (size == 0) return (Tag_val(v) << 2) + 2;
  if (2 * extern_table_used >= extern_table_size) resize_table();
  h = Hash(v);
  while (extern_table[h].obj != 0) {
    if (extern_table[h].obj == v) return extern_table[h].ofs;
    h++;
    if (h >= extern_table_size) h = 0;
  }
  end_pos = extern_pos + 1 + size;
  while (end_pos >= extern_size) resize_result();
  extern_block[extern_pos++] = Hd_val(v);
  res = extern_pos * sizeof(unsigned long);
  extern_table[h].obj = v;
  extern_table[h].ofs = res;
  extern_table_used++;
  for (n = size, p = &Field(v, 0), q = &extern_block[extern_pos]; n > 0; n--) {
    *q++ = *p++;
  }
  extern_pos = end_pos;
  return res;
}

static unsigned long emit_all(root)
     value root;
{
  unsigned long read_pos;
  unsigned long res;
  header_t hd;
  mlsize_t sz;

  read_pos = extern_pos;
  res = emit(root);
  while (read_pos < extern_pos) {
    hd = (header_t) extern_block[read_pos++];
    sz = Wosize_hd(hd);
    switch(Tag_hd(hd)) {
    case String_tag:
    case Double_tag:
      read_pos += sz;
      break;
    case Abstract_tag:
    case Final_tag:
      end_blocking_section ();
      failwith("extern: abstract value");
      break;
    case Closure_tag:
      end_blocking_section ();
      failwith("extern: code value");
      break;
    default:
      while (sz > 0) {
        extern_block[read_pos] = emit((value) extern_block[read_pos]);
        read_pos++;
        sz--;
      }
      break;
    }
  }
  return res;
}

value extern_val(chan, v)       /* ML */
     struct channel * chan;
     value v;
{
  unsigned long res;

  begin_blocking_section ();
  extern_size = 4096;
  extern_block =
    (unsigned long *) stat_alloc(extern_size * sizeof(unsigned long));
  extern_pos = 0;
  extern_table_size = 2039;
  alloc_table();
  extern_table_used = 0;
  res = emit_all(v);
  stat_free((uchar *) extern_table);
  if (extern_pos >= Max_wosize) {
    end_blocking_section ();
    failwith("extern");
  }
  mutex_lock(&chan->mutex);
#ifdef BIG_ENDIAN
  putword(chan, Big_endian_magic_number);
#else
  putword(chan, Little_endian_magic_number);
#endif
  putword(chan, extern_pos);
  if (extern_pos == 0)
    putword(chan, res);
  else
    putblock(chan, (value) extern_block, 0, extern_pos*sizeof(unsigned long));
  mutex_unlock(&chan->mutex);
  stat_free((uchar *) extern_block);
  end_blocking_section ();
  return Atom(0);
}

void adjust_pointers(start, actual_hd)
     value start;
     header_t actual_hd;
{
  mlsize_t size;
  unsigned long size_in_bytes;
  color_t color;
  value * p, * q;
  mlsize_t sz;
  header_t hd;
  tag_t tag;
  value v;

  size = Wosize_val(start);
  size_in_bytes = size * sizeof(value);
  color = Color_val(start);
  sz = Wosize_hd(actual_hd);
  tag = Tag_hd(actual_hd);
  p = Op_val(start);
  q = p + size;
  while (1) {
    if (tag >= No_scan_tag)
      p += sz;
    else
      for( ; sz > 0; sz--, p++) {
        v = *p;
        switch(v & 3) {
        case 0:                 /* 0 -> bloc represente par son offset */
          Assert(v >= 0 && v <= size_in_bytes && (v & 3) == 0);
          *p = (value) ((unsigned long) start - 4 + v);
          break;
        case 2:                 /* 2 -> atome */
          v = v >> 2;
          Assert(v >= 0 && v < 256);
          *p = Atom(v);
          break;
        default:                /* 1 ou 3 -> entier */
          break;
        }
      }
    if (p >= q) break;
    hd = *p;
    sz = Wosize_hd(hd);
    tag = Tag_hd(hd);
    *p++ = Make_header(sz, tag, color);
  }
  sz = Wosize_hd(actual_hd);
  tag = Tag_hd(actual_hd);
  Hd_val(start) = Make_header(sz, tag, color);
}

#define Reverse_word(w) {                                                     \
  char * _p;                                                                  \
  int _a;                                                                     \
  _p = (char *) (w);                                                          \
  _a = _p[0];                                                                 \
  _p[0] = _p[3];                                                              \
  _p[3] = _a;                                                                 \
  _a = _p[1];                                                                 \
  _p[1] = _p[2];                                                              \
  _p[2] = _a;                                                                 \
}

#define Reverse_double(d) {                                                   \
  char * _p;                                                                  \
  int _a;                                                                     \
  _p = (char *) (d);                                                          \
  _a = _p[0];                                                                 \
  _p[0] = _p[7];                                                              \
  _p[7] = _a;                                                                 \
  _a = _p[1];                                                                 \
  _p[1] = _p[6];                                                              \
  _p[6] = _a;                                                                 \
  _a = _p[2];                                                                 \
  _p[2] = _p[5];                                                              \
  _p[5] = _a;                                                                 \
  _a = _p[3];                                                                 \
  _p[3] = _p[4];                                                              \
  _p[4] = _a;                                                                 \
}


void rev_pointers(start, actual_hd)
     value start;
     header_t actual_hd;
{
  value * p, * q;
  header_t hd;
  mlsize_t n;

  hd = actual_hd;
  p = Op_val(start);
  q = p + Wosize_val(start);
  while (1) {
    n = Wosize_hd(hd);
    switch(Tag_hd(hd)) {
    case String_tag:
    case Abstract_tag:
    case Final_tag:
      p += n;
      break;
    case Double_tag:
      Reverse_double(p);
      p += n;
      break;
    default:
      for( ; n > 0; n--, p++) {
        Reverse_word(p);
      }
    }
    if (p >= q) break;
    Reverse_word(p);
    hd = (header_t) *p++;
  }
}

value intern_val(chan)          /* ML */
     struct channel * chan;
{
  value res;
  unsigned long magic;
  mlsize_t size;
  unsigned long size_in_bytes;
  header_t actual_hd;

  begin_blocking_section ();
  mutex_lock(&chan->mutex);
  magic = getword(chan);
  if (magic != Big_endian_magic_number && magic != Little_endian_magic_number)
    goto error;
  size = getword(chan);
  end_blocking_section ();
  if (size == 0) {
    res = getword(chan);
    mutex_unlock(&chan->mutex);
    if (Is_long(res))
      return res;
    else
      return Atom(res >> 2);
  }
  size--;
  size_in_bytes = size * sizeof(value);
  res = alloc_shr(size);
  begin_blocking_section ();
  if (getblock(chan, (value) &actual_hd, 0, sizeof(actual_hd))
      != sizeof(actual_hd))
    goto error;
  if (getblock(chan, res, 0, size_in_bytes) != size_in_bytes)
    goto error;
  end_blocking_section ();
  end_alloc(res, size, Abstract_tag);
  mutex_unlock(&chan->mutex);
#ifdef BIG_ENDIAN
  if (magic == Little_endian_magic_number) {
    Reverse_word(&actual_hd);
    rev_pointers(res, actual_hd);
  }
#else
  if (magic == Big_endian_magic_number) {
    Reverse_word(&actual_hd);
    rev_pointers(res, actual_hd);
  }
#endif
  adjust_pointers(res, actual_hd);
  return res;

 error:
  mutex_unlock(&chan->mutex);
  end_blocking_section ();
  failwith ("intern");
}
  
