/* main.c : all together, now ! */

#include <fcntl.h>
#include <stdio.h>
#include "config.h"
#include "control.h"
#include "debugger.h"
#include "exec.h"
#include "fail.h"
#include "gc.h"
#include "globals.h"
#include "i_interp.h"
#include "major_gc.h"
#include "mem_misc.h"
#include "memory.h"
#include "mlvalues.h"
#include "private.h"
#include "profil.h"
#include "stacks.h"

header_t first_atoms[256];

static void init_atoms()
{
  int i;
  for(i = 0; i < 256; i++) first_atoms[i] = Make_header(0, i, Blue);
}

static void read_header(fd, hdr)
     int fd;
     struct exec_header * hdr;
{
  unsigned char header_buffer[Exec_comment_size + sizeof(struct exec_header)];
  unsigned char * p;
  int n;
  
  n = read(fd, header_buffer, Exec_comment_size + sizeof(struct exec_header));
  p = header_buffer;
  while (*p != '\012') {
    p++;
    n--;
    if (n <= 0) goto truncated;
  }
  p++;
  n--;

#define Read_int(dest) {                                                      \
  int i;                                                                      \
  if (n < 4) goto truncated;                                                  \
  for ((dest) = 0, i = 4; i > 0; i--) {                                       \
    (dest) = ((dest) << 8) + *p;                                              \
    p++;                                                                      \
  }                                                                           \
  n -= 4;                                                                     \
}
  Read_int(hdr->code_size);
  Read_int(hdr->data_size);
  Read_int(hdr->symbol_size);
  Read_int(hdr->debug_size);

#undef Read_int

  lseek(fd, (unsigned long)(p - header_buffer), 0);
  return;

 truncated:
  fatal_error("truncated bytecode file");
}

code_t start_code;

static void read_code(fd, size)
     int fd;
     mlsize_t size;
{
  start_code = (code_t) stat_alloc(size);
  if (read(fd, start_code, size) != size)
    fatal_error("truncated bytecode file");
}

value global_data;
mutex_t global_data_mutex;

static void read_global_data(fd, size)
     int fd;
     mlsize_t size;
{
  value block;
  mlsize_t block_size;
  header_t actual_hd;
  unsigned long n;

  global_data_mutex = mutex_alloc ();
  if (size < 12 || size & 3 != 0) goto truncated;
  size -= sizeof(unsigned long);  /* Le dernier mot est un ptr vers le res. */
  block_size = (size - sizeof(header_t)) / sizeof(value);
		                 /* Taille du bloc en mots */
  block = alloc_shr(block_size);
  if (read(fd, (char *) &actual_hd, sizeof(header_t)) != sizeof(header_t))
    goto truncated;
  size -= sizeof(header_t);
  if (read(fd, (char *) block, size) != size)
    goto truncated;
  end_alloc(block, block_size, Abstract_tag);
  adjust_pointers(block, actual_hd);
  if (read(fd, (char *) &n, sizeof(n)) != sizeof(n))
    goto truncated;
  global_data = (value) ((unsigned long) block - 4 + n);
  Field(global_data, Global_data) = global_data;
  return;

 truncated:
  fatal_error("Truncated bytecode file");
}

char usage[] =
  "usage: camlrun [-v] [-h heap size] [-g generation size] \n               [-a arg stack size] [-r ret stack size] [-f %free] <file>\n";

int main(argc, argv)
     int argc;
     char * argv[];
{
  struct exec_header hdr;
  int fd;
  int i;
  unsigned long n;
  unsigned long heap_size, generation_size;
  unsigned long arg_size, ret_size;
  jmp_buf raise_buf;

#ifdef PROFIL
  { int i;
    profil *p_empty;

    p_empty = profil_new ("Empty");
    for (i = 0; i < 1000; i++){
      profil_enter (p_empty);
      profil_exit (p_empty);
    }
  }
  profil_enter (profil_new ("Whole_program"));
#endif

  cthread_init ();
  mutex_spin_limit = 1000;
  condition_spin_limit = 1000;

#ifdef TRACE
  tr = (struct trace *) malloc (Trace_buffer_size * sizeof (struct trace));
  if (tr == NULL) fatal_error ("cannot allocate trace buffer");
  for (n = 0; n < Trace_buffer_size; n++){
    tr [n].msg = "";
    tr [n].arg1 = 0;
    tr [n].arg2 = 0;
  }
#endif
  heap_size = Heap_size;
  generation_size = Generation_size;
  arg_size = Arg_stack_size;
  ret_size = Ret_stack_size;
  mem_low_percent = Memory_low_percent;
  verb_gc = 0;

  for(i = 1; i < argc && argv[i][0] == '-'; i++) {
    unsigned long * param;
#ifdef DEBUG
    extern int trace_flag;
#endif
    switch(argv[i][1]) {
    case 'v':
      verb_gc = 1;
      continue;
#ifdef DEBUG
    case 't':
      trace_flag = 1;
      continue;
    case 'T':
      trace_flag = 2;
      continue;
#endif
#ifdef PROFILING
    case 'p':
      { extern char * profile_name;
        ++i;
        profile_name = argv[i];
        continue;
      }
#endif
    case 'h':
      param = &heap_size;
      break;
    case 'g':
      param = &generation_size;
      break;
    case 'a':
      param = &arg_size;
      break;
    case 'r':
      param = &ret_size;
      break;
    case 'f':
      mem_low_percent = atoi(argv[i][2] ? argv[i]+2 : argv[++i]);
      continue;
    default:
      fatal_error("unknown option");
    }
    *param = atoi(argv[i][2] ? argv[i]+2 : argv[++i]) * 1024;
  }

  if (i >= argc) {
    write(2, usage, strlen(usage));
    exit(2);
  }

#ifdef msdos
  fd = open(argv[i], O_RDONLY | O_BINARY);
#else
  fd = open(argv[i], O_RDONLY);
#endif
  if (fd == -1)
    fatal_unix_error("cannot open ", argv[i]);
  read_header(fd, &hdr);

  n = (hdr.data_size + 3) / 4 + 1024;
  heap_size += n;

  { gc_cycle_t starting_gc_cycle;

    if (generation_size < Max_young_size * 2){
      generation_size = Max_young_size * 2;
    }
    init_memory (generation_size, arg_size, ret_size);
    init_atoms();
    starting_gc_cycle = init_heap (heap_size);
    init_interp (starting_gc_cycle);
  }

#ifdef DEBUG
  init_debugger();
#endif

  if (setjmp(raise_buf) == 0) {
    external_raise = (jmp_buf *) raise_buf;

    read_code(fd, hdr.code_size);
    read_global_data(fd, hdr.data_size);
    close(fd);

    unix_init(argc - i, argv + i);
    init_signals();
#ifdef PROFILING
    start_profiling(hdr.code_size);
#endif
      /* The major GC must not start before the global data are read and the
	 main interpreter has registered. */
    start_major_gc ();
    interprete(start_code, Val_long(0));
    unix_exit(0);
  } else {
    if (exn_bucket == Atom(Out_of_memory_exn))
      fatal_error ("out of memory");
    else
      fatal_error ("uncaught exception");
  }
}
