#ifndef _major_gc_
#define _major_gc_


#include "control.h"
#include "private.h"

typedef struct gc_cycle {
  struct mutex lock;		/* Protects [head] and [head->next]. */
  volatile long phase;		/* Accessed without synchronization. */
  struct private_data *head;
} *gc_cycle_t;

#define Phase_mark 0
#define Phase_sweep 1

typedef struct {
  unsigned long size;
  unsigned char *next;
} heap_chunk_head;

extern struct gc_cycle gc_cycles [];
extern gc_cycle_t current_gc_cycle;

extern mutex_t master_fl_mutex;	/* Protects [master_fl]. */
extern condition_t memory_low;
extern mutex_t heap_chunk_mutex; /* Protects the [heap_chunk_head.next]s. */
/* master_fl_mutex < heap_chunk_mutex */

extern unsigned char *master_fl;
extern volatile mlsize_t last_free_size;
extern volatile unsigned long running_procs;

extern unsigned char *heap_start;
  /* These are accessed without synchronization. */
extern unsigned char * volatile heap_end;
extern unsigned char * last_chunk;
extern unsigned long total_heap_size;
extern unsigned char * volatile page_table;
extern volatile unsigned long page_table_size;

#define In_heap 1
#define Not_in_heap 0
#define Page_hp(hp) (((unsigned char *) (hp) - heap_start) / Page_size)
#define Page_val(val) Page_hp (Hp_val (val))
#define Is_in_heap(v) \
  (Bp_val (v) > heap_start && Bp_val (v) <= heap_end \
   && page_table [Page_val (v)] == In_heap)

extern unsigned char * volatile gc_sweep_hp;

#ifdef ANSI
extern gc_cycle_t init_heap (unsigned long);
extern void start_major_gc (void);
#else
gc_cycle_t init_major_gc ();
void start_major_gc ();
#endif


#endif /* _major_gc_ */
