/* mem_misc.c */

/* Miscellaneous functions. */

#include "config.h"
#include "debugger.h"
#include "mem_misc.h"

extern unsigned char *malloc ();

int verb_gc;

#ifdef ANSI
extern char * itoa(char * buf, unsigned long num);
#else
extern char * itoa();
#endif

void gc_message (str1, num, str2)
     char *str1, *str2;
     unsigned long num;
{
  char buff [13];
  char *snum = itoa (buff, num);

  if (verb_gc){
    write (2, str1, strlen (str1));
    write (2, snum, strlen (snum));
    write (2, str2, strlen (str2));
  }
}

void gc_message_str (s)
     char *s;
{
  if (verb_gc){
    write (2, s, strlen (s));
  }
}

void fatal_error (s)
     char *s;
{
  write (2, "Fatal error: ", 13);
  write (2, s, strlen (s));
  write (2, ".\n", 2);
  exit(2);
}

extern int errno;
extern char * error_message();

void fatal_unix_error (s1, s2)
     char * s1, * s2;
{
  char * s;
  write (2, "Fatal error: ", 13);
  write(2, s1, strlen(s1));
  write(2, s2, strlen(s2));
  write(2, ": ", 2);
  s = error_message(errno);
  write(2, s, strlen(s));
  write(2, ".\n", 2);
  exit(2);
}


#ifdef USING_MEMMOV

/* This should work on 64-bit machines as well as 32-bit machines.
   It assumes a long is the natural size for memory reads and writes.
*/
void memmov (dst, src, length)
     uchar *dst, *src;
     unsigned long length;
{
  unsigned long i;

  if ((unsigned long) dst <= (unsigned long) src){
    /* Copy in ascending order. */
    if (((unsigned long) src - (unsigned long) dst) % sizeof (long) != 0){
      /* The pointers are not equal modulo sizeof (long). Copy byte by byte. */
      for (; length > 0; length--){
	*dst++ = *src++;
      }
    }else{
      /* Copy the first few bytes. */
      i = (unsigned long) dst % sizeof (long);
      if (i != 0){
	for (; i < sizeof (long); i++){
	  *dst++ = *src++; --length;
	}
      }
      /* Then copy as many entire words as possible. */
      for (i = length / sizeof (long); i > 0; i--){
	*(long *) dst = *(long *) src;
	dst += sizeof (long); src += sizeof (long);
      }
      /* Then copy the last few bytes. */
      for (i = length % sizeof (long); i > 0; i--){
	*dst++ = *src++;
      }
    }
  }else{
    /* Copy in descending order. */
    src += length; dst += length;
    if (((unsigned long) dst - (unsigned long) src) % sizeof (long) != 0){
      /* The pointers are not equal modulo sizeof (long). Copy byte by byte. */
      for (; length > 0; length--){
	*--dst = *--src;
      }
    }else{
      /* Copy the first few bytes. */
      for (i = (unsigned long) dst % sizeof (long); i > 0; i--){
	*--dst = *--src; --length;
      }
      /* Then copy as many entire words as possible. */
      for (i = length / sizeof (long); i > 0; i--){
	dst -= sizeof (long); src -= sizeof (long);
	*(long *) dst = *(long *) src;
      }
      /* Then copy the last few bytes. */
      for (i = length % sizeof (long); i > 0; i--){
	*--dst = *--src;
      }
    }
  }
}

#endif /* USING_MEMMOV */

uchar *aligned_malloc (size, modulo)
     unsigned long size;
     long modulo;
{
  uchar *raw_mem;
  unsigned long aligned_mem;

  Assert (modulo < Page_size);
  raw_mem = malloc (size + Page_size);
  if (raw_mem == NULL) return NULL;
  raw_mem -= modulo;		/* Address of raw_mem with origin [modulo]. */
  aligned_mem = (((unsigned long) raw_mem / Page_size + 1) * Page_size);
				/* Aligned mem with origin [modulo]. */
  return (uchar *) (aligned_mem + modulo); /* Aligned mem with origin 0. */
}
