/* meta.c : primitives pour le toplevel */

#include "mlvalues.h"
#include "memory.h"
#include "globals.h"

value realloc_global(size)      /* ML */
     value size;
{
  mlsize_t requested_size, actual_size, i;
  Push_roots (new_global_data, 1);

  requested_size = Long_val(size);
  actual_size = Wosize_val(global_data);
  if (requested_size >= actual_size) {
    requested_size = (requested_size + 0x100) & ~0xFF;
    gc_message ("Expanding globals to ", requested_size, " entries.\n");
    new_global_data [0] = alloc_shr (requested_size);
    for (i = 0; i < requested_size; i++){
      Field (new_global_data [0], i) = Val_long (0);
    }
    end_alloc (new_global_data [0], requested_size, 0);
    begin_blocking_section ();
    mutex_lock (global_data_mutex);
    end_blocking_section ();
      for (i = 0; i < actual_size; i++)
        Modify(&Field(new_global_data [0], i), Field(global_data, i));
      Field(new_global_data [0], Global_data) = new_global_data [0];
        /* Ne peut pas creer de racines inverses -> on n'appelle pas MODIFY */
      global_data = new_global_data [0];
    mutex_unlock (global_data_mutex);
  }
  Pop_roots ();
  return Atom(0);
}
    
    
value static_alloc(size)        /* ML */
     value size;
{
  return (value) stat_alloc(Long_val(size));
}

value static_free(blk)          /* ML */
     value blk;
{
  stat_free((uchar *) blk);
  return Atom(0);
}

value static_resize(blk, new_size) /* ML */
     value blk, new_size;
{
  return (value) stat_resize((uchar *) blk, Long_val(new_size));
}

value obj_block(tag, size) /* ML */
     value tag, size;
{
  value res;
  mlsize_t sz, i;
  int tg;

  sz = Long_val(size);
  tg = Long_val(tag);
  if (sz == 0) return Atom(tg);
  res = alloc (sz);
  for (i = 0; i < sz; i++)
    Field(res, i) = Val_long(0);
  end_alloc (res, sz, tg);
  return res;
}

