/* minor_gc.c */

#include "debugger.h"
#include "gc.h"
#include "major_gc.h"
#include "mlvalues.h"
#include "mem_misc.h"
#include "roots.h"
#include "private.h"
#include "profil.h"

unsigned long mem_low_percent;

static value oldify_aux (v)
     value v;
{
  value result;
  unsigned long i;

  if (Is_block (v) && Hp_val (v) >= young_start && Hp_val (v) < young_ptr){
    if (Lhd_val (v).forward != NULL){
      return Lhd_val (v).forward;
    }else if (Tag_val (v) >= No_scan_tag){
#if 0 /* def PROFIL */
      ++survivors;
#endif
      result = alloc_shr (Wosize_val (v));
      Lhd_val (v).forward = result;
      bcopy (Bp_val (v), Bp_val (result), Bosize_val (v));
      end_alloc (result, Wosize_val (v), Tag_val (v));
      return result;
    }else{
      result = alloc_shr (Wosize_val (v));
#if 0 /* def PROFIL */
      ++survivors;
#endif
      Lhd_val (v).forward = result;
      for (i = 0; i < Wosize_val (v); i++){
	Field (result, i) = oldify_aux (Field (v, i));
      }
      end_alloc (result, Wosize_val (v), Tag_val (v));
      return result;
    }
  }else{
    return v;
  }
}

value oldify (v)
     value v;
{
  value result;

  result = oldify_aux (v);
  return result;
}

#ifdef PROFIL
value timed_oldify (v)
     value v;
{
  value result;

  profil_enter (p_modify);
  result = oldify (v);
  profil_exit (p_modify);
  return result;
}
#endif

value oldify_gray (v)
     value v;
{
  value result, fv;
  mlsize_t i;
  header_t hd;

  Assert (Is_block (v));
  if (Hp_val (v) >= young_start && Hp_val (v) < young_ptr){
    hd = Grayhd_hd (Hd_val (v));
    fv = Lhd_val (v).forward;
    if (fv != NULL){
      if (Is_white_val (fv)) Hd_val (fv) = Grayhd_hd (Hd_val (fv));
      return fv;
    }else if (Tag_val (v) >= No_scan_tag){
      result = alloc_shr (Wosize_val (v));
#if 0 /* def PROFIL */
      ++survivors;
#endif
      Lhd_val (v).forward = result;
      bcopy (Bp_val (v), Bp_val (result), Bosize_val (v));
      Hd_val (result) = hd;	/* end_alloc */
      return result;
    }else{
      result = alloc_shr (Wosize_val (v));
#if 0 /* def PROFIL */
      ++survivors;
#endif
      Lhd_val (v).forward = result;
      for (i = 0; i < Wosize_val (v); i++){
	if (Is_block (Field (v, i))){
	  Field (result, i) = oldify_gray (Field (v, i));
	}else{
	  Field (result, i) = Field (v, i);
	}
      }
      Hd_val (result) = hd;	/* end_alloc */
      return result;
    }
  }else if (Is_in_heap (v)){
    if (Is_white_val (v)) Hd_val (v) = Grayhd_hd (Hd_val (v));
    return v;
  }else{
    return v;
  }
}

void gc_minor (root)
     value *root;
{
  if (Is_block (*root)){
    *root = oldify (*root);
  }
}

void gc_minor_gray (root)
     value *root;
{
  if (Is_block (*root)){
    *root = oldify_gray (*root);
  }
}

void gc_minor_blocked (root)
     value *root;
{
  if (Is_block (*root)){
    oldify_gray (*root);
  }
}

void minor_collection ()
{
  gc_message_str ("<");
  profil_enter (p_minor_gc);
  if (local_current_cycle == current_gc_cycle){
    /* Normal case. */
#ifdef PROFIL
    survivors = 0;
#endif
    local_roots (gc_minor);
    profil_count (p_survivors, survivors);
  }else{
    /* A new major gc cycle has started.  We must color all objects in gray. */
#ifdef PROFIL
    survivors = 0;
#endif
    local_roots (gc_minor_gray);
    profil_count (p_survivors, survivors);
    /* We must also tell the major gc we have given our roots. */
    local_current_cycle = current_gc_cycle;
    gc_message_str ("[");
    sem_V (&new_gc_cycle);
  }
  young_ptr = young_start;
  profil_exit (p_minor_gc);
  gc_message_str (">");
}

/* This function is called by the GC thread to get the roots from
   blocked threads. */
void minor_collection_blocked (pdata)
     struct private_data *pdata;
{
  gc_message_str ("{");
  Set_private_data (pdata);
  local_roots (gc_minor_blocked);
  local_current_cycle = current_gc_cycle;
  Set_private_data (NULL);
  gc_message_str ("}");
}
