/* The PDA automaton for parsers generated by camlyacc */

#include "config.h"
#include "mlvalues.h"
#include "memory.h"
#include "private.h"

struct parser_tables {    /* Mirrors parse_tables in ../lib/parsing.mli */
  value actions;
  value transl;
  char * lhs;
  char * len;
  char * defred;
  char * dgoto;
  char * sindex;
  char * rindex;
  char * gindex;
  value tablesize;
  char * table;
  char * check;
};

struct parser_env {       /* Mirrors parser_env in ../lib/iparsing.mli */
  value s_stack;
  value v_stack;
  value symb_start_stack;
  value symb_end_stack;
  value stacksize;
  value curr_char;
  value lval;
  value symb_start;
  value symb_end;
  value sp;
  value lhs_start;
  value lhs_end;
  value rule_number;
};

#ifdef BIG_ENDIAN
#define Short(tbl,n) \
  (*((unsigned char *)((tbl) + (n) * sizeof(short))) + \
          (*((schar *)((tbl) + (n) * sizeof(short) + 1)) << 8))
#else
#define Short(tbl,n) (((short *)(tbl))[n])
#endif

#ifdef DEBUG
int parser_trace = 0;
#define Parser_trace(act) if(parser_trace) act
#else
#define Parser_trace(act)
#endif

/* Input codes */

#define START 0            /* Mirrors parser_input in ../lib/iparsing.mli */
#define TOKEN_READ 1
#define STACKS_GROWN_1 2
#define STACKS_GROWN_2 3
#define SEMANTIC_ACTION_COMPUTED 4

/* Output codes */

#define READ_TOKEN Atom(0) /* Mirrors parser_output in ../lib/iparsing.mli */
#define RAISE_PARSE_ERROR Atom(1)
#define GROW_STACKS_1 Atom(2)
#define GROW_STACKS_2 Atom(3)
#define COMPUTE_SEMANTIC_ACTION Atom(4)

/* The pushdown automata */

value parse_engine(tables, env, cmd, arg) /* ML */
     struct parser_tables * tables;
     struct parser_env * env;
     value cmd;
     value arg;
{
  int n, n1, n2, m, state1;
  mlsize_t sp2;

  switch(Tag_val(cmd)) {

  case START:
    parsing_state = 0;
    parsing_sp = Long_val(env->sp);

  loop:
    Parser_trace(printf("Loop %d\n", parsing_state));
    n = Short(tables->defred, parsing_state);
    if (n != 0) goto reduce;
    if (Long_val(env->curr_char) >= 0) goto testshift;
    return READ_TOKEN;
                                /* The ML code calls the lexer and updates */
                                /* symb_start and symb_end */
  case TOKEN_READ:
    env->curr_char = Field(tables->transl, Tag_val(arg));
    if (Wosize_val(arg) == 0) {
      env->lval = Val_long(0);
    } else {
      Modify(&env->lval, Field(arg, 0));
    }
    Parser_trace(printf("Token %d (0x%lx)\n", Long_val(env->curr_char), env->lval));
    
  testshift:
    n1 = Short(tables->sindex, parsing_state);
    n2 = n1 + Long_val(env->curr_char);
    if (n1 != 0 && n2 >= 0 && n2 <= Long_val(tables->tablesize) &&
        Short(tables->check, n2) == Long_val(env->curr_char)) goto shift;
    n1 = Short(tables->rindex, parsing_state);
    n2 = n1 + Long_val(env->curr_char);
    if (n1 != 0 && n2 >= 0 && n2 <= Long_val(tables->tablesize) &&
        Short(tables->check, n2) == Long_val(env->curr_char)) {
      n = Short(tables->table, n2);
      goto reduce;
    }
    return RAISE_PARSE_ERROR;
                                /* The ML code raises the Parse_error exn */
  shift:
    parsing_state = Short(tables->table, n2);
    Parser_trace(printf("Shift %d\n", parsing_state));
    parsing_sp++;
    if (parsing_sp < Long_val(env->stacksize)) goto push;
    return GROW_STACKS_1;
                                /* The ML code resizes the stacks */
  case STACKS_GROWN_1:
  push:
    Field(env->s_stack, parsing_sp) = Val_long(parsing_state);
    Modify(&Field(env->v_stack, parsing_sp), env->lval);
    Field(env->symb_start_stack, parsing_sp) = env->symb_start;
    Field(env->symb_end_stack, parsing_sp) = env->symb_end;
    env->curr_char = Val_long(-1);
    goto loop;

  reduce:
    Parser_trace(printf("Reduce %d\n", n));
    sp2 = parsing_sp - Short(tables->len, n) + 1;
    env->lhs_start = Field(env->symb_start_stack, sp2);
    env->lhs_end = Field(env->symb_end_stack, parsing_sp);
    env->sp = Val_long(parsing_sp);
    env->rule_number = Val_long(n);
    parsing_sp = sp2;
    m = Short(tables->lhs, n);
    state1 = Long_val(Field(env->s_stack, parsing_sp - 1));
    n1 = Short(tables->gindex, m);
    n2 = n1 + state1;
    if (n1 != 0 && n2 >= 0 && n2 <= Long_val(tables->tablesize) &&
        Short(tables->check, n2) == state1) {
      parsing_state = Short(tables->table, n2);
    } else {
      parsing_state = Short(tables->dgoto, m);
    }
    if (parsing_sp < Long_val(env->stacksize)) goto semantic_action;
    return GROW_STACKS_2;
                                /* The ML code resizes the stacks */
  case STACKS_GROWN_2:
  semantic_action:
    return COMPUTE_SEMANTIC_ACTION;
                                /* The ML code calls the semantic action */
  case SEMANTIC_ACTION_COMPUTED:
    Field(env->s_stack, parsing_sp) = Val_long(parsing_state);
    Modify(&Field(env->v_stack, parsing_sp), arg);
    Field(env->symb_end_stack, parsing_sp) = env->lhs_end;
    goto loop;
  }
}
