/* private.h */

#ifndef _private_
#define _private_


#include "control.h"
#include "debugger.h"
#include "freelist.h"
#include "intext.h"
#include "major_gc.h"
#include "mlvalues.h"
#include "profil.h"
#include "semaph.h"

typedef struct private_data {
  struct private_data *next;
  int pd_am_dead;
  struct gc_cycle *pd_local_current_cycle;
  struct semaphore pd_new_gc_cycle;
  int pd_force_minor_gc;
  struct mutex pd_running;
  jmp_buf *pd_external_raise;
  unsigned char *pd_young_start;
  unsigned char *pd_young_end;
  unsigned char *pd_young_ptr;
  free_list_t pd_private_fl;
  value *pd_arg_stack_low;
  value *pd_arg_stack_high;
  value *pd_arg_stack_threshold;
  value *pd_ret_stack_low;
  value *pd_ret_stack_high;
  value *pd_ret_stack_threshold;
  value *pd_extern_asp;
  value *pd_extern_rsp;
  struct trap_frame *pd_gtp;
  value *pd_c_roots_head;
  value pd_exn_bucket;
  unsigned long * pd_extern_block;
  unsigned long pd_extern_size, pd_extern_pos;
  struct extern_obj * pd_extern_table;
  unsigned long pd_extern_table_size, pd_extern_table_used;
  unsigned long pd_hash_accu, pd_hash_univ_limit, pd_hash_univ_count;
  jmp_buf pd_hash_exit;
  value pd_per_thread_data;
  int pd_parsing_state;
  int pd_parsing_sp;
#ifdef PROFILING
  code_t pd_pc;
#endif
#ifdef PROFIL
  profil *pd_p_modify;
  profil *pd_p_minor_gc;
  unsigned long pd_survivors;
  profil *pd_p_survivors;
#endif
#ifdef DEBUG
  code_t pd_log_buffer[Log_buffer_size];
  code_t * pd_log_ptr;
#endif
} private_data;

#ifdef __GNUC__
#ifdef ns32000
register private_data * priv_data asm("r7");
#define PRIVATE_DATA_REG
#endif
#endif

#ifdef PRIVATE_DATA_REG
#define Set_private_data(pdata) priv_data = pdata
#else
#define Set_private_data(pdata) \
  cthread_set_data (cthread_self (), (any_t) pdata)
#define priv_data ((private_data *) (cthread_data (cthread_self ())))
#endif

#define am_dead (priv_data->pd_am_dead)
#define local_current_cycle (priv_data->pd_local_current_cycle)
#define new_gc_cycle (priv_data->pd_new_gc_cycle)
#define force_minor_gc (priv_data->pd_force_minor_gc)
#define running (priv_data->pd_running)
#define external_raise (priv_data->pd_external_raise)
#define young_start (priv_data->pd_young_start)
#define young_end (priv_data->pd_young_end)
#define young_ptr (priv_data->pd_young_ptr)
#define private_fl (priv_data->pd_private_fl)
#define arg_stack_low (priv_data->pd_arg_stack_low)
#define arg_stack_high (priv_data->pd_arg_stack_high)
#define arg_stack_threshold (priv_data->pd_arg_stack_threshold)
#define ret_stack_low (priv_data->pd_ret_stack_low)
#define ret_stack_high (priv_data->pd_ret_stack_high)
#define ret_stack_threshold (priv_data->pd_ret_stack_threshold)
#define extern_asp (priv_data->pd_extern_asp)
#define extern_rsp (priv_data->pd_extern_rsp)
#define gtp (priv_data->pd_gtp)
#define c_roots_head (priv_data->pd_c_roots_head)
#define exn_bucket (priv_data->pd_exn_bucket)
#define extern_block (priv_data->pd_extern_block)
#define extern_size (priv_data->pd_extern_size)
#define extern_pos (priv_data->pd_extern_pos)
#define extern_table (priv_data->pd_extern_table)
#define extern_table_size (priv_data->pd_extern_table_size)
#define extern_table_used (priv_data->pd_extern_table_used)
#define hash_accu (priv_data->pd_hash_accu)
#define hash_univ_limit (priv_data->pd_hash_univ_limit)
#define hash_univ_count (priv_data->pd_hash_univ_count)
#define hash_exit (priv_data->pd_hash_exit)
#define per_thread_data (priv_data->pd_per_thread_data)
#define parsing_state (priv_data->pd_parsing_state)
#define parsing_sp (priv_data->pd_parsing_sp)

#ifdef PROFILING
#define pc (priv_data->pd_pc)
#endif
#ifdef PROFIL
#define p_modify (priv_data->pd_p_modify)
#define p_minor_gc (priv_data->pd_p_minor_gc)
#define survivors (priv_data->pd_survivors)
#define p_survivors (priv_data->pd_p_survivors)
#endif
#ifdef DEBUG
#define log_buffer (priv_data->pd_log_buffer)
#define log_ptr (priv_data->pd_log_ptr)
#endif


#endif /* _private_ */
