#include <sys/time.h>
#include <signal.h>
#include <fcntl.h>
#include <cthreads.h>
#include "mlvalues.h"
#include "private.h"

#ifdef PROFILING

struct thread_profile {
  cthread_t pid;
  unsigned * table;
};
  
code start_code;
unsigned size_code;
char * profile_name = 0;
struct thread_profile profile_data[10];
unsigned num_threads;

void sigalrm_handler()
{
  unsigned i;
  code_t p;

  for (i = 0; i < num_threads; i++) {
    p = ((private_data *) cthread_data(profile_data[i].pid))->pd_pc;
    profile_data[i].table[p - start_code]++;
  }
}

struct mutex profile_mutex;

void register_thread(pid)
     cthread_t pid;
{
  unsigned * p;
  unsigned i;

  mutex_lock(&profile_mutex);
  if (num_threads < sizeof(profile_data) / sizeof(struct thread_profile)) {
    profile_data[num_threads].pid = pid;
    profile_data[num_threads].table =
      (unsigned *) malloc(code_size * sizeof(unsigned));
    for (p = profile_data[num_threads].table, size = code_size;
         size > 0;
         size--, p++) *p = 0;
    num_threads++;
  }
  mutex_unlock(&profile_mutex);
}

void start_profile(size)
     unsigned size;
{
  int fd;
  unsigned * p;
  struct itimerval timerval;

  if (profile_name) {
    size_code = size;
    num_threads = 0;
    mutex_init(&profile_mutex);
    register_thread(cthread_self());
    pc = start_code;
    signal(SIGALRM, sigalrm_handler);
    timerval.it_interval.tv_sec = 0;
    timerval.it_interval.tv_usec = 10000;
    timerval.it_value = timerval.it_interval;
    setitimer(ITIMER_REAL, &timerval, 0);
  }
}

void stop_profile()
{
  int fd;
  struct itimerval timerval;
  unsigned i;
  int n;
  char file_name[256];

  if (profile_name) {
    signal(SIGALRM, SIG_IGN);
    timerval.it_interval.tv_sec = 0;
    timerval.it_interval.tv_usec = 0;
    timerval.it_value = timerval.it_interval;
    setitimer(ITIMER_REAL, &timerval, 0);
    mutex_lock(&profile_mutex);
    n = strlen(profile_name);
    strcpy(file_name, profile_name);
    file_name[n+1] = '\0';
    for (i = 0; i < num_threads; i++) {
      file_name[n] = 'A' + i;
      fd = open(file_name, O_WRONLY | O_CREAT | O_TRUNC, 0666);
      write(fd, profile_data[i].table, size_code * sizeof(unsigned));
      close(fd);
    }
    mutex_unlock(&profile_mutex);
  }
}

#endif
