#ifdef PROFIL

#include <stdio.h>
#include <sys/time_stamp.h>
char *malloc ();

#include "profil.h"

static profil *head = NULL;

profil *profil_new (name)
     char *name;
{
  profil *result;
  int i;

  result = (profil *) malloc (sizeof (profil));
  result->total = 0;
  result->max = 0;
  result->num = 0;
  result->lastentry.low_val = 0;
  for (i = 0; i < 32; i++) result->histo[i] = 0;
  result->name = name;
  result->next = head;
  head = result;
  return result;
}

void profil_enter (p)
     profil *p;
{
  if (p->lastentry.low_val != 0){
    fprintf (stderr, "profil_enter: double entry into \"%s\".\n", p->name);
  }else{
    kern_timestamp (&(p->lastentry));
    if (p->lastentry.low_val == 0) p->lastentry.low_val = 1;
  }
}

void profil_exit (p)
     profil *p;
{
  struct tsval tv;
  unsigned long cur;
  int i;
  unsigned long max;

  if (p->lastentry.low_val == 0){
    fprintf (stderr, "profil_exit: exiting \"%s\", not entered.\n", p->name);
  }else{
    kern_timestamp (&tv);
    cur = tv.low_val - p->lastentry.low_val;
    p->total += cur;
    ++p->num;
    if (cur > p->max) p->max = cur;
    p->lastentry.low_val = 0;
    i = 0;
    max = 2;
    while (cur >= max) {
      i++;
      max <<= 1;
    }
    ++(p->histo[i]);
  }
}

void profil_count (p, n)
     profil *p;
     unsigned long n;
{
  int i;
  unsigned long max;

  p->total += n;
  ++p->num;
  if (n > p->max) p->max = n;
  i = 0;
  max = 1;
  n >>= 1;
  while (n >= max){
    i++;
    max <<= 1;
  }
  ++(p->histo [i]);
}

void profil_finish ()
{
  profil *cur;
  int i;

  for (cur = head; cur != NULL; cur = cur->next){
    if (cur->lastentry.low_val != 0){
      profil_exit (cur);
      fprintf (stderr, "profil_finish: exiting \"%s\".\n", cur->name);
    }
  }
  for (cur = head; cur != NULL; cur = cur->next){
    fprintf (stderr, "%s: max %lu\n", cur->name, cur->max);
    fprintf (stderr, "%s: tot %lu\n", cur->name, cur->total);
    fprintf (stderr, "%s: num %lu\n", cur->name, cur->num);
    fprintf (stderr, "%s: avg %lu\n", cur->name,
	     cur->num != 0 ? cur->total / cur->num : 0);
    fprintf (stderr, "%s: hist", cur->name);
    for (i = 0; i < 24; i++) fprintf(stderr, " %lu", cur->histo[i]);
    fprintf (stderr, "\n");
  }
}

#endif /* PROFIL */
