/* Gestion des signaux */

#include <signal.h>
#include <sys/time.h>
#include <fcntl.h>
#include "private.h"
#include "control.h"
#include "fail.h"
#include "globals.h"
#include "memory.h"
#include "mlvalues.h"
#include "signals.h"

int signal_is_pending;
int block_signals;
int raise_on_signal;
value pending_signal;

#ifndef NSIG
#define NSIG 32
#endif

void init_signals()
{
  value signal_table;
  int i;

  signal_is_pending = 0;
  pending_signal = Val_long (0);
  block_signals = 0;
  raise_on_signal = 0;
  signal_table = alloc_shr (NSIG);
  for (i = 0; i < NSIG; i++)
    Field(signal_table, i) = Atom(Bottom_exn);
  end_alloc (signal_table, NSIG, 0);
  Modify(&Field(global_data, Signal_table), signal_table);
}

void raise_pending_signal()
{
  value exc;
  exc = pending_signal;
  pending_signal = Val_long (1);
  signal_is_pending = 0;
  raise_on_signal = 0;
  mlraise(exc);
}

#ifdef ANSI
static void signal_handler(sig)
#else
static signal_handler(sig)
#endif
     int sig;
{
  value signal_table;

#ifndef BSD_SIGNALS
  (void) signal (sig, signal_handler);
#endif
  if (block_signals) return;
  signal_table = Field(global_data, Signal_table);
  if (sig >= 0 && sig < Wosize_val(signal_table)) {
    pending_signal = Field(signal_table, sig);
    signal_is_pending = 1;
    if (raise_on_signal){
      end_blocking_section ();
      raise_pending_signal();
    }
  }
}

value set_signal(sign, behavior)	/* ML */
     value sign, behavior;
{
  value signal_table;
  int sig;

  signal_table = Field(global_data, Signal_table);
  sig = Long_val(sign);
  if (sig < 0) {
    block_signals = (Tag_val(behavior) == 1);
  } else {
    switch(Tag_val(behavior)) {
    case 0:                       /* Signal_default */
      signal(sig, SIG_DFL);
      Field(signal_table, sig) = Atom(Bottom_exn);
      break;
    case 1:                       /* Signal_ignore */
      signal(sig, SIG_IGN);
      Field(signal_table, sig) = Atom(Bottom_exn);
      break;
    case 2:                       /* Signal_raise */
      Modify(&Field(signal_table, sig), Field(behavior, 0));
      signal(sig, signal_handler);
      break;
    }
  }
  return Atom(0);
}

#ifdef PROFILING

struct thread_profile {
  cthread_t pid;
  unsigned * table;
};
  
code_t start_code;
unsigned size_code;
char * profile_name = 0;
struct thread_profile profile_data[10];
unsigned num_threads;

void sigalrm_handler()
{
  unsigned i;
  code_t p;

  for (i = 0; i < num_threads; i++) {
    p = ((private_data *) cthread_data(profile_data[i].pid))->pd_pc;
    profile_data[i].table[p - start_code]++;
  }
}

struct mutex profile_mutex;

void register_thread(pid)
     cthread_t pid;
{
  unsigned * p;
  unsigned i;

  mutex_lock(&profile_mutex);
  if (num_threads < sizeof(profile_data) / sizeof(struct thread_profile)) {
    profile_data[num_threads].pid = pid;
    profile_data[num_threads].table =
      (unsigned *) malloc(size_code * sizeof(unsigned));
    for (p = profile_data[num_threads].table, i = size_code;
         i > 0;
         i--, p++) *p = 0;
    num_threads++;
  }
  mutex_unlock(&profile_mutex);
}

void start_profiling(size)
     unsigned size;
{
  int fd;
  unsigned * p;
  struct itimerval timerval;

  if (profile_name) {
    size_code = size;
    num_threads = 0;
    mutex_init(&profile_mutex);
    register_thread(cthread_self());
    pc = start_code;
    signal(SIGALRM, sigalrm_handler);
    timerval.it_interval.tv_sec = 0;
    timerval.it_interval.tv_usec = 10000;
    timerval.it_value = timerval.it_interval;
    setitimer(ITIMER_REAL, &timerval, 0);
  }
}

void stop_profiling()
{
  int fd;
  struct itimerval timerval;
  unsigned i;
  int n;
  char file_name[256];

  if (profile_name) {
    signal(SIGALRM, SIG_IGN);
    timerval.it_interval.tv_sec = 0;
    timerval.it_interval.tv_usec = 0;
    timerval.it_value = timerval.it_interval;
    setitimer(ITIMER_REAL, &timerval, 0);
    mutex_lock(&profile_mutex);
    n = strlen(profile_name);
    strcpy(file_name, profile_name);
    file_name[n+1] = '\0';
    for (i = 0; i < num_threads; i++) {
      file_name[n] = 'A' + i;
      fd = open(file_name, O_WRONLY | O_CREAT | O_TRUNC, 0666);
      write(fd, profile_data[i].table, size_code * sizeof(unsigned));
      close(fd);
    }
    mutex_unlock(&profile_mutex);
  }
}

#endif
