/* Initialisation des piles et de la table des globaux */

#include "config.h"
#include "debugger.h"
#include "fail.h"
#include "mem_misc.h"
#include "mlvalues.h"
#include "private.h"
#include "stacks.h"

static void realloc_arg_stack()
{        
  unsigned long new_size;
  value * new_low, * new_high, * new_asp;
  struct trap_frame * p;

  Assert(extern_asp >= arg_stack_low);
  new_size = (arg_stack_high - arg_stack_low) * 2;
  gc_message("Growing argument stack to ",
             new_size * sizeof(value) / 1024,
             "k.\n");
  new_low = (value *) stat_alloc(new_size * sizeof(value));
  new_high = new_low + new_size;
  new_asp = new_high - (arg_stack_high - extern_asp);

  bcopy((unsigned char *) extern_asp,
        (unsigned char *) new_asp,
        (arg_stack_high - extern_asp) * sizeof(value));
  stat_free((char_ptr) arg_stack_low);
  for (p = gtp; p < (struct trap_frame *) ret_stack_high; p = p->tp)
    p->asp = new_high - (arg_stack_high - p->asp);
  arg_stack_low = new_low;
  arg_stack_high = new_high;
  arg_stack_threshold = arg_stack_low + Arg_stack_threshold;
  extern_asp = new_asp;
}

static void realloc_ret_stack()
{        
  unsigned long new_size;
  value * new_low, * new_high, * new_rsp;
  struct trap_frame * p;

  Assert(extern_rsp >= ret_stack_low);
  new_size = (ret_stack_high - ret_stack_low) * 2;
  gc_message("Growing return stack to ",
             new_size * sizeof(value) / 1024,
             "k.\n");
  new_low = (value *) stat_alloc(new_size * sizeof(value));
  new_high = new_low + new_size;
  new_rsp = new_high - (ret_stack_high - extern_rsp);
  bcopy((uchar *) extern_rsp,
        (uchar *) new_rsp,
        (ret_stack_high - extern_rsp) * sizeof(value));
  stat_free((char_ptr) ret_stack_low);
  gtp = (struct trap_frame *) (new_high - (ret_stack_high - (value *) gtp));
  for (p = gtp; p < (struct trap_frame *) new_high; p = p->tp) {
    p->tp = (struct trap_frame *)
      (new_high - (ret_stack_high - (value *) p->tp));
  }
  ret_stack_low = new_low;
  ret_stack_high = new_high;
  ret_stack_threshold = ret_stack_low + Ret_stack_threshold;
  extern_rsp = new_rsp;
}

void realloc_stacks()
{
  if (extern_rsp < ret_stack_threshold)
    realloc_ret_stack();
  if (extern_asp < arg_stack_threshold)
    realloc_arg_stack();
}

