/* unix.c : appels systeme Unix */

#include <fcntl.h>
#include "config.h"
#include "control.h"
#include "debugger.h"
#include "globals.h"
#include "memory.h"
#include "mlvalues.h"
#include "profil.h"
#include "stacks.h"

#ifdef macintosh

char * error_message(err)
     long err;
{
  extern char * strerror();
  
  return strerror(err);
}

#else

char * error_message(err)
     long err;
{
  extern int sys_nerr;
  extern char * sys_errlist [];
  if (err < 0 || err >= sys_nerr)
    return "unknown error";
  else
    return sys_errlist[err];
}

#endif /* macintosh */

value os_error_message(errcode)    /* ML */
     value errcode;
{
  return alloc_string(error_message(Long_val(errcode)));
}

extern int errno;

value os_errno()                /* ML */
{
  return Val_long(errno);
}

value unix_exit(status)         /* ML */
     value status;
{
#ifdef PROFILING
  stop_profiling();
#endif
#ifdef PROFIL
  profil_finish ();
#endif
#ifdef TRACE
  { extern void something_went_wrong ();
    something_went_wrong ();
  }
#endif
  exit(Long_val(status));
  return Val_long (0);		/* Not reached. */
}

value unix_open(path, flags, mode) /* ML */
     value path, flags, mode;
{
#ifdef macintosh
  return Val_long (open (String_val (path), Long_val (flags)));
#else
#ifdef msdos
  return Val_long (open (String_val (path),
                         (Long_val (flags) | O_BINARY),
			 (int) Long_val (mode)));
#else
  return Val_long (open (String_val (path), Long_val (flags),
			 Long_val (mode)));
#endif /* msdos */
#endif /* macintosh */
}

value unix_close(des)           /* ML */
     value des;
{
  return Val_long (close (Long_val (des)));
}

value unix_read(d, buf, offset, nbytes) /* ML */
     value nbytes, offset, buf, d;
{
  begin_blocking_section ();
  return Val_long (read (Long_val (d), &Byte(buf, Long_val (offset)),
			 Long_val (nbytes)));
  end_blocking_section ();
}

value unix_write(d, buf, offset, nbytes) /* ML */
     value nbytes, offset, buf, d;
{
  begin_blocking_section ();
  return Val_long (write (Long_val (d), &Byte(buf, Long_val (offset)),
			  Long_val (nbytes)));
  end_blocking_section ();
}

value unix_lseek(des, offset, whence) /* ML */
     value whence, offset, des;
{
  return
    Val_long (lseek (Long_val (des), Long_val (offset), Long_val (whence)));
}

value unix_unlink(name)         /* ML */
     value name;
{
  return Val_long(unlink(String_val(name)));
}

/* Ce qui suit n'est en general defini que sur les vraies Unix boxes. */

#ifdef UNIX

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/times.h>

value unix_truncate(path, length) /* ML */
     value path, length;
{
  return Val_long(truncate(String_val(path), Long_val(length)));
}

static value stat_aux(buf)
     struct stat * buf;
{
  value v;
  v = alloc_small_unshr (11, 0);
  Field (v, 0) = Val_long (buf->st_dev);
  Field (v, 1) = Val_long (buf->st_ino);
  Field (v, 2) = Val_long (buf->st_mode);
  Field (v, 3) = Val_long (buf->st_nlink);
  Field (v, 4) = Val_long (buf->st_uid);
  Field (v, 5) = Val_long (buf->st_gid);
  Field (v, 6) = Val_long (buf->st_rdev);
  Field (v, 7) = Val_long (buf->st_size);
  Field (v, 8) = Val_long (buf->st_atime);
  Field (v, 9) = Val_long (buf->st_mtime);
  Field (v, 10) = Val_long (buf->st_ctime);
  return v;
}

value unix_stat(path)           /* ML */
     value path;
{
  struct stat stat_buf;

  if (stat(String_val(path), &stat_buf) == -1)
    os_error();
  return stat_aux(&stat_buf);
}

value unix_fstat(fd)           /* ML */
     value fd;
{
  struct stat stat_buf;

  if (fstat(Long_val(fd), &stat_buf) == -1)
    os_error();
  return stat_aux(&stat_buf);
}

value unix_link(name1, name2)   /* ML */
     value name1, name2;
{
  return Val_long(link(String_val(name1), String_val(name2)));
}

value unix_pipe()               /* ML */
{
  int fd[2];
  value res;

  if (pipe(fd) == -1)
    os_error();
  res = alloc_small_unshr(2, 0);
  Field(res, 0) = Val_long(fd[0]);
  Field(res, 1) = Val_long(fd[1]);
  return res;
}

value mlunix_fork()		/* ML */
{
  return Val_long(fork());
}

value unix_execv(name, args)    /* ML */
     value name, args;
{
  char ** argv;
  mlsize_t argc, i;

  argc = Wosize_val(args);
  argv = (char **) stat_alloc((argc + 1) * sizeof(char *));
  for (i = 0; i < argc; i++)
    argv[i] = (char *) String_val(Field(args, i));
  argv[argc] = 0;
  execv(String_val(name), argv);
  free(argv);
  return Atom(0);
}

value unix_wait()               /* ML */
{
  int status, pid;
  value res;

  status = 0;
  begin_blocking_section ();
  pid = wait(&status);
  end_blocking_section ();
  res = alloc_small_unshr(2, 0);
  Field(res, 0) = Val_long(pid);
  Field(res, 1) = Val_long(status);
  return res;
}

value unix_times()              /* ML */
{
  value res;
  struct tms buffer;
  long real_time;
  int i;
  Push_roots(t,5);

#ifndef HZ
#define HZ 60
#endif

  times(&buffer);
  time(&real_time);
  t[0] = alloc_double((double) real_time);
  t[1] = alloc_double((double) buffer.tms_utime / HZ);
  t[2] = alloc_double((double) buffer.tms_stime / HZ);
  t[3] = alloc_double((double) buffer.tms_cutime / HZ);
  t[4] = alloc_double((double) buffer.tms_cstime / HZ);
  res = alloc_small_unshr(5, 0);
  for (i = 0; i < 5; i++)
    Field(res, i) = t[i];
  Pop_roots();
  return res;
}

#else /* UNIX */

/* The return statements are here to suppress warnings from smart compilers. */

value unix_truncate() { failwith("unix_truncate"); return Val_long (0); }
value unix_stat() { failwith("unix_stat"); return Val_long (0); }
value unix_fstat() { failwith("unix_fstat"); return Val_long (0); }
value unix_link() { failwith("unix_link"); return Val_long (0); }
value unix_pipe() { failwith("unix_pipe"); return Val_long (0); }
value unix_fork_process() { failwith("unix_fork"); return Val_long (0); }
value unix_execv() { failwith("unix_execv"); return Val_long (0); }
value unix_wait() { failwith("unix_wait"); return Val_long (0); }
value unix_times() { failwith("unix_times"); return Val_long (0); }

#endif /* UNIX */

static int system_flags[] =
{ O_RDONLY, O_WRONLY, O_RDWR, O_CREAT, O_TRUNC, O_EXCL,
#ifdef TURBOC
  S_IREAD, S_IWRITE, S_IEXEC, 1, 0, 0, S_IREAD+S_IWRITE+S_IEXEC, 1
#else
  4, 2, 1, 01, 010, 0100, 7, 0111
#endif
};

void unix_init(argc, argv)
     int argc;
     char * argv [];
{
  int i;
  value tmp;
  Push_roots (v, 1);

  if (argc <= 0){
    v [0] = Atom (0);
  }else{
    v [0] = alloc_shr (argc);
    for (i = 0; i < argc; i++){
      Field (v [0], i) = Val_long (0);
    }
    end_alloc (v [0], argc, 0);
    for (i = 0; i < argc; i++){
      tmp = alloc_string (argv [i]);
      Modify (&Field (v [0], i), tmp);
    }
  }
  Modify(&Field(global_data, Unix_command_line), v [0]);
  for (i = Mode_rdonly; i <= Rights_everybody; i++)
    Field(global_data, i) = Val_long(system_flags[i - Mode_rdonly]);
  Pop_roots ();
}
