(* mecfile for the Caml Light compiler. *)

#open "mec";;
#open "rules_ml";;

let ZC = ref "camlc >/dev/null";;
let ZFLAGS = ref "-O fast";;
let ZCONFIG = ref "";;
let YACC = ref "../mlyacc -s";;
let LEX = ref "../run ../mllex";;

let OBJS =
  ["config.zo"; "misc.zo"; "const.zo"; "globals.zo";
   "location.zo"; "syntax.zo";
   "errors.zo"; "hashtbl.zo"; "modules.zo"; "builtins.zo"; "types.zo";
   "pr_type.zo"; "ty_error.zo"; "typing.zo"; "ty_decl.zo"; "ty_intf.zo";
   "match.zo"; "front.zo"; "instruct.zo"; "back.zo";
   "opcodes.zo"; "prim_c.zo"; "prim_opc.zo";
   "buffcode.zo"; "labels.zo"; "predef.zo"; "reloc.zo"; "emitcode.zo";
   "primdecl.zo"; "lexer.zo"; "par_aux.zo"; "parser.zo"; "compiler.zo";
   "symtable.zo";
   "emitdata.zo"; "link.zo"; "readword.zo"
];;

let ZC_LIBS =
  ["../lib/eq.zo"; "../lib/exc.zo"; "../lib/fstring.zo"; "../lib/fchar.zo";
   "../lib/list.zo"; "../lib/fvect.zo"; "../lib/io.zo"; "../lib/pair.zo";
   "../lib/lexing.zo"; "../lib/parsing.zo"
];;

let ZC_OBJS = OBJS @ ["version.zo"; "main.zo"];;

let ZINC_LIBS =
  ["../lib/eq.zo"; "../lib/exc.zo"; "../lib/fstring.zo";
   "../lib/cstring.zo"; "../lib/fchar.zo"; "../lib/cchar.zo";
   "../lib/list.zo"; "../lib/fvect.zo"; "../lib/cvect.zo"; "../lib/io.zo";
   "../lib/pair.zo"; "../lib/signal.zo";
   "../lib/lexing.zo"; "../lib/parsing.zo"
];;

let ZINC_OBJS = OBJS @ ["version.zo"; "pr_value.zo"; "toplevel.zo"];;

rule "programs" ["zc"; "zinc"; "../lib/toplvl.zi"] [];;

rule "zc" ZC_OBJS
  [command ([!ZC; !ZCONFIG; "-O none -o zc -I ."] @ ZC_LIBS @ ZC_OBJS)]
;;

rule "zinc" ("toplevel.zi" :: "expunge" :: ZINC_OBJS)
  [command ([!ZC; !ZCONFIG; "-O none -o zinc -I . -g"] @ ZINC_LIBS @ ZINC_OBJS);
   command ["../run expunge zinc"]]
;;

rule "../lib/toplvl.zi" ["toplevel.zi"]
  [command ["cp toplevel.zi ../lib/toplvl.zi"]]
;;

let EXPUNGE_LIBS =
  ["../lib/exc.zo"; "../lib/io.zo"; "../lib/list.zo"; "../lib/fvect.zo"]
;;
let EXPUNGE_OBJS = ["hashtbl.zo"; "expunge.zo"];;

rule "expunge" EXPUNGE_OBJS
  [command ([!ZC; "-O none -I . -o expunge"] @ EXPUNGE_LIBS @ EXPUNGE_OBJS)]
;;

force "clean" []
  [command ["rm -f *.zi & rm -f *.zo & rm -f expunge zc zinc ../lib/toplvl.zi \
             & wait"]]
;;

force "scratch" []
  [command ["rm -f *.zi & rm -f *.zo \
             & rm -f opcodes.ml prim_c.ml predef.ml \
                     parser.ml parser.mli lexer.ml \
             & rm -f expunge zc zinc ../lib/toplvl.zi \
             & wait"]]
;;

rule "opcodes.ml" ["make_instruct"; "../runtime/instruct.h"]
  [command ["./make_instruct ../runtime/instruct.h opcodes.ml"]]
;;

rule "prim_c.ml" ["../runtime/primitives"]
  [command ["(echo 'let c_primitives_list = ['; \
             sed -e 's/.*/  \"&\";/' -e '$s/;$//' ../runtime/primitives; \
             echo '];;') > prim_c.ml"]]
;;

rule "predef.ml" ["../runtime/globals.h"; "../runtime/fail.h"]
  [command ["(echo '#open \"const\";;'; \
             echo 'let predef_variables = ['; \
             sed -n -e 's|.*/\* \(\".*\"\),\(\".*\"\) \*/$|{qual=\1; id=\2};|p' ../runtime/globals.h\
             | sed -e '$s|;$||'; \
             echo '];;'; \
             echo 'let predef_exn = ['; \
             sed -n -e 's|.*/\* \(\".*\"\),\(\".*\"\) \*/$|{qual=\1; id=\2};|p' ../runtime/fail.h \
             | sed -e '$s|;$||'; \
             echo '];;') > predef.ml"]]
;;

rule "lexer.ml" ["lexer.mll"]
  [command ["if test -f ../mllex;                                           \
             then " ^ !LEX ^ " lexer.mll;                                   \
             else touch lexer.ml;                                           \
             fi"]]
;;

rule "parser.ml" ["parser.mli"] [];;
rule "parser.mli" ["parser.mly"]
  [command [!YACC; "parser.mly"];
   command ["touch parser.ml"]]
;;

let newversion = ref false;;

let versiondeps = ["version.tpl"];;
let versioncommands =
  [command ["sed -e \"s/XXXX/`date`/\" version.tpl > version.ml"];
   command [!ZC; !ZFLAGS; "-c version.ml"]]
;;
rule2 "version.ml"
  (fun _ -> (versiondeps,
             fun () -> if !newversion or not newer "version.ml" versiondeps
                       then launch_comms ["version.ml"] versioncommands))
;;

rules_ml__default [!ZC; !ZFLAGS];;

rule2 "*.ml" (fun _ -> ([], fun () -> ()));;

dep "back.zi" ["instruct.zi"; "lambda.zi"];;
dep "back.zo" ["const.zi"; "instruct.zi"; "lambda.zi"; "misc.zi"; "prim.zi"];;
dep "buffcode.zi" [];;
dep "buffcode.zo" ["config.zi"; "meta.zi"; "misc.zi"; "opcodes.zi"];;
dep "builtins.zi" ["const.zi"; "globals.zi"];;
dep "builtins.zo" ["const.zi"; "globals.zi"; "hashtbl.zi"; "misc.zi"; "modules.zi"];;
dep "compiler.zi" ["syntax.zi"];;
dep "compiler.zo" ["back.zi"; "buffcode.zi"; "builtins.zi"; "const.zi"; "emitcode.zi"; "front.zi"; "globals.zi"; "hashtbl.zi"; "instruct.zi"; "lexer.zi"; "location.zi"; "misc.zi"; "modules.zi"; "parser.zi"; "reloc.zi"; "syntax.zi"; "ty_decl.zi"; "ty_error.zi"; "ty_intf.zi"; "types.zi"; "typing.zi"];;
dep "config.zo" ["config.zi"];;
dep "const.zi" [];;
dep "const.zo" ["misc.zi"];;
dep "emitcode.zi" ["instruct.zi"];;
dep "emitcode.zo" ["buffcode.zi"; "config.zi"; "const.zi"; "instruct.zi"; "labels.zi"; "misc.zi"; "opcodes.zi"; "prim.zi"; "prim_opc.zi"; "reloc.zi"];;
dep "emitdata.zi" [];;
dep "emitdata.zo" ["buffcode.zi"; "config.zi"; "const.zi"; "meta.zi"; "misc.zi"; "symtable.zi"];;
dep "errors.zi" ["const.zi"; "globals.zi"; "location.zi"];;
dep "errors.zo" ["const.zi"; "globals.zi"; "location.zi"; "misc.zi"];;
dep "expunge.zo" ["const.zi"; "hashtbl.zi"];;
dep "front.zi" ["lambda.zi"; "syntax.zi"; "location.zi"];;
dep "front.zo" ["builtins.zi"; "const.zi"; "globals.zi"; "lambda.zi"; "location.zi"; "match.zi"; "misc.zi"; "modules.zi"; "prim.zi"; "syntax.zi"];;
dep "globals.zi" ["const.zi"; "prim.zi"];;
dep "globals.zo" ["const.zi"; "prim.zi"];;
dep "hashtbl.zi" [];;
dep "hashtbl.zo" ["hashtbl.zi"];;
dep "instruct.zi" ["const.zi"; "prim.zi"];;
dep "instruct.zo" ["instruct.zi"];;
dep "labels.zi" [];;
dep "labels.zo" ["buffcode.zi"; "instruct.zi"; "misc.zi"];;
dep "lambda.zi" ["const.zi"; "prim.zi"];;
dep "lexer.zi" ["parser.zi"];;
dep "lexer.zo" ["config.zi"; "misc.zi"; "parser.zi"];;
dep "link.zi" [];;
dep "link.zo" ["config.zi"; "emitdata.zi"; "instruct.zi"; "misc.zi"; "modules.zi"; "opcodes.zi"; "reloc.zi"; "symtable.zi"];;
dep "location.zi" [];;
dep "location.zo" ["config.zi"; "misc.zi"];;
dep "main.zo" ["compiler.zi"; "config.zi"; "link.zi"; "meta.zi"; "misc.zi"; "modules.zi"; "readword.zi"; "symtable.zi"; "ty_decl.zi"; "version.zi"];;
dep "match.zi" ["lambda.zi"; "syntax.zi"];;
dep "match.zo" ["const.zi"; "globals.zi"; "lambda.zi"; "location.zi"; "misc.zi"; "prim.zi"; "syntax.zi"];;
dep "misc.zi" [];;
dep "misc.zo" ["config.zi"];;
dep "modules.zi" ["globals.zi"; "hashtbl.zi"];;
dep "modules.zo" ["const.zi"; "errors.zi"; "globals.zi"; "hashtbl.zi"; "misc.zi"];;
dep "opcodes.zo" [];;
dep "par_aux.zi" ["globals.zi"; "syntax.zi"];;
dep "par_aux.zo" ["builtins.zi"; "const.zi"; "errors.zi"; "globals.zi"; "location.zi"; "modules.zi"; "syntax.zi"];;
dep "parser.zi" ["syntax.zi"];;
dep "parser.zo" ["builtins.zi"; "const.zi"; "globals.zi"; "par_aux.zi"; "primdecl.zi"; "syntax.zi"; "types.zi"; "typing.zi"];;
dep "pr_type.zi" ["globals.zi"];;
dep "pr_type.zo" ["builtins.zi"; "const.zi"; "globals.zi"; "misc.zi"; "modules.zi"; "types.zi" ];;
dep "pr_value.zi" ["globals.zi"];;
dep "pr_value.zo" ["builtins.zi"; "const.zi"; "globals.zi"; "misc.zi"; "modules.zi"; "pr_type.zi"; "symtable.zi"; "types.zi"];;
dep "predef.zi" ["const.zi"];;
dep "predef.zo" ["const.zi"];;
dep "prim.zi" ["const.zi"];;
dep "prim_c.zo" ["prim_c.zi"];;
dep "prim_opc.zi" ["hashtbl.zi"; "misc.zi"; "opcodes.zi"; "prim.zi"; "prim_c.zi"];;
dep "prim_opc.zo" ["misc.zi"; "opcodes.zi"; "prim.zi"];;
dep "primdecl.zi" ["globals.zi"];;
dep "primdecl.zo" ["globals.zi"; "prim.zi"];;
dep "readword.zi" [];;
dep "readword.zo" ["readword.zi"];;
dep "reloc.zi" ["const.zi"];;
dep "reloc.zo" ["buffcode.zi"; "const.zi"];;
dep "symtable.zi" ["const.zi"; "reloc.zi"];;
dep "symtable.zo" ["buffcode.zi"; "config.zi"; "const.zi"; "errors.zi"; "hashtbl.zi"; "misc.zi"; "predef.zi"; "reloc.zi"];;
dep "syntax.zi" ["const.zi"; "globals.zi"; "location.zi"];;
dep "syntax.zo" ["syntax.zi"];;
dep "toplevel.zi" [];;
dep "toplevel.zo" ["back.zi"; "buffcode.zi"; "builtins.zi"; "compiler.zi"; "config.zi"; "const.zi"; "emitcode.zi"; "emitdata.zi"; "front.zi"; "globals.zi"; "instruct.zi"; "labels.zi"; "link.zi"; "location.zi"; "meta.zi"; "misc.zi"; "modules.zi"; "opcodes.zi"; "pr_value.zi"; "readword.zi"; "reloc.zi"; "symtable.zi"; "syntax.zi"; "ty_decl.zi"; "types.zi"; "typing.zi"; "version.zi"];;
dep "ty_decl.zi" ["globals.zi"; "location.zi"; "syntax.zi"];;
dep "ty_decl.zo" ["builtins.zi"; "const.zi"; "globals.zi"; "location.zi"; "misc.zi"; "modules.zi"; "pr_type.zi"; "syntax.zi"; "ty_error.zi"; "types.zi"; "typing.zi"];;
dep "ty_error.zi" ["globals.zi"; "location.zi"; "syntax.zi"];;
dep "ty_error.zo" ["const.zi"; "errors.zi"; "globals.zi"; "location.zi"; "misc.zi"; "pr_type.zi"; "syntax.zi"; "types.zi"];;
dep "ty_intf.zi" ["modules.zi"];;
dep "ty_intf.zo" ["globals.zi"; "hashtbl.zi"; "misc.zi"; "modules.zi"; "ty_error.zi"; "types.zi"];;
dep "types.zi" ["globals.zi"];;
dep "types.zo" ["builtins.zi"; "const.zi"; "globals.zi"; "misc.zi"; "modules.zi"];;
dep "typing.zi" ["globals.zi"; "location.zi"; "syntax.zi"];;
dep "typing.zo" ["builtins.zi"; "const.zi"; "errors.zi"; "globals.zi"; "misc.zi"; "modules.zi"; "syntax.zi"; "ty_error.zi"; "types.zi"];;


match get_args arg__warn []
with ([] | ["all"]) -> newversion := true; build rules "programs"
   | [targ] -> build rules targ
   | targs -> do_list (build rules) targs
;;
