#open "obj";;
#open "meta";;
#open "misc";;
#open "config";;
#open "opcodes";;

(* buffcode.ml : bufferisation du code en sortie *)

let out_buffer = ref (create_string 256)
and out_pos = ref 0
and abs_out_pos = ref 0
;;

let realloc_out_buffer () =
  let new_buffer = create_string (2 * string_length !out_buffer) in
    replace_string new_buffer !out_buffer 0;
    out_buffer := new_buffer;
    ()
;;

let init_out_code () =
  out_pos := 0;
  ()
;;

let get_out_position() =
  !out_pos
and get_abs_out_position() =
  !abs_out_pos + !out_pos
and set_out_position p =
  out_pos := p; ()
;;

let out b =
  if !out_pos >= string_length !out_buffer then realloc_out_buffer();
  set_nth_char !out_buffer (pred(incr out_pos)) (char_of_int b);
  ()
;;

let out_short s =
  if !big_endian then
    (out (lshift_right s 8); out s)
  else
    (out s; out (lshift_right s 8))
;;

let align02 () =
  if !access_alignment >= 2 & land !out_pos 1 == 1 then out NOP1
;;

let align12 () =
  if !access_alignment >= 2 & land !out_pos 1 == 0 then out NOP1
;;

let out_long l =
  if !big_endian then
    (out (lshift_right l 24);
     out (lshift_right l 16);
     out (lshift_right l 8);
     out l)
  else
    (out l;
     out (lshift_right l 8);
     out (lshift_right l 16);
     out (lshift_right l 24))
;;

let out_coded_long l =			(* en codage 2l + 1 *)
  if !big_endian then
    (out (lshift_right l 23);
     out (lshift_right l 15);
     out (lshift_right l 7);
     out (l+l+1))
  else
    (out (l+l+1);
     out (lshift_right l 7);
     out (lshift_right l 15);
     out (lshift_right l 23))
;;

let out_header (size, tag) =
  out_long (lshift_left size 10 + tag)
;;

let align04 () =
  match !access_alignment with
    1 -> ()
  | 4 -> begin match land !out_pos 3 with
           0 -> ()
         | 1 -> out NOP3; out NOP2; out NOP1
         | 2 -> out NOP2; out NOP1
         | _ -> out NOP1
         end
  | 2 -> if land !out_pos 1 == 1 then out NOP1
  | _ -> fatal_error "align04"
;;

let align34 () =
  match !access_alignment with
    1 -> ()
  | 4 -> begin match land !out_pos 3 with
           0 -> out NOP3; out NOP2; out NOP1
         | 1 -> out NOP2; out NOP1
         | 2 -> out NOP1
         | _ -> ()
         end
  | 2 -> if land !out_pos 1 == 0 then out NOP1
  | _ -> fatal_error "align34"
;;

let rec out_string s =
  if string_length s + !out_pos >= string_length !out_buffer then
    (realloc_out_buffer(); out_string s)
  else
    (replace_string !out_buffer s !out_pos;
     out_pos := !out_pos + string_length s;
     ())
;;

let out_float (f : float) = fatal_error "out_float"; ()
;;

let dump_output_to_file outchan =
  output outchan !out_buffer 0 !out_pos;
  abs_out_pos := !out_pos + !abs_out_pos;
  ()

and dump_output_to_heap () =
  let code = static_alloc !out_pos in
    blit_string !out_buffer 0 code 0 !out_pos;
    code
;;
