#open "misc";;
#open "const";;
#open "instruct";;
#open "prim";;
#open "opcodes";;
#open "prim_opc";;
#open "buffcode";;
#open "config";;
#open "labels";;
#open "reloc";;

(* emit.ml : generation du bytecode *)

let out_bool_test tst =
  function PTeq -> out tst
      |    PTnoteq -> out (tst + 1)
      |    PTlt -> out (tst + 2)
      |    PTgt -> out (tst + 3)
      |    PTle -> out (tst + 4)
      |    PTge -> out (tst + 5)
      |    _ -> fatal_error "out_bool_test"
;;

let out_prim_test imm_emit tst imm_tst =
  function PTnoteqimm x -> out imm_tst; imm_emit x
      |    x -> out_bool_test tst x
;;

let out_int_const i =
  let j = i + i + 1 in
  if i <= lshift_right maxint_byte 1
   & i >= lshift_right minint_byte 1 then
    (out CONSTBYTE; out j)
  else if i <= lshift_right maxint_short 1
        & i >= lshift_right minint_short 1 then
    (align12(); out CONSTSHORT; out_short j)
  else
    (align34(); out CONSTLONG; out_coded_long i)
;;

(***

let out_push_int_const i =
  let j = i + i + 1 in
  if i <= lshift_right maxint_byte 1
   & i >= lshift_right minint_byte 1 then
    (out PUSH_CONSTBYTE; out j)
  else if i <= lshift_right maxint_short 1
        & i >= lshift_right minint_short 1 then
    (align12(); out PUSH_CONSTSHORT; out_short j)
  else
    (align34(); out PUSH; out CONSTLONG; out_coded_long i)
;;

***)

let out_tag = function
    ConstrRegular(t, _) ->
      out t
  | ConstrExtensible name ->
      slot_for_tag name
;;

let out_header size tag =
  if !big_endian then begin
    out (lshift_right size 14);
    out (lshift_right size 6);
    out (lshift_left size 2);
    out_tag tag
  end else begin
    out_tag tag;
    out (lshift_left size 2);
    out (lshift_right size 6);
    out (lshift_right size 14)
  end
;;

let rec emit = function
      [] -> ()
    | Kquote(SCatom(ACint i)) :: C ->
        out_int_const i;
        emit C
    | Kquote(SCatom(ACchar c)) :: C ->
        out_int_const (int_of_char c);
        emit C
    | Kquote(SCblock(tag,[])) :: C ->
        begin match tag with
          ConstrRegular(t, _) ->
            if t < 10 then out (ATOM0 + t) else (out ATOM; out t)
        | ConstrExtensible name ->
            out ATOM; slot_for_tag name
        end;
        emit C
    | Kquote(sc) :: C ->
        align12();
        out GETGLOBAL;
        slot_for_literal sc;
        emit C
    | Kget_global qualid :: C ->
        align12();
        out GETGLOBAL;
        slot_for_get_global qualid;
        emit C
    | Kset_global qualid :: C ->
        align12();
        out SETGLOBAL;
        slot_for_set_global qualid;
        emit C
    | Kaccess n :: C ->
        if n < 6 then out(ACC0 + n) else (out ACCESS; out n);
        emit C
    | Kendlet n :: Kendlet p :: C ->
        emit(Kendlet(n+p) :: C)
    | Kendlet 1 :: C ->
        out ENDLET1; emit C
    | Kendlet n :: C ->
        out ENDLET; out n; emit C
    | Kdummy 2 :: Kclosure lbl :: Kupdate 0 :: C ->
        align12(); out LETREC1; out_label lbl; emit C
    | Kdummy 0 :: C ->
        out DUMMY; out 1; emit C
    | Kdummy n :: C ->
        out DUMMY; out n; emit C
    | Kupdate n :: C ->
        out UPDATE; out n; emit C
    | Kmakeblock(tag,n,shared) :: C ->
        if n <= 0 or n >= rangeint_byte then
          fatal_error "emit : Kmakeblock";
        begin match shared with
          Shared ->
            if n <= 2 then begin
              out (MAKESHAREDBLOCK1 + n - 1); out_tag tag
            end else begin
              align34(); out MAKESHAREDBLOCK; out_header n tag
            end
        | Unshared ->
            if n <= 4 then begin
              out (MAKEBLOCK1 + n - 1); out_tag tag
            end else begin
              align34(); out MAKEBLOCK; out_header n tag
            end
        end;
        emit C
    | Klabel lbl :: C ->
        define_label lbl; emit C
    | Kclosure lbl :: C ->
        align12(); out CUR; out_label lbl; emit C
    | Kpushtrap lbl :: C ->
        align12(); out PUSHTRAP; out_label lbl; emit C
    | Kbranch lbl :: C ->
        align12(); out BRANCH; out_label lbl; emit C
    | Kbranchif lbl :: C ->
        align12(); out BRANCHIF; out_label lbl; emit C
    | Kbranchifnot lbl :: C ->
        align12(); out BRANCHIFNOT; out_label lbl; emit C
    | Kstrictbranchif lbl :: C ->
        align12(); out BRANCHIF; out_label lbl; emit C
    | Kstrictbranchifnot lbl :: C ->
        align12(); out BRANCHIFNOT; out_label lbl; emit C
    | Kswitch lblvect :: C ->
        align12();
        out SWITCH;
        let orig = get_out_position() in
        do_vect (out_label_with_orig orig) lblvect;
        emit C
    | Ktest(tst,lbl) :: C ->
        (match tst with
            Peq_test ->
              align12(); out BRANCHIFEQ; out_label lbl
          | Pnoteq_test ->
              align12(); out BRANCHIFNEQ; out_label lbl
          | Pint_test(PTnoteqimm i) ->
              align34(); out BRANCHIFNEQIMM; out_coded_long i; out_label lbl
          | Pint_test x ->
              align12(); out_bool_test BRANCHIFEQ x; out_label lbl
          | Pfloat_test(PTnoteqimm i) ->
              align34(); out BRANCHIFNEQIMMFLOAT; out_float i; out_label lbl
          | Pfloat_test x ->
              out_bool_test EQFLOAT x; align12(); out BRANCHIF; out_label lbl
          | Pstring_test(PTnoteqimm i) ->
              align12(); out BRANCHIFNEQIMMSTRING;
              out_label lbl; out_string i; out 0
          | Pstring_test x ->
              out_bool_test EQSTRING x; align12(); out BRANCHIF; out_label lbl
          | Peqtag_test tag ->
              align02(); out BRANCHIFEQTAG;
              out_tag tag; out_label lbl
          | Pnoteqtag_test tag ->
              align02(); out BRANCHIFNEQTAG;
              out_tag tag; out_label lbl);
        emit C
    | Kprim Pidentity :: C ->
        emit C
    | Kprim p :: C ->
        (match p with
            Ptest tst ->
              (match tst with
                  Peq_test -> out EQ
                | Pnoteq_test -> out NEQ
                | Pint_test tst -> out_bool_test EQ tst
                | Pfloat_test tst -> out_bool_test EQFLOAT tst
                | Pstring_test tst -> out_bool_test EQSTRING tst
                | _ -> fatal_error "emit : Kprim, Ptest")
          | Pfield n ->
              if n < 4 then out (GETFIELD0 + n) else (out GETFIELD; out n)
          | Psetfield n ->
              if n < 4 then out (SETFIELD0 + n) else (out SETFIELD; out n)
          | Pccall(name, arity, does_allocate) ->
              let n = get_c_primitive name in
                if arity <= 5 then
                  out ((if does_allocate then C_CALL_ALLOC1 else C_CALL1) +
                       arity - 1)
                else
                  (out (if does_allocate then C_CALL_ALLOCN else C_CALLN);
                   out arity);
                out n
          | Pfloatprim p ->
              out FLOATOP;
              out(opcode_for_float_primitive p)
          | p ->
              out(opcode_for_primitive p));
        emit C
    | Kstartfun name :: C ->
(***
        if string_length name > 0 then
          (function_beginnings :=
            (name, !object_name, !abs_out_pos + !out_pos) ::
               !function_beginnings; ());
***)
        out STARTFUN; emit C
    | Kpush :: Kget_global qualid :: Kapply :: C ->
        align12();
        out PUSH_GETGLOBAL_APPLY;
        slot_for_get_global qualid;
        emit C
    | Kpush :: Kget_global qualid :: Ktermapply :: C ->
        align12();
        out PUSH_GETGLOBAL_APPTERM;
        slot_for_get_global qualid;
        emit C
(***
    | Kpush :: Kget_global qualid :: C ->
        align12();
        out PUSH_GETGLOBAL;
        slot_for_get_global qualid;
        emit C
    | Kpush :: Kaccess n :: C ->
        if n < 6 then out(PUSH_ACC0 + n) else (out PUSH_ACCESS; out n);
        emit C
    | Kpush :: Kquote(SCatom(ACint i)) :: C ->
        out_push_int_const i;
        emit C
    | Kpush :: Kquote(SCatom(ACchar c)) :: C ->
        out_push_int_const (int_of_char c);
        emit C
    | Kpush :: Kquote(SCblock(tag,[])) :: C ->
        begin match tag with
          ConstrRegular(t, _) ->
            if t < 2 then out (PUSH_ATOM0 + t) else (out PUSH; out ATOM; out t)
        | ConstrExtensible name ->
            out PUSH; out ATOM; slot_for_tag name
        end;
        emit C
    | Kpush :: Kquote(sc) :: C ->
        align12();
        out PUSH_GETGLOBAL;
        slot_for_literal sc;
        emit C
***)
    | instr :: C ->
        out(match instr with
           Kreturn -> RETURN
        |  Kgrab -> GRAB
        |  Kpush -> PUSH
        |  Kpushmark -> PUSHMARK
        |  Klet -> LET
        |  Kapply -> APPLY
        |  Ktermapply -> APPTERM
        |  Kpoptrap -> POPTRAP
        |  _  -> fatal_error "emit: should not happen");
        emit C
;;

let emit_phrase outchan phr =
  init_out_code();
  reset_label_table();
  begin match phr with
    { kph_fcts = [] } ->
        emit phr.kph_init;
        align04()
  | { kph_rec = false } ->
        emit [Kbranch 0];
        emit phr.kph_fcts;
        emit [Klabel 0];
        emit phr.kph_init;
        align04()
  | { kph_rec = true } ->
        emit phr.kph_init;
        emit [Kbranch 0];
        emit phr.kph_fcts;
        align04();
        emit [Klabel 0]
  end;
  dump_output_to_file outchan
;;
    
