#open "const";;
#open "prim";;

(* Dans le source, on fait reference aux valeurs globales par leur nom,
   ou bien par leur nom et le nom du module (notation mod__x). *)

type global_reference =
    GRname of string
  | GRmodname of qualified_ident
;;

(* Une fois tout bien desambigue, un global c'est le nom de son module de
   definition, son nom, et des informations. *)

type 'a global =
  { qualid: qualified_ident; (* Nom complet *)
    info: 'a }               (* Description *)
;;

value little_name_of_global : 'a global -> string;;

(* Les descriptions des constructeurs de types *)

type type_desc =
  { ty_stamp: int;                      (* Stamp *)
    ty_arity: int;                      (* Arite *)
    mutable ty_dang: bool;              (* Dangereux ou pas *)
    mutable ty_desc: type_components }  (* Description *)

and type_components =
    Abstract_type
  | Variant_type of constr_desc global list
                                        (* Somme -> liste des constructeurs *)
  | Record_type of label_desc global list
                                        (* Produit -> liste des etiquettes *)
  | Abbrev_type of typ list * typ
                                        (* Abreviation -> liste des parametres
                                           et corps de l'abreviation *)
(* Les expressions de types *)

and typ =
    Typ of typ_desc * mutable int
and typ_desc =
    Tvar of mutable typ_link
  | Tarrow of typ * typ
  | Tconstr of type_desc global * typ list
and typ_link =
    Tnolink
  | Tlinkto of typ

(* Les descriptions des constructeurs de types concrets *)

and mutable_flag =
  Mutable | Notmutable

and constr_desc =
  { cs_res: typ;                       (* Type du resultat *)
    cs_args: (typ * mutable_flag) list;(* Type des arguments + flag mutable *)
    cs_tag: constr_tag;                (* Son tag *)
    cs_pure: bool }                    (* Si sans aucun champ mutable *)

(* Les descriptions des etiquettes *)

and label_desc =
  { lbl_res: typ;                      (* Type du resultat *)
    lbl_arg: typ;                      (* Type de l'argument *)
    lbl_mut: mutable_flag;             (* Mutable ou pas *)
    lbl_pos: int }                     (* Position dans le tuple *)

;;

value notgeneric : int;;
value generic : int;;

type value_desc =
  { val_typ: typ;              (* Son type *)
    val_prim: prim_desc }      (* Est-ce une primitive ? *)

and prim_desc =
    ValueNotPrim
  | ValuePrim of int * primitive
;;

