#open "obj";;
#open "meta";;
#open "unix";;
#open "misc";;
#open "modules";;
#open "compiler";;
#open "link";;
#open "config";;

(* Parsing de la ligne de commande *)

let link_also = ref true
and object_files = ref ([] : string list)
and exec_file = ref default_exec_name
and default_linked_modules = ref ([] : string list)
and current_arg = ref 1
;;

let process_filename s =
  let pre3 = sub_string s 0 (string_length s - 3)
  and post3 = sub_string s (string_length s - 3) 3
  and pre4 = sub_string s 0 (string_length s - 4)
  and post4 = sub_string s (string_length s - 4) 4 in
    if post3 = ".ml" then begin
      compile_implementation pre3 s;
      if !link_also then
       begin object_files := pre3 :: !object_files; () end
    end else if post4 = ".mli" then begin
      compile_intf pre4 s
    end else if post3 = ".zo" then begin
      object_files := pre3 :: !object_files;
      ()
    end else begin
      prerr_begline "> Unrecognized argument ";
      prerr_endline2 command_line.(!current_arg)
    end 
;;

let arg_of_option () =
  let l = string_length command_line.(!current_arg) in
    if l > 2 then
      sub_string command_line.(!current_arg) 2 (l - 2)
    else if incr current_arg >= vect_length command_line then
     (prerr_begline ">> Missing parameter to option ";
      prerr_endline command_line.(pred !current_arg);
      raise Toplevel)
    else command_line.(!current_arg)
;;

try
  default_used_modules := assoc "cautious" default_used_interfaces;
  default_linked_modules := assoc "cautious" default_used_implementations;
  path_interfaces := [path_library; current_dir_name];
  while !current_arg < vect_length command_line do
    let s = command_line.(!current_arg) in
      begin
        if nth_char s 0 == `-` then
          match nth_char s 1 with
            `c` -> link_also := false; ()
          | `o` -> exec_file := arg_of_option(); ()
          | `I` -> path_interfaces := arg_of_option() :: !path_interfaces; ()
          | `O` -> begin let set = arg_of_option() in
                     try
                       default_used_modules :=
                         assoc set default_used_interfaces;
                       default_linked_modules :=
                        assoc set default_used_implementations; ()
                     with failure "find" ->
                       prerr_begline ">> Unknown module set: ";
                       prerr_endline2 set
                   end
          | `e` -> begin match arg_of_option() with
                     "b" -> config__big_endian := true; ()
                   | "l" -> config__big_endian := false; ()
                   |  e  -> prerr_begline ">> Bad endianness: ";
                            prerr_endline2 e
                   end
          | `a` -> begin match arg_of_option() with
                     "1" -> config__access_alignment := 1; ()
                   | "2" -> config__access_alignment := 2; ()
                   | "4" -> config__access_alignment := 4; ()
                   |  a  -> prerr_begline ">> Bad alignment: ";
                            prerr_endline2 a
                   end
          | `v` -> prerr_begline version__banner; prerr_endline2 ""
          | `g` -> write_symbols := true; ()
          | `s` -> ty_decl__verbose := false; ()    (* silent *)
          | `@` -> do_list process_filename
                     (readword__input_words(arg_of_option()))
          |  _  -> prerr_begline "> Unrecognized option ";
                   prerr_endline2 s
        else
          process_filename s
      end;
      flush std_out;
      incr current_arg
  done;
  begin match !object_files with
    [] -> ()
  | _  -> link (!default_linked_modules @ rev !object_files) !exec_file
  end;
  flush std_out;
  exit 0

with x ->
  let retcode =
    match x with
      Toplevel ->
        2
    | Out_of_memory ->
        prerr_begline ">> Not enough memory.";
        prerr_endline "";
        3
    | failure s ->
        prerr_begline ">> Evaluation failed: ";
        prerr_endline s;
        3
    | End_of_file ->
        prerr_begline ">> End of file encountered during input.";
	prerr_endline "";
        3
    | OS_error n ->
        prerr_begline ">> System call failed: ";
        prerr_endline (os_error_message n);
        3
    | Zinc s ->
        prerr_begline ">> CAML internal error: ";
        prerr_endline s;
        3
    | x ->
        prerr_begline ">> Uncaught compiler exception, tag ";
        prerr_int (obj_tag (Repr x));
        prerr_endline ""; 
        4
  in
    flush std_out;
    exit retcode
;;

