/* Definition de l'analyseur syntaxique */

%{
#open "par_aux";;
#open "const";;
#open "globals";;
#open "builtins";;
#open "syntax";;
#open "types";;
#open "typing";;
#open "primdecl";;
%}

/* Tokens */

%token <string> IDENT
%token <string> INFIX
%token <int> INT
%token <char> CHAR
%token <float> FLOAT
%token <string> STRING
%token EOF
%token SHARP          /* "#" */
%token BANG           /* "!" */
%token BANGEQUAL      /* "!=" */
%token AMPERSAND      /* "&" */
%token QUOTE          /* "'" */
%token LPAREN         /* "(" */
%token RPAREN         /* ")" */
%token STAR           /* "*" */
%token PLUS           /* "+" */
%token COMMA          /* "," */
%token MINUS          /* "-" */
%token MINUSGREATER   /* "->" */
%token DOT            /* "." */
%token DOTLPAREN      /* ".(" */
%token SLASH          /* "/" */
%token COLON          /* ":" */
%token COLONCOLON     /* "::" */
%token COLONEQUAL     /* ":=" */
%token SEMI           /* ";" */
%token SEMISEMI       /* ";;" */
%token LESS           /* "<" */
%token LESSMINUS      /* "<-" */
%token LESSEQUAL      /* "<=" */
%token LESSGREATER    /* "<>" */
%token EQUAL          /* "=" */
%token EQUALEQUAL     /* "==" */
%token GREATER        /* ">" */
%token GREATEREQUAL   /* ">=" */
%token AT             /* "@" */
%token LBRACKET       /* "[" */
%token LBRACKETBAR    /* "[|" */
%token RBRACKET       /* "]" */
%token CARET          /* "^" */
%token UNDERSCORE     /* "_" */
%token UNDERUNDER     /* "__" */
%token LBRACE         /* "{" */
%token BAR            /* "|" */
%token BARRBRACKET    /* "|]" */
%token RBRACE         /* "}" */
%token AND            /* "and" */
%token AS             /* "as" */
%token BEGIN          /* "begin" */
%token DO             /* "do" */
%token DONE           /* "done" */
%token DOWN           /* "down" */
%token ELSE           /* "else" */
%token END            /* "end" */
%token EXCEPTION      /* "exception" */
%token FOR            /* "for" */
%token FUN            /* "fun" */
%token FUNCTION       /* "function" */
%token IF             /* "if" */
%token IN             /* "in" */
%token LET            /* "let" */
%token MATCH          /* "match" */
%token MUTABLE        /* "mutable" */
%token NOT            /* "not" */
%token OF             /* "of" */
%token OR             /* "or" */
%token PREFIX         /* "prefix" */
%token REC            /* "rec" */
%token THEN           /* "then" */
%token TO             /* "to" */
%token TRY            /* "try" */
%token TYPE           /* "type" */
%token VALUE          /* "value" */
%token WHERE          /* "where" */
%token WHILE          /* "while" */
%token WITH           /* "with" */

/* Priorites */

%right prec_let
%right prec_where
%right prec_define
%right SEMI
%right prec_list
%right prec_if
%right COLONEQUAL LESSMINUS
%left  AS
%left  BAR
%right COMMA
%right OR
%left  AMPERSAND
%left  NOT
%left  LESS GREATER EQUAL GREATEREQUAL LESSEQUAL LESSGREATER
       GREATEREQUAL EQUALEQUAL BANGEQUAL
%right AT CARET
%right COLONCOLON
%left  PLUS MINUS
%right prec_typearrow
%left  STAR SLASH
%left  INFIX
%right prec_uminus
%right prec_app
%left  DOT
%right BANG

%start Implementation
%type <syntax__impl_phrase> Implementation 
%start Interface
%type <syntax__intf_phrase> Interface

%%

/* Une phrase d'une implementation */

Implementation :
        Expr SEMISEMI
          { make_impl(Zexpr $1) }
      | LET Binding_list SEMISEMI  %prec prec_let
          { make_impl(Zletdef(false, $2)) }
      | LET REC Binding_list SEMISEMI  %prec prec_let
          { make_impl(Zletdef(true, $3)) }
      | TYPE Type_decl SEMISEMI
          { make_impl(Ztypedef $2) }
      | EXCEPTION Exc_decl SEMISEMI
          { make_impl(Zexcdef $2) }
      | SHARP Directive SEMISEMI
          { make_impl(Zimpldirective $2) }
      | EOF
          { raise End_of_file }
;

/* Une phrase d'une interface */

Interface :
        VALUE Value_decl SEMISEMI
          { make_intf(Zvaluedecl $2) }
      | TYPE Type_decl SEMISEMI
          { make_intf(Ztypedecl $2) }
      | EXCEPTION Exc_decl  SEMISEMI
          { make_intf(Zexcdecl $2) }
      | SHARP Directive SEMISEMI
          { make_intf(Zintfdirective $2) }
      | EOF
          { raise End_of_file }
;

/* Auxiliaires pour les expressions. Places devant Expr pour des
   raisons de priorites. */

Simple_expr_list :
        Simple_expr Simple_expr_list
          { $1 :: $2 }
      | Simple_expr
          { [$1] }
;

Expr_comma_list :
        Expr COMMA Expr_comma_list
          { $1 :: $3 }
      | Expr  %prec COMMA
          { [$1] }
;

Expr_sm_list :
        Expr_sm_list SEMI Expr  %prec prec_list
          { $3 :: $1 }
      | Expr  %prec prec_list
          { [$1] }
;

Expr_label :
        Ext_ident EQUAL Expr
          { (find_label $1, $3)  }
;

Expr_label_list :
        Expr_label_list SEMI Expr_label  %prec prec_list
          { $3 :: $1 }
      | Expr_label  %prec prec_list
          { [$1] }
;

/* Expressions */

Expr :
        Simple_expr
          { $1 }
      | Simple_expr Simple_expr_list   %prec prec_app
          { make_apply ($1, $2) }
      | Expr COMMA Expr_comma_list
          { make_tuple_expr($1::$3) }
      | MINUS Expr  %prec prec_uminus
          { make_unary_minus $2 }
      | NOT Expr
          { make_unop "not" $2 }
      | Ide LESSMINUS Expr
          { make_expr (Zassign($1, $3)) }
      | Expr INFIX Expr
          { make_binop $2 $1 $3 }
      | Expr STAR Expr
          { make_binop "*" $1 $3 }
      | Expr SLASH Expr
          { make_binop "/" $1 $3 }
      | Expr PLUS Expr
          { make_binop "+" $1 $3 }
      | Expr MINUS Expr
          { make_binop "-" $1 $3 }
      | Expr COLONCOLON Expr
          { make_expr(Zconstruct(constr_cons, [$1; $3])) }
      | Expr AT Expr
          { make_binop "@" $1 $3 }
      | Expr CARET Expr
          { make_binop "^" $1 $3 }
      | Expr LESS Expr
          { make_binop "<" $1 $3 }
      | Expr GREATER Expr
          { make_binop ">" $1 $3 }
      | Expr EQUAL Expr
          { make_binop "=" $1 $3 }
      | Expr EQUALEQUAL Expr
          { make_binop "==" $1 $3 }
      | Expr GREATEREQUAL Expr
          { make_binop ">=" $1 $3 }
      | Expr LESSEQUAL Expr
          { make_binop "<=" $1 $3 }
      | Expr LESSGREATER Expr
          { make_binop "<>" $1 $3 }
      | Expr BANGEQUAL Expr
          { make_binop "!=" $1 $3 }
      | Expr COLONEQUAL Expr
          { make_binop ":=" $1 $3 }
      | Expr AMPERSAND Expr 
          { make_expr(Zsequand($1, $3)) }
      | Expr OR Expr
          { make_expr(Zsequor($1, $3)) }
      | Simple_expr DOT Ext_ident LESSMINUS Expr
          { make_expr(Zrecord_update($1, find_label $3, $5)) }
      | Simple_expr DOTLPAREN Expr RPAREN LESSMINUS Expr
          { make_ternop "vect_assign" $1 $3 $6 }
      | IF Expr THEN Expr ELSE Expr  %prec prec_if
          { make_expr(Zcondition($2, $4, $6)) }
      | IF Expr THEN Expr  %prec prec_if
          { make_expr(Zcondition($2, $4, 
                                 make_expr(Zconstruct(constr_void,[])))) }
      | WHILE Expr DO Expr DONE
          { make_expr(Zwhile($2, $4)) }
      | FOR Ide EQUAL Expr TO Expr DO Expr DONE
          { make_expr(Zfor($2, $4, $6, true, $8)) }
      | FOR Ide EQUAL Expr DOWN TO Expr DO Expr DONE
          { make_expr(Zfor($2, $4, $7, false, $9)) }
      | Expr SEMI Expr
          { make_expr(Zsequence($1, $3)) }
      | MATCH Expr WITH Function_match
          { make_expr(Zapply(make_expr(Zfunction $4), [$2])) }
      | LET Binding_list IN Expr  %prec prec_let
          { make_expr(Zlet(false, $2, $4)) }
      | LET REC Binding_list IN Expr  %prec prec_let
          { make_expr(Zlet(true, $3, $5)) }
      | FUN Fun_match
          { make_expr(Zfunction $2) }
      | FUNCTION Function_match
          { make_expr(Zfunction $2) }
      | TRY Expr WITH Try_match
	  { make_expr(Ztrywith($2, $4)) }
      | Expr WHERE Binding_list  %prec prec_where
          { make_expr(Zlet(false, $3, $1)) }
      | Expr WHERE REC Binding_list  %prec prec_where
          { make_expr(Zlet(true, $4, $1)) }
;

Simple_expr :
        Struct_constant
          { make_expr(Zconstant $1) }
      | Ext_ident
          { expr_constr_or_ident $1 }
      | LPAREN RPAREN
          { make_expr(Zconstruct(constr_void,[])) }
      | LBRACKET Expr_sm_list RBRACKET
          { make_list $2 }
      | LBRACKET RBRACKET
          { make_expr(Zconstruct (constr_nil, [])) }
      | LBRACKETBAR Expr_sm_list BARRBRACKET
          { make_expr(Zvector(rev $2)) }
      | LBRACKETBAR BARRBRACKET
          { make_expr(Zvector []) }
      | LPAREN Expr COLON Type RPAREN
          { make_expr(Zconstraint($2, $4)) }
      | LPAREN Expr RPAREN
          { $2 }
      | BEGIN Expr END
          { $2 }
      | LBRACE Expr_label_list RBRACE
          { make_expr (Zrecord $2) }
      | BANG Simple_expr
          { make_unop "!" $2 }
      | Simple_expr DOT Ext_ident
          { make_expr(Zrecord_access($1, find_label $3)) }
      | Simple_expr DOTLPAREN Expr RPAREN  %prec DOT
          { make_binop "vect_item" $1 $3 }
;

Struct_constant :
        Atomic_constant
          { SCatom $1 }
;

Atomic_constant :
        INT
          { ACint $1 }
      | FLOAT
          { ACfloat $1 }
      | STRING
          { ACstring $1 }
      | CHAR
          { ACchar $1 }
;

/* Les liaisons par filtrage */

Fun_match :
        Simple_pattern_list MINUSGREATER Expr BAR Fun_match
          { ($1, $3) :: $5 }
      | Simple_pattern_list MINUSGREATER Expr
	  { [$1, $3] }
;

Function_match :
        Pattern MINUSGREATER Expr BAR Function_match
          { ([$1], $3) :: $5 }
      | Pattern MINUSGREATER Expr
	  { [[$1], $3] }
;

Try_match :
        Pattern MINUSGREATER Expr BAR Try_match
          { ($1, $3) :: $5 }
      | Pattern MINUSGREATER Expr
          { [$1, $3] }
;

Binding_list :
        Binding AND Binding_list
          { $1 :: $3 }
      | Binding
          { [$1] }
;

Binding :
        Pattern EQUAL Expr  %prec prec_define
          { ($1, $3, ref 0) }
      | Ide Simple_pattern_list EQUAL Expr  %prec prec_define
          { (make_pat(Zvarpat $1), make_expr(Zfunction [$2, $4]), ref 0) }
;

/* Les motifs */

Pattern_sm_list :
        Pattern SEMI Pattern_sm_list
          { make_pat(Zconstructpat(constr_cons, [$1; $3])) }
      | Pattern
          { make_pat(Zconstructpat(constr_cons,
                       [$1; make_pat(Zconstructpat(constr_nil,[]))])) }
;

Pattern_label_list :
        Pattern_label SEMI Pattern_label_list
          { $1 :: $3 }
      | Pattern_label
          { [$1] }
      | UNDERSCORE
          { [] }
;

Pattern_label :
        Ext_ident EQUAL Pattern
          { (find_label $1, $3) }
;

Pattern_constr_list :
        LPAREN Pattern_comma_list RPAREN
          { $2 }
      | Simple_pattern
          { [$1] }
;

Pattern_comma_list :
        Pattern COMMA Pattern_comma_list
          { $1 :: $3 }
      | Pattern  %prec COMMA
          { [$1] }
;
  
Simple_pattern_list :
        Simple_pattern Simple_pattern_list
          { $1 :: $2 }
      | Simple_pattern
          { [$1] }
;

Pattern :
        Simple_pattern
          { $1 }
      | Pattern AS IDENT
          { make_pat(Zaliaspat($1, $3)) }
      | Pattern COLONCOLON Pattern
          { make_pat(Zconstructpat(constr_cons, [$1; $3])) }
      | Pattern COMMA Pattern_comma_list
          { make_tuple_pat($1 :: $3) }
      | Ext_ident Pattern_constr_list
          { make_pat(Zconstructpat (find_constructor $1, $2)) }
      | Pattern BAR Pattern
          { make_pat(Zorpat($1, $3)) }
;

Simple_pattern :
        Atomic_constant
          { make_pat(Zconstantpat $1) }
      | MINUS INT
          { make_pat(Zconstantpat(ACint(minus_int $2))) }
      | MINUS FLOAT
          { make_pat(Zconstantpat(ACfloat(minus_float $2))) }
      | UNDERSCORE
          { make_pat(Zwildpat) }
      | Ide
          { pat_constr_or_var $1 }
      | Qual_ident
          { make_pat(Zconstructpat(find_constructor (GRmodname $1),[])) }
      | LPAREN RPAREN
          { make_pat(Zconstructpat(constr_void,[])) }
      | LBRACKET RBRACKET
          { make_pat(Zconstructpat(constr_nil,[])) }
      | LBRACKET Pattern_sm_list RBRACKET
          { $2 }
      | LPAREN Pattern COLON Type RPAREN
          { make_pat(Zconstraintpat($2, $4)) }
      | LBRACE Pattern_label_list RBRACE
          { make_pat(Zrecordpat $2) }
      | LPAREN Pattern RPAREN
          { $2 }
;

/* Noms de variables */

Ide :
        IDENT
          { $1 }
      | PREFIX Infx
          { $2 }
;

Infx :
        INFIX       { $1 }
      | PLUS        { "+" }   | MINUS        { "-" }
      | STAR        { "*" }   | SLASH        { "/" }
      | LESSEQUAL   { "<=" }  | GREATEREQUAL { ">=" }
      | LESSGREATER { "<>" }  | LESS         { "<" }
      | GREATER     { ">" }   | COLONCOLON   { "::" }
      | AT          { "@" }   | AMPERSAND    { "&" }
      | OR          { "or" }  | CARET        { "^" }
      | BANGEQUAL   { "!=" }  | COLONEQUAL   { ":=" }
      | EQUALEQUAL  { "==" }  | EQUAL        { "=" }
      | NOT         { "not" } | BANG         { "!" }
;

Qual_ident :
        IDENT UNDERUNDER Ide
          { {qual=$1; id=$3} }
;

Ext_ident :
        Qual_ident
          { GRmodname $1 }
      | Ide
          { GRname $1 }
;

/* Les expressions de type */

Type :
        Simple_type
          { $1 }
      | Type STAR Type_star_list
          { make_typ(Ztypeconstr(gr_type_tuple, $1 :: $3)) }
      | Type MINUSGREATER Type  %prec prec_typearrow
          { make_typ(Ztypearrow($1, $3)) }
;

Simple_type :
        Type_var
          { make_typ(Ztypevar $1) }
      | Ext_ident
          { make_typ(Ztypeconstr($1, [])) }
      | Simple_type Ext_ident
          { make_typ(Ztypeconstr($2, [$1])) }
      | LPAREN Type COMMA Type_comma_list RPAREN Ext_ident
          { make_typ(Ztypeconstr($6, $2 :: $4)) }
      | LPAREN Type RPAREN
          { $2 }
;

Type_star_list :
        Simple_type
          { [$1] }
      | Simple_type STAR Type_star_list
          { $1 :: $3 }
;

Type_var :
        QUOTE IDENT
          { $2 }
;

Type_comma_list :
        Type COMMA Type_comma_list
          { $1 :: $3 }
      | Type
          { [$1] }
;

/* Les declarations */

Value_decl :
        Value1_decl AND Value_decl
          { $1 :: $3 }
      | Value1_decl
          { [$1] }
;

Type_decl :
        Type1_decl AND Type_decl
          { $1 :: $3 }
      | Type1_decl
          { [$1] }
;

Exc_decl :
        Constr1_decl AND Exc_decl
          { $1 :: $3 }
      | Constr1_decl
          { [$1] }
;

Constr_decl :
        Constr1_decl BAR Constr_decl
          { $1 :: $3 }
      | Constr1_decl
          { [$1] }
;

Label_decl :
        Label1_decl SEMI Label_decl
          { $1 :: $3 }
      | Label1_decl
          { [$1] }
;

Value1_decl :
        Ide COLON Type
          { ($1, $3, ValueNotPrim) }
      | Ide COLON Type EQUAL Prim_decl
          { ($1, $3, $5) }
;

Prim_decl :
        INT STRING
          { find_primitive $1 $2 "" }
      | INT STRING STRING
          { find_primitive $1 $2 $3 }
;

Type1_decl :
        Type_params IDENT Type1_def
          { ($2, $1, $3) }
;

Type1_def :
        /* epsilon */
          { Zabstract_type }
      | EQUAL Constr_decl
          { Zvariant_type $2 }
      | EQUAL LBRACE Label_decl RBRACE
          { Zrecord_type $3 }
      | EQUALEQUAL Type
          { Zabbrev_type $2 }
;

Constr1_decl :
        IDENT OF Constr_args
          { ($1, $3) }
      | IDENT
          { ($1, []) }
;

Label1_decl :
        Mutable_option IDENT COLON Type
          { ($2, $4, $1) }
;

Constr_args :
        Constr_arg
          { [$1] }
      | Constr_arg STAR Constr_args
          { $1 :: $3 }
;

Constr_arg :
        Mutable_option Simple_type
          { ($2, $1) }
;

Mutable_option :
        MUTABLE
          { Mutable }
      | /* epsilon */
          { Notmutable }
;

Type_params :
        LPAREN Type_var_list RPAREN
          { $2 }
      | Type_var
          { [$1] }
      |
          { [] }
;

Type_var_list :
        Type_var COMMA Type_var_list
          { $1 :: $3 }
      | Type_var
          { [$1] }
;
       
/* Les directives */

Directive :
        IDENT STRING
          { Zdir($1, $2) }
;

%%
