#open "misc";;
#open "const";;
#open "globals";;
#open "builtins";;
#open "types";;
#open "modules";;

type chan = Std | Err;;
let cur_chan = ref Std;;
let print_on c = (cur_chan := c);;
let print_string s =
    match !cur_chan
	with Std -> io__print_string s
	   | Err -> io__prerr_string s
;;

let print_global sel_fct gl =
  let rec can_omit_qualifier = function
      [] -> false
    | md1::mdl ->
        try
          find_in_symtable (sel_fct md1) gl.qualid.id;
          gl.qualid.qual = md1.mod_name
        with failure "find" ->
          can_omit_qualifier mdl
  in
    if not (can_omit_qualifier (!defined_module :: !used_modules)) then
      (print_string gl.qualid.qual; print_string "__");
    print_string gl.qualid.id
;;

let print_type_constructor = print_global types_of_module;;

(* Affichage des expressions de type *)

let rec int_to_alpha i =
  if i < 26
  then make_string 1 (char_of_int (i+97))
  else (int_to_alpha (i/26) ^ make_string 1 (char_of_int ((i mod 26)+97)))
;;

let name_of_type_vars = ref ([] : (typ * string) list)
and type_vars_counter = ref 0
;;

let reset_type_var_name () =
  name_of_type_vars := []; type_vars_counter := (-1); ()
;;

let name_of_type_var var =
  let rec find_var = function
    [] ->
      let var_name = int_to_alpha(incr type_vars_counter) in
        name_of_type_vars := (var, var_name) :: !name_of_type_vars; var_name
  | (v, name_v) :: rest ->
      if v == var then name_v else find_var rest in
  find_var !name_of_type_vars
;;

let rec print_typ priority ty =
  match type_repr ty with
    Typ(Tvar _, _) as var ->
      print_string "'"; print_string (name_of_type_var var)
  | Typ(Tarrow(ty1, ty2), _) ->
      if priority >= 1 then print_string "(";
      print_typ 1 ty1;
      print_string " -> ";
      print_typ 0 ty2;
      if priority >= 1 then print_string ")"
  | Typ(Tconstr(cstr, args), _) ->
      if same_type_constr cstr constr_type_tuple then begin
        if priority >= 2 then print_string "(";
        print_typ_list 2 " * " args;
        if priority >= 2 then print_string ")"
      end else begin
        begin match args with
           []    -> ()
         | [ty1] ->
             print_typ 2 ty1; print_string " "
         | tyl ->
             print_string "("; print_typ_list 0 ", " tyl; print_string ") "
         end;
         print_type_constructor cstr
       end

and print_typ_list priority sep = function
    [] ->
      fatal_error "print_typ_list"
  | [ty] ->
      print_typ priority ty
  | ty::rest ->
      print_typ priority ty;
      print_string sep;
      print_typ_list priority sep rest
;;

let print_constr c = print_on Std; print_global constrs_of_module c
and print_type_constr tc = print_on Std; print_global types_of_module tc
and print_value v = print_on Std; print_global values_of_module v
and print_label l = print_on Std; print_global labels_of_module l
;;

let prerr_constr c = print_on Err; print_global constrs_of_module c
and prerr_type_constr tc = print_on Err; print_global types_of_module tc
and prerr_value v = print_on Err; print_global values_of_module v
and prerr_label l = print_on Err; print_global labels_of_module l
;;

let print_type ty =
  print_on Std; print_typ 0 ty
;;

let print_one_type ty =
  print_on Std; reset_type_var_name(); print_typ 0 ty
;;

let prerr_type ty =
  print_on Err; print_typ 0 ty
;;

let prerr_one_type ty =
  print_on Err; reset_type_var_name(); print_typ 0 ty
;;
