#open "misc";;
#open "const";;
#open "globals";;
#open "hashtbl";;
#open "modules";;

(* builtins.ml : les globaux predefinis *)

let builtin n d = {qualid={qual="builtin"; id=n}; info=d}
;;

(* Quelques types speciaux *)

let constr_type_tuple = builtin "*"
  {ty_stamp=1; ty_arity=-1; ty_dang=false; ty_desc=Abstract_type}
and constr_type_unit = builtin "unit"
   {ty_stamp=2; ty_arity=0; ty_dang=false; ty_desc=Abstract_type}
and constr_type_exn = builtin "exn"
   {ty_stamp=3; ty_arity=0; ty_dang=false; ty_desc=Variant_type []}
and constr_type_bool = builtin "bool"
   {ty_stamp=4; ty_arity=0; ty_dang=false; ty_desc=Abstract_type}
and constr_type_int = builtin "int"
   {ty_stamp=5; ty_arity=0; ty_dang=false; ty_desc=Abstract_type}
and constr_type_float = builtin "float"
   {ty_stamp=6; ty_arity=0; ty_dang=false; ty_desc=Abstract_type}
and constr_type_string = builtin "string"
   {ty_stamp=7; ty_arity=0; ty_dang=false; ty_desc=Abstract_type}
and constr_type_char = builtin "char"
   {ty_stamp=8; ty_arity=0; ty_dang=false; ty_desc=Abstract_type}
and constr_type_list = builtin "list"
   {ty_stamp=9; ty_arity=1; ty_dang=false; ty_desc=Abstract_type}
and constr_type_vect = builtin "vect"
   {ty_stamp=10; ty_arity=1; ty_dang=true; ty_desc=Abstract_type}
;;

let type_arrow (t1,t2) = Typ(Tarrow(t1, t2), notgeneric)
and type_unit = Typ(Tconstr(constr_type_unit, []), notgeneric)
and type_exn = Typ(Tconstr(constr_type_exn, []), notgeneric)
and type_bool = Typ(Tconstr(constr_type_bool, []), notgeneric)
and type_int = Typ(Tconstr(constr_type_int, []), notgeneric)
and type_float = Typ(Tconstr(constr_type_float, []), notgeneric)
and type_string = Typ(Tconstr(constr_type_string, []), notgeneric)
and type_char = Typ(Tconstr(constr_type_char, []), notgeneric)
and type_vect t = Typ(Tconstr(constr_type_vect, [t]), notgeneric)
;;

(* Quelques constructeurs speciaux *)

(* Quelques constructeurs utiles (pour le parsing) *)

let constr_void =
  builtin "()"
    { cs_res = Typ(Tconstr(constr_type_unit,[]), notgeneric);
      cs_args= [];
      cs_tag = ConstrRegular(0,1);
      cs_pure= true }
;;
constr_type_unit.info.ty_desc <- Variant_type [constr_void]
;;
let constr_nil =
  let arg = Typ(Tvar(Tnolink), generic) in
  builtin "[]"
    { cs_res = Typ(Tconstr(constr_type_list, [arg]), generic);
      cs_args= [];
      cs_tag = ConstrRegular(0,2);
      cs_pure= true }

and constr_cons =
  let arg1 = Typ(Tvar(Tnolink), generic) in
  let arg2 = Typ(Tconstr(constr_type_list, [arg1]), generic) in
  builtin "::"
    { cs_res = arg2;
      cs_args= [arg1, Notmutable; arg2, Notmutable];
      cs_tag = ConstrRegular(1,2);
      cs_pure= true }
;;
constr_type_list.info.ty_desc <- Variant_type [constr_nil; constr_cons]
;;
let constr_false =
  builtin "false"
    { cs_res = Typ(Tconstr(constr_type_bool,[]), notgeneric);
      cs_args= [];
      cs_tag = ConstrRegular(0,2);
      cs_pure= true }

and constr_true =
  builtin "true"
    { cs_res = Typ(Tconstr(constr_type_bool,[]), notgeneric);
      cs_args= [];
      cs_tag = ConstrRegular(1,2);
      cs_pure= true }
;;
constr_type_bool.info.ty_desc <- Variant_type [constr_false; constr_true]
;;

let match_failure_tag =
  ConstrExtensible {qual="builtin"; id="match_failure"}
;;

let constr_match_failure =
  builtin "match_failure"
    { cs_res = Typ(Tconstr(constr_type_exn,[]), notgeneric);
      cs_args= [ type_string, Notmutable;
                 type_int, Notmutable;
                 type_int, Notmutable ];
      cs_tag = match_failure_tag;
      cs_pure= true }
;;

(* L'abomination des tuples: construction a la demande des descripteurs *)

(* Pour l'instant: limitation a, disons, 16 constructeurs *)

let tuple_name = {qual="builtin"; id=","}
;;

let create_constr_tuple n =
  let rec make_vars i =
    if i <= 0 then [] else
      Typ(Tvar(Tnolink), generic) :: make_vars (pred i) in
  let vars = make_vars n in
    { qualid = tuple_name;
      info   = { cs_res = Typ(Tconstr(constr_type_tuple, vars), generic);
                 cs_args= map (fun v -> (v, Notmutable)) vars;
                 cs_tag = ConstrRegular(0,1);
                 cs_pure= true } }
;;

let constr_tuple =
  let v = make_vect 16 (create_constr_tuple 0) in
    for i = 1 to 15 do v.(i) <- create_constr_tuple i done;
    (fun i -> if i < 2 or i > 15 then fatal_error "constr_tuple" else v.(i))
;;

let is_constr_tuple (g : constr_desc global) =
  g.qualid = tuple_name
;;

(* Le module builtin *)

let module_builtin = new_module "builtin";;

do_list (add_to_symtable module_builtin.mod_types)
  [constr_type_tuple; constr_type_unit; constr_type_exn;
   constr_type_bool; constr_type_int; constr_type_float; constr_type_string;
   constr_type_char; constr_type_list; constr_type_vect ]
;;
do_list (add_to_symtable module_builtin.mod_constrs)
  [constr_void; constr_nil; constr_cons; constr_true; constr_false;
   constr_match_failure ]
;;

let reset_module_table () =
  reset_hashtable module_table;
  add_to_assoctable module_table ("builtin", module_builtin)
;;

