#open "obj";;
#open "misc";;
#open "const";;
#open "config";;
#open "buffcode";;
#open "symtable";;

(* emit_data.ml : sortie du segment de donnees *)

let num_of_tag = function
    ConstrRegular(t, _) -> t
  | ConstrExtensible qualid -> get_num_of_tag qualid
;;

let emit_data_to_file outstream =
  let pos =
    ref 0 in
  let out =
    output_byte outstream in
  let out_long =
    if !big_endian then
      (fun l -> out (lshift_right l 24);
                out (lshift_right l 16);
                out (lshift_right l 8);
                out l;
                pos := !pos + 4)
    else
      (fun l -> out l;
                out (lshift_right l 8);
                out (lshift_right l 16);
                out (lshift_right l 24);
                pos := !pos + 4) in
  let out_header =
    (* La couleur de l'en-tete est White (0 << 8)
       les marques sont a 0 *)
    if !big_endian then
      (fun size tag ->
        out (lshift_right size 14);
        out (lshift_right size 6);
        out (lshift_left size 2);
        out tag;
        pos := !pos + 4)
    else
      (fun size tag ->
        out tag;
        out (lshift_left size 2);
        out (lshift_right size 6);
        out (lshift_right size 14);
        pos := !pos + 4) in
  let emit_atomic_constant = function
      ACint i ->
        2 * i + 1
    | ACchar c ->
        2 * int_of_char c + 1
    | ACstring s ->
        let size = lshift_right (string_length s + 5) 2 in
          out_header size 253;
          let p = !pos in
            output_string outstream s;
            (match land (string_length s) 3 with
                0 -> out 0; out 0; out 0; out 4
              | 1 -> out 0; out 0; out 3
              | 2 -> out 0; out 2
              | _ -> out 0; out 0; out 0; out 0; out 5);
            pos := !pos + size * 4;
            p
    | ACfloat f ->
        out_header 2 254;               (* ce code suppose sizeof(double)=8 *)
        let p = !pos in
          output outstream (magic f) 0 8; (* ne marche pas si on *)
          pos := !pos + 8;              (* compile pour une endianness *)
          p in                          (* differente *)
  let rec emit_struct_constant = function
      SCatom a ->
        emit_atomic_constant a
    | SCblock(constr, []) ->
        4 * (num_of_tag constr) + 2
    | SCblock(constr, args) ->
        let transl_args = map emit_struct_constant args in
        out_header (list_length args) (num_of_tag constr);
        let p = !pos in
          do_list out_long transl_args;
          p in
  let globals = make_vect (number_of_globals()) (SCatom(ACint 0)) in
    do_list (function (n, sc) -> globals.(n) <- sc) !literal_table;
    literal_table := [];
    let transl_globals =
      map_vect_list emit_struct_constant globals in
    out_header (vect_length globals) 0;
    let p = !pos in
      do_list out_long transl_globals;
      out_long p;
      ()
;;

let obj_for_atomic_const = function
    ACint i -> Repr i
  | ACchar i -> Repr i
  | ACfloat i -> Repr i
  | ACstring i -> Repr i
;;

let rec obj_for_struct_const = function
    SCatom a -> obj_for_atomic_const a
  | SCblock(constr, args) ->
      let res = obj_block (num_of_tag constr) (list_length args) in
      let rec set_field i = function
          [] -> ()
        | sc::rest ->
            set_obj_field res i (obj_for_struct_const sc);
            set_field (succ i) rest;
            ()
      in
        set_field 0 args;
        res
;;

let emit_data_to_heap () =
  let enter_constant (n, sc) =
    if n >= vect_length meta__global_data then meta__realloc_global_data n;
    meta__global_data.(n) <- obj_for_struct_const sc
  in
    do_list enter_constant !literal_table;
    literal_table := [];
    ()
;;
