#open "unix";;
#open "config";;

exception Zinc of string;;

let fatal_error s = raise (Zinc s);;

exception Toplevel;;

let toplevel = ref false;;

let print_begline s =
  (if !toplevel then print_string toplevel_output_prompt
                else print_string batch_output_prompt);
  print_string s
;;

let prerr_begline s =
  (if !toplevel then prerr_string toplevel_error_prompt
                else prerr_string batch_error_prompt);
  prerr_string s
;;

let prerr_endline2 s =
  prerr_endline s; prerr_endline ""
;;

let file_exists filename =
  let fd = open filename 0 0 in
    if fd == -1 then false else (close fd; true)
;;

(**********
let string_contains s c
  = let result = ref false
    in for i = 0 to (string_length s)
       do if nth_char s i == c then (result := true; ())
       done;
       !result
;;
**********)

let current_dir_name
  = match file_system_type
    with "unix" -> "."
       | "macintosh" -> ":"
       | "ms-dos" -> "."
       | _ -> fatal_error "current_dir_name"
;;

exception Found;;

let is_absolute =
  match file_system_type with
    "unix" ->
      (function n  ->
        (string_length n >= 1 & sub_string n 0 1 = "/")
        or (string_length n >= 2 & sub_string n 0 2 = "./")
        or (string_length n >= 3 & sub_string n 0 3 = "../"))
  | "macintosh" ->
      (function n ->
        try
          for i = 0 to string_length n - 1 do
            if nth_char n i == `:` then raise Found
          done;
          false
        with Found ->
          true)
  | "ms-dos" ->
      (function n ->
        (string_length n >= 2 & nth_char n 1 == `:`)
        or (string_length n >= 1 & sub_string n 0 1 = "\\")
        or (string_length n >= 2 & sub_string n 0 2 = ".\\")
        or (string_length n >= 3 & sub_string n 0 3 = "..\\"))
  | _ -> fatal_error "is_absolute"
;;

let append_name
  = let (sep_str, sep_ch) = match file_system_type
                            with "unix" -> ("/", `/`)
                               | "macintosh" -> (":", `:`)
                               | "ms-dos" -> ("\\", `\\`)
                               | _ -> fatal_error "append_name"
    in fun dirname filename
        -> let l = string_length dirname - 1
           in let d = (if l == -1 or nth_char dirname l == sep_ch
		       then dirname
		       else dirname ^ sep_str)
              in d ^ filename
;;

let find_in_path filename path
  = if is_absolute filename then filename
    else let rec find
           = function [] -> prerr_begline ">> Cannot find file ";
                            prerr_endline filename;
                            raise Toplevel
                    | a::rest -> let b = append_name a filename
                                 in if file_exists b then b else find rest
         in find path
;;
