#open "misc";;
#open "prim";;
#open "opcodes";;
#open "hashtbl";;

(* Les opcodes des primitives.
   Tout changement dans le type "primitive" doit etre repercute ici *)

let opcode_for_primitive = function
    Praise -> RAISE
  | Pnot -> BOOLNOT
  | Ptag_of -> TAGOF
  | Pnegint -> NEGINT
  | Psuccint -> SUCCINT
  | Ppredint -> PREDINT
  | Paddint -> ADDINT
  | Psubint -> SUBINT
  | Pmulint -> MULINT
  | Pdivint -> DIVINT
  | Pmodint -> MODINT
  | Pandint -> ANDINT
  | Porint -> ORINT
  | Pxorint -> XORINT
  | Pshiftleftint -> SHIFTLEFTINT
  | Pshiftrightint -> SHIFTRIGHTINT
  | Pincr -> INCR
  | Pdecr -> DECR
  | Pintoffloat -> INTOFFLOAT
  | Pmakestring -> MAKESTRING
  | Pstringlength -> STRINGLENGTH
  | Pgetstringchar -> GETSTRINGCHAR
  | Psetstringchar -> SETSTRINGCHAR
  | Pmakevector -> MAKEVECTOR
  | Pvectlength -> VECTLENGTH
  | Pgetvectitem -> GETVECTITEM
  | Psetvectitem -> SETVECTITEM
  | _ -> fatal_error "opcode_for_primitive"
;;

let opcode_for_float_primitive = function
    Pfloatofint -> FLOATOFINT
  | Pnegfloat -> NEGFLOAT
  | Paddfloat -> ADDFLOAT
  | Psubfloat -> SUBFLOAT
  | Pmulfloat -> MULFLOAT
  | Pdivfloat -> DIVFLOAT
;;

(* Les codes des primitives C *)

let c_primitives_table = (new_hashtable 29 : (string * int) hashtable)
;;

let count = ref 0 in
  do_list (fun name ->
    add_to_assoctable c_primitives_table (name, !count);
    incr count)
  prim_c__c_primitives_list
;;

let get_c_primitive name =
  try
    find_in_assoctable c_primitives_table name
  with failure "find" ->
    prerr_begline ">> Unknown C primitive ";
    prerr_endline name;
    raise Toplevel
;;

