#open "unix";;
#open "obj";;
#open "misc";;
#open "const";;
#open "lexer";;
#open "parser";;
#open "location";;
#open "syntax";;
#open "builtins";;
#open "hashtbl";;
#open "globals";;
#open "modules";;
#open "types";;
#open "ty_error";;
#open "typing";;
#open "ty_decl";;
#open "ty_intf";;
#open "front";;
#open "instruct";;
#open "back";;
#open "buffcode";;
#open "emitcode";;

(* Le processus qui produit les tokens *)

let lexing_process lexbuf () =
  let rec build_list n =
    if n <= 0 then [] else
      let tok =
        Main lexbuf in
      let triple =
        (tok, lexing__get_lexeme_start lexbuf, lexing__get_lexeme_end lexbuf)
      in
        triple :: build_list (pred n)
  in
    build_list 20
;;

(* Le wrapper qui casse les listes de tokens en tokens *)

let get_lexeme_buffer = ref ([] : (token * int * int) list)
;;

let rec get_lexeme lexing_prod =
  match !get_lexeme_buffer with
    tok::rest ->
      get_lexeme_buffer := rest; tok
  | [] ->
      get_lexeme_buffer := producer__get lexing_prod;
      get_lexeme lexing_prod
;;

(* Le processus qui produit les phrases *)

let parsing_impl_process lexing_prod typing_is_over =
  let wait_typing = ref false in
    function () ->
      if !wait_typing then
        semaphore__P(typing_is_over);
      let phr =
        Implementation get_lexeme lexing_prod in
      parsing__clear_parser();
      begin match phr with
        Impl((Zexpr _ | Zletdef(_,_)), _) ->
          wait_typing := false
      | _ ->
          wait_typing := true
      end;
      phr
;;

let parsing_intf_process lexing_prod typing_is_over =
  let wait_typing = ref false in
    function () ->
      if !wait_typing then
        semaphore__P(typing_is_over);
      let phr =
        Interface get_lexeme lexing_prod in
      parsing__clear_parser();
      begin match phr with
        Intf(Zvaluedecl _, _) ->
          wait_typing := false
      | _ ->
          wait_typing := true
      end;
      phr
;;

(* Le truc qui parse et qui recupere les erreurs de parsing *)

let parse_phrase lexing_prod parsing_prod =
  let rec skip () =
    try
      match get_lexeme lexing_prod with
        (EOF | SEMISEMI), _, _ -> ()
      | _ -> skip()
    with lexer__Lexical_error(_,_,_) ->
      skip() in
  try
    producer__get parsing_prod
  with
    parsing__Parse_error f ->
      let (tok, pos1, pos2) = get_lexeme lexing_prod in
      if f (obj__Repr EOF) or f (obj__Repr SEMISEMI) then () else skip();
      prerr_location (Loc(pos1, pos2));
      prerr_begline ">> Syntax error.";
      prerr_endline "";
      raise Toplevel
  | lexer__Lexical_error(msg, pos1, pos2) ->
      if pos1 >= 0 & pos2 >= 0 then prerr_location (Loc(pos1, pos2));
      prerr_begline ">> Lexical error: ";
      prerr_string msg;
      prerr_endline ".";
      skip();
      raise Toplevel
  | Toplevel ->
      skip();
      raise Toplevel
;;

(* Execution des directives *)

let do_directive = function
    Zdir("open", name) ->
      used_modules := find_module name :: !used_modules; ()
  | Zdir("close", name) ->
      let rec remove_module = function
        [] ->
          prerr_begline "> Module \""; prerr_string name;
          prerr_endline "\" was not open.";
          []
      | md :: rest ->
          if md.mod_name = name then rest else md :: remove_module rest in
      used_modules := remove_module !used_modules; ()
  | Zdir(d, name) ->
      prerr_begline "> Unknown directive \"";
      prerr_string d;
      prerr_endline2 "\", ignored."
;;

(****

let typing_impl_process typing_is_over lexing_prod parsing_prod () =
  let (Impl(desc, loc) as phr) =
    parse_phrase lexing_prod parsing_prod in
  reset_type_expression_vars [];
  begin match desc with
    Zexpr expr ->
      type_expression loc expr; ()
  | Zletdef(rec_flag, pat_expr_list) ->
      type_letdef loc rec_flag pat_expr_list; ()
  | Ztypedef decl ->
      type_typedecl loc decl;
      semaphore__V typing_is_over
  | Zexcdef decl ->
      type_excdecl loc decl;
      semaphore__V typing_is_over
  | Zimpldirective dir ->
      do_directive dir;
      semaphore__V typing_is_over
  end;
  flush std_out;
  phr
;;

(* Compilation d'une phrase *)

let compile_impl_phrase outstream (Impl(desc,loc)) =
  match desc with
    Zexpr expr ->
      emit_phrase outstream
        (compile_lambda false (translate_expression expr))
  | Zletdef(rec_flag, pat_expr_list) ->
      emit_phrase outstream
       (if rec_flag
        then compile_lambda true  (translate_letdef_rec loc pat_expr_list)
        else compile_lambda false (translate_letdef loc pat_expr_list))
  | _ ->
      ()
;;

****)

let compile_impl_phrase typing_is_over outstream (Impl(desc,loc)) =
  reset_type_expression_vars [];
  begin match desc with
    Zexpr expr ->
      type_expression loc expr;
      emit_phrase outstream
        (compile_lambda false (translate_expression expr))
  | Zletdef(rec_flag, pat_expr_list) ->
      type_letdef loc rec_flag pat_expr_list;
      emit_phrase outstream
       (if rec_flag
        then compile_lambda true  (translate_letdef_rec loc pat_expr_list)
        else compile_lambda false (translate_letdef loc pat_expr_list))
  | Ztypedef decl ->
      type_typedecl loc decl;
      semaphore__V typing_is_over
  | Zexcdef decl ->
      type_excdecl loc decl;
      semaphore__V typing_is_over
  | Zimpldirective dir ->
      do_directive dir;
      semaphore__V typing_is_over
  end;
  flush std_out
;;

let compile_intf_phrase typing_is_over (Intf(desc,loc)) =
  begin match desc with
    Zvaluedecl decl ->
      type_valuedecl loc decl
  | Ztypedecl decl ->
      type_typedecl loc decl;
      semaphore__V typing_is_over
  | Zexcdecl decl ->
      type_excdecl loc decl;
      semaphore__V typing_is_over
  | Zintfdirective dir ->
      do_directive dir;
      semaphore__V typing_is_over
  end;
  flush std_out
;;

(* Compilation d'une implementation *)

let compile_impl name filename =
  let impl_name =
    name ^ ".zo" in
  let ic =
    open_in filename
  and oc =
    open_out impl_name in
  let lexbuf =
    lexing__create_lexer_channel ic in
  let lexing_prod =
    producer__new (lexing_process lexbuf) in
  get_lexeme_buffer := [];
  let typing_is_over =
    semaphore__new () in
  let parsing_prod =
    producer__new (parsing_impl_process lexing_prod typing_is_over) in
(***
  let typing_prod =
    producer__new
      (typing_impl_process typing_is_over lexing_prod parsing_prod) in
***)
  input_name := filename;
  output_int oc 0;
  abs_out_pos := 0;
  reloc__reset();
  try
    while true do
      compile_impl_phrase typing_is_over oc
                          (parse_phrase lexing_prod parsing_prod)
    done
  with End_of_file ->
    output_value oc (reloc__get_info ());
    seek_out oc 0;
    output_int oc !abs_out_pos;
    prerr_string "Parsing: "; producer__kill parsing_prod;
    prerr_string "Lexing:  "; producer__kill lexing_prod;
    close_in ic;
    close_out oc
  | x ->
    producer__kill parsing_prod;
    producer__kill lexing_prod;
    close_in ic;
    close_out oc;
    unix__unlink impl_name;
    raise x
;;

(* Compilation d'une implementation *)

let compile_implementation name filename =
  new_globalenv name;
  if file_exists (name ^ ".mli") then begin
    let intf = load_module name in
      enter_interface_definitions intf;
      compile_impl name filename;
      check_interface intf
  end else begin
    let intf_name = name ^ ".zi" in
    let oc = open_out intf_name in
      try
        compile_impl name filename;
        write_compiled_interface oc;
        close_out oc
      with x ->
        close_out oc;
        unix__unlink intf_name;
        raise x
  end
;;

(* Compilation d'une interface *)

let compile_interface name filename =
  let intf_name =
    name ^ ".zi" in
  let ic =
    open_in filename
  and oc =
    open_out intf_name in
  new_globalenv name;
  type_stamp := 268435456;
(* Hack pour eviter de donner le meme stamp a un type defini ou declare dans
   l'interface et a un type defini dans l'implementation *)
  let lexbuf =
    lexing__create_lexer_channel ic in
  let lexing_prod =
    producer__new (lexing_process lexbuf) in
  get_lexeme_buffer := [];
  let typing_is_over =
    semaphore__new () in
  let parsing_prod =
    producer__new (parsing_intf_process lexing_prod typing_is_over) in
  input_name := filename;
  try
    while true do
      compile_intf_phrase typing_is_over (parse_phrase lexing_prod parsing_prod)
    done
  with End_of_file ->
    prerr_string "Parsing: "; producer__kill parsing_prod;
    prerr_string "Lexing:  "; producer__kill lexing_prod;
    close_in ic;
    write_compiled_interface oc;
    close_out oc
  | x ->
    producer__kill parsing_prod;
    producer__kill lexing_prod;
    close_in ic;
    close_out oc;
    unix__unlink intf_name;
    raise x
;;

