(* config.ml.unix.l1 *)

let maxint_byte = 127
and minint_byte = (-128)
and rangeint_byte = 256
and maxint_short = 32767
and minint_short = (-32768)
;;

(* Endianness of the default target machine. *)
let big_endian = ref false;;

(* Default alignment of the target machine.
 * Depends also on the target byte-code interpreter.
 * 1 if no alignment constraints OR if the runtime was compiled
 *    with the INTERP_ALIGN flag
 * 2 if 16-bits or 32-bits accesses require an even address
 * 4 if 16-bits accesses require an even address
 *   and 32-bits accesses require an address = 0 mod 4
 *)
let access_alignment = ref 1;;

(* The available libraries. *)
let default_used_interfaces =
  ["none",     ["builtin"];
   "fast",     ["io"; "eq"; "int"; "float"; "ref"; "pair"; "list";
                "fvect"; "fchar"; "fstring"; "bool";
                "exc"; "builtin"];
   "cautious", ["io"; "eq"; "int"; "float"; "ref"; "pair"; "list";
                "cvect"; "cchar"; "cstring"; "bool"; "exc"; "builtin"]]

and default_used_implementations =
  ["none",     [];
   "fast",     ["eq"; "exc"; "fstring"; "fchar"; "list"; "fvect";
                "io"; "pair"];
   "cautious", ["eq"; "exc"; "fstring"; "cstring";
                "fchar"; "cchar"; "list"; "fvect";
                "cvect"; "io"; "list"; "pair"]]

and default_exec_name =
  "a.out"
;;

(* The ususal line-feed/new-line/carriage-return mess.
 * Numbers are in decimal.
 *)
let newline_char = `\010` and return_char = `\013`;;

(* System-dependent constants.
 * file_system_type: "unix", "macintosh", or "ms-dos".
 *
 * toplevel_input_prompt: Printed before user input.
 * toplevel_output_prompt: Printed before compiler ouput in the toplevel system.
 * toplevel_error_prompt: Printed before compiler error and warning messages.
 * batch_output_prompt: Printed before compiler output in the batch system.
 * batch_error_prompt: Printed before error and warning messages.
 *)
let file_system_type = "unix";;
let toplevel_input_prompt = "#";;
let toplevel_output_prompt = "";;
let toplevel_error_prompt = ">";;
let batch_output_prompt = "";;
let batch_error_prompt = "#";;

(* Site-dependent constants.
 * path_library: Location of the standard library .zo and .zi files.
 * path_runtime: Header of bytecode files. If possible, it runs the bytecode
 *               interpreter automatically. It must end with a \010 char and
 *               be less than 256 chars long.
 * On BSD machines: "#!/usr/local/lib/caml-light/run\010"
 * (the full path to the bytecode interpreter).
 * works fine, provided the whole string is less than 32 chars long.
 * On other flavors of Unix: try
 *    "exec /usr/local/lib/caml-light/run $0 $*\010"
 *)
let path_library = "/usr/local/lib/caml-light/";;
let path_runtime = "#!/usr/local/bin/camlrun\010";;
