#open "obj";;
#open "lexing";;


#open "misc";;
#open "parser";;
#open "hash";;

(* Pour singaler les erreurs lexicales *)

exception Lexical_error of string * int * int
;;

(* Pour les commentaires emboites *)

let comment_depth = ref 0;;

(* La table des mots-cles *)

let keyword_table = make_vect 31 ([] : (string * token) list)
and infix_table = make_vect 17  ([] : string list);;

let add_keyword (s,tok as pair) =
  let i = hash_univ_param 1 1 s mod  31 in
    keyword_table.(i) <- pair :: keyword_table.(i)
;;

do_list add_keyword [
  "and", AND;
  "as", AS;
  "begin", BEGIN;
  "do", DO;
  "done", DONE;
  "down", DOWN;
  "else", ELSE;
  "end", END;
  "exception", EXCEPTION;
  "for", FOR;
  "fun", FUN;
  "function", FUNCTION;
  "if", IF;
  "in", IN;
  "let", LET;
  "match", MATCH;
  "mutable", MUTABLE;
  "not", NOT;
  "of", OF;
  "or", OR;
  "prefix", PREFIX;
  "rec", REC;
  "then", THEN;
  "to", TO;
  "try", TRY;
  "type", TYPE;
  "value", VALUE;
  "where", WHERE;
  "while", WHILE;
  "with", WITH
];;

let find_keyword s =
  match keyword_table.(hash_univ_param 1 1 s  mod  31) with
    [] -> raise (failure "find")
  | (s1,tok1)::rest1 ->
      if s = s1 then tok1 else match rest1 with
        [] -> raise (failure "find")
      | (s2,tok2)::rest2 ->
          if s = s2 then tok2 else assoc s rest2
;;

(* Les infixes *)

let is_infix s =
  match infix_table.(hash_univ_param 1 1 s  mod  17) with
    [] -> false
  | s1::rest1 ->
      s = s1 or match rest1 with
        [] -> false
      | s2::rest2 ->
          s = s2 or mem s rest2
;;

let add_infix s =
  let i = hash_univ_param 1 1 s  mod  17 in
    infix_table.(i) <- s :: infix_table.(i); ()
;;

let remove_infix s =
  let i = hash_univ_param 1 1 s  mod  17 in
    infix_table.(i) <- except s infix_table.(i); ()
;;

do_list add_infix ["quo"; "mod"]
;;

(* Bufferisation des chaines a la lecture *)

let initial_string_buffer = create_string 256;;
let string_buff = ref initial_string_buffer;;
let string_index = ref 0;;

let reset_string_buffer () =
  string_buff := initial_string_buffer;
  string_index := 0;
  ()
;;

let store_string_char c =
  (if !string_index >= string_length (!string_buff) then
    let new_buff = create_string (string_length (!string_buff) * 2) in
      blit_string (!string_buff) 0 new_buff 0 (string_length (!string_buff));
      string_buff := new_buff;
      ());
  set_nth_char (!string_buff) (!string_index) c;
  incr string_index
;;

let get_stored_string () =
  let s = sub_string (!string_buff) 0 (!string_index) in
    string_buff := initial_string_buffer;
    s
;;

(* Les caracteres speciaux *)

let char_for_backslash = function
    `n` -> config__newline_char
  | `r` -> config__return_char
  | `b` -> `\008`
  | `t` -> `\009`
  | c   -> c
;;

let char_for_decimal_code lexbuf i =
  char_of_int(100 * (int_of_char(get_lexeme_char lexbuf i) - 48) +
               10 * (int_of_char(get_lexeme_char lexbuf (i+1)) - 48) +
                    (int_of_char(get_lexeme_char lexbuf (i+2)) - 48))
;;


let rec action_62 lexbuf = (
 raise (Lexical_error("illegal character",
                            get_lexeme_start lexbuf, get_lexeme_end lexbuf)) )
and action_61 lexbuf = (
 EOF )
and action_60 lexbuf = (
 RBRACE )
and action_59 lexbuf = (
 BARRBRACKET )
and action_58 lexbuf = (
 BAR )
and action_57 lexbuf = (
 LBRACE )
and action_56 lexbuf = (
 UNDERUNDER )
and action_55 lexbuf = (
 UNDERSCORE )
and action_54 lexbuf = (
 CARET )
and action_53 lexbuf = (
 RBRACKET )
and action_52 lexbuf = (
 LBRACKETBAR )
and action_51 lexbuf = (
 LBRACKET )
and action_50 lexbuf = (
 AT )
and action_49 lexbuf = (
 GREATEREQUAL )
and action_48 lexbuf = (
 GREATER )
and action_47 lexbuf = (
 EQUALEQUAL )
and action_46 lexbuf = (
 EQUAL )
and action_45 lexbuf = (
 LESSGREATER )
and action_44 lexbuf = (
 LESSEQUAL )
and action_43 lexbuf = (
 LESSMINUS )
and action_42 lexbuf = (
 LESS )
and action_41 lexbuf = (
 SEMISEMI )
and action_40 lexbuf = (
 SEMI )
and action_39 lexbuf = (
 COLONEQUAL )
and action_38 lexbuf = (
 COLONCOLON )
and action_37 lexbuf = (
 COLON )
and action_36 lexbuf = (
 SLASH )
and action_35 lexbuf = (
 DOTLPAREN )
and action_34 lexbuf = (
 DOT )
and action_33 lexbuf = (
 MINUSGREATER )
and action_32 lexbuf = (
 MINUS )
and action_31 lexbuf = (
 COMMA )
and action_30 lexbuf = (
 PLUS )
and action_29 lexbuf = (
 STAR )
and action_28 lexbuf = (
 RPAREN )
and action_27 lexbuf = (
 LPAREN )
and action_26 lexbuf = (
 QUOTE )
and action_25 lexbuf = (
 AMPERSAND )
and action_24 lexbuf = (
 BANGEQUAL )
and action_23 lexbuf = (
 BANG )
and action_22 lexbuf = (
 SHARP )
and action_21 lexbuf = (
 comment_depth := 1; Comment lexbuf; Main lexbuf )
and action_20 lexbuf = (
 CHAR (Char lexbuf) )
and action_19 lexbuf = (
 reset_string_buffer();
        String lexbuf;
        STRING (get_stored_string()) )
and action_18 lexbuf = (
 INT (int_of_string(get_lexeme lexbuf)) )
and action_17 lexbuf = (
 FLOAT (float_of_string(get_lexeme lexbuf)) )
and action_16 lexbuf = (
 let s = get_lexeme lexbuf in
          try
            find_keyword s
          with failure "find" ->
            if is_infix s then INFIX s else IDENT s )
and action_15 lexbuf = (
 Main lexbuf )
and action_14 lexbuf = (
 Comment lexbuf )
and action_13 lexbuf = (
 raise (Lexical_error("comment not terminated", -1, -1)) )
and action_12 lexbuf = (
 reset_string_buffer();
        String lexbuf;
        reset_string_buffer();
        Comment lexbuf )
and action_11 lexbuf = (
 comment_depth := pred !comment_depth;
        if !comment_depth > 0 then Comment lexbuf )
and action_10 lexbuf = (
 comment_depth := succ !comment_depth; Comment lexbuf )
and action_9 lexbuf = (
 raise (Lexical_error("bad character constant",
                            get_lexeme_start lexbuf, get_lexeme_end lexbuf)) )
and action_8 lexbuf = (
 char_for_decimal_code lexbuf 1 )
and action_7 lexbuf = (
 char_for_backslash (get_lexeme_char lexbuf 1) )
and action_6 lexbuf = (
 get_lexeme_char lexbuf 0 )
and action_5 lexbuf = (
 store_string_char(get_lexeme_char lexbuf 0);
        String lexbuf )
and action_4 lexbuf = (
 raise (Lexical_error("string not terminated", -1, -1)) )
and action_3 lexbuf = (
 store_string_char(char_for_decimal_code lexbuf 1);
         String lexbuf )
and action_2 lexbuf = (
 store_string_char(char_for_backslash(get_lexeme_char lexbuf 1));
        String lexbuf )
and action_1 lexbuf = (
 String lexbuf )
and action_0 lexbuf = (
 () )
and state_0 lexbuf =
  match get_next_char lexbuf with
    `\255` -> state_81 lexbuf
 |  `\\` -> state_83 lexbuf
 |  `"` -> state_84 lexbuf
 |  _ -> state_82 lexbuf
and state_1 lexbuf =
  match get_next_char lexbuf with
    `\255` -> backtrack lexbuf
 |  `\`` -> state_71 lexbuf
 |  `\\` -> state_72 lexbuf
 |  _ -> state_70 lexbuf
and state_2 lexbuf =
  match get_next_char lexbuf with
    `\255` -> state_63 lexbuf
 |  `*` -> state_65 lexbuf
 |  `(` -> state_66 lexbuf
 |  `"` -> state_67 lexbuf
 |  _ -> state_64 lexbuf
and state_3 lexbuf =
  match get_next_char lexbuf with
    `z`|`y`|`x`|`w`|`v`|`u`|`t`|`s`|`r`|`q`|`p`|`o`|`n`|`m`|`l`|`k`|`j`|`i`|`h`|`g`|`f`|`e`|`d`|`c`|`b`|`a`|`Z`|`Y`|`X`|`W`|`V`|`U`|`T`|`S`|`R`|`Q`|`P`|`O`|`N`|`M`|`L`|`K`|`J`|`I`|`H`|`G`|`F`|`E`|`D`|`C`|`B`|`A` -> state_9 lexbuf
 |  `9`|`8`|`7`|`6`|`5`|`4`|`3`|`2`|`1`|`0` -> state_21 lexbuf
 |  ` `|`\026`|`\013`|`\012`|`\n`|`\t` -> state_35 lexbuf
 |  `\255` -> state_4 lexbuf
 |  `}` -> state_6 lexbuf
 |  `|` -> state_7 lexbuf
 |  `{` -> state_8 lexbuf
 |  `\`` -> state_10 lexbuf
 |  `_` -> state_11 lexbuf
 |  `^` -> state_12 lexbuf
 |  `]` -> state_13 lexbuf
 |  `[` -> state_14 lexbuf
 |  `@` -> state_15 lexbuf
 |  `>` -> state_16 lexbuf
 |  `=` -> state_17 lexbuf
 |  `<` -> state_18 lexbuf
 |  `;` -> state_19 lexbuf
 |  `:` -> state_20 lexbuf
 |  `/` -> state_22 lexbuf
 |  `.` -> state_23 lexbuf
 |  `-` -> state_24 lexbuf
 |  `,` -> state_25 lexbuf
 |  `+` -> state_26 lexbuf
 |  `*` -> state_27 lexbuf
 |  `)` -> state_28 lexbuf
 |  `(` -> state_29 lexbuf
 |  `'` -> state_30 lexbuf
 |  `&` -> state_31 lexbuf
 |  `#` -> state_32 lexbuf
 |  `"` -> state_33 lexbuf
 |  `!` -> state_34 lexbuf
 |  _ -> state_5 lexbuf
and state_4 = action_61
and state_5 = action_62
and state_6 = action_60
and state_7 lexbuf =
  lexbuf.lex_last_pos <- lexbuf.lex_curr_pos;
  lexbuf.lex_last_action <- magic action_58;
  match get_next_char lexbuf with
    `]` -> state_62 lexbuf
 |  _ -> backtrack lexbuf
and state_8 = action_57
and state_9 lexbuf =
  lexbuf.lex_last_pos <- lexbuf.lex_curr_pos;
  lexbuf.lex_last_action <- magic action_16;
  match get_next_char lexbuf with
    `z`|`y`|`x`|`w`|`v`|`u`|`t`|`s`|`r`|`q`|`p`|`o`|`n`|`m`|`l`|`k`|`j`|`i`|`h`|`g`|`f`|`e`|`d`|`c`|`b`|`a`|`Z`|`Y`|`X`|`W`|`V`|`U`|`T`|`S`|`R`|`Q`|`P`|`O`|`N`|`M`|`L`|`K`|`J`|`I`|`H`|`G`|`F`|`E`|`D`|`C`|`B`|`A`|`9`|`8`|`7`|`6`|`5`|`4`|`3`|`2`|`1`|`0`|`'` -> state_59 lexbuf
 |  `_` -> state_61 lexbuf
 |  _ -> backtrack lexbuf
and state_10 = action_20
and state_11 lexbuf =
  lexbuf.lex_last_pos <- lexbuf.lex_curr_pos;
  lexbuf.lex_last_action <- magic action_55;
  match get_next_char lexbuf with
    `z`|`y`|`x`|`w`|`v`|`u`|`t`|`s`|`r`|`q`|`p`|`o`|`n`|`m`|`l`|`k`|`j`|`i`|`h`|`g`|`f`|`e`|`d`|`c`|`b`|`a`|`Z`|`Y`|`X`|`W`|`V`|`U`|`T`|`S`|`R`|`Q`|`P`|`O`|`N`|`M`|`L`|`K`|`J`|`I`|`H`|`G`|`F`|`E`|`D`|`C`|`B`|`A`|`9`|`8`|`7`|`6`|`5`|`4`|`3`|`2`|`1`|`0`|`'` -> state_59 lexbuf
 |  `_` -> state_60 lexbuf
 |  _ -> backtrack lexbuf
and state_12 = action_54
and state_13 = action_53
and state_14 lexbuf =
  lexbuf.lex_last_pos <- lexbuf.lex_curr_pos;
  lexbuf.lex_last_action <- magic action_51;
  match get_next_char lexbuf with
    `|` -> state_58 lexbuf
 |  _ -> backtrack lexbuf
and state_15 = action_50
and state_16 lexbuf =
  lexbuf.lex_last_pos <- lexbuf.lex_curr_pos;
  lexbuf.lex_last_action <- magic action_48;
  match get_next_char lexbuf with
    `=` -> state_57 lexbuf
 |  _ -> backtrack lexbuf
and state_17 lexbuf =
  lexbuf.lex_last_pos <- lexbuf.lex_curr_pos;
  lexbuf.lex_last_action <- magic action_46;
  match get_next_char lexbuf with
    `=` -> state_56 lexbuf
 |  _ -> backtrack lexbuf
and state_18 lexbuf =
  lexbuf.lex_last_pos <- lexbuf.lex_curr_pos;
  lexbuf.lex_last_action <- magic action_42;
  match get_next_char lexbuf with
    `>` -> state_53 lexbuf
 |  `=` -> state_54 lexbuf
 |  `-` -> state_55 lexbuf
 |  _ -> backtrack lexbuf
and state_19 lexbuf =
  lexbuf.lex_last_pos <- lexbuf.lex_curr_pos;
  lexbuf.lex_last_action <- magic action_40;
  match get_next_char lexbuf with
    `;` -> state_52 lexbuf
 |  _ -> backtrack lexbuf
and state_20 lexbuf =
  lexbuf.lex_last_pos <- lexbuf.lex_curr_pos;
  lexbuf.lex_last_action <- magic action_37;
  match get_next_char lexbuf with
    `=` -> state_50 lexbuf
 |  `:` -> state_51 lexbuf
 |  _ -> backtrack lexbuf
and state_21 lexbuf =
  lexbuf.lex_last_pos <- lexbuf.lex_curr_pos;
  lexbuf.lex_last_action <- magic action_18;
  match get_next_char lexbuf with
    `9`|`8`|`7`|`6`|`5`|`4`|`3`|`2`|`1`|`0` -> state_42 lexbuf
 |  `e`|`E` -> state_41 lexbuf
 |  `.` -> state_43 lexbuf
 |  _ -> backtrack lexbuf
and state_22 = action_36
and state_23 lexbuf =
  lexbuf.lex_last_pos <- lexbuf.lex_curr_pos;
  lexbuf.lex_last_action <- magic action_34;
  match get_next_char lexbuf with
    `(` -> state_40 lexbuf
 |  _ -> backtrack lexbuf
and state_24 lexbuf =
  lexbuf.lex_last_pos <- lexbuf.lex_curr_pos;
  lexbuf.lex_last_action <- magic action_32;
  match get_next_char lexbuf with
    `>` -> state_39 lexbuf
 |  _ -> backtrack lexbuf
and state_25 = action_31
and state_26 = action_30
and state_27 = action_29
and state_28 = action_28
and state_29 lexbuf =
  lexbuf.lex_last_pos <- lexbuf.lex_curr_pos;
  lexbuf.lex_last_action <- magic action_27;
  match get_next_char lexbuf with
    `*` -> state_38 lexbuf
 |  _ -> backtrack lexbuf
and state_30 = action_26
and state_31 = action_25
and state_32 = action_22
and state_33 = action_19
and state_34 lexbuf =
  lexbuf.lex_last_pos <- lexbuf.lex_curr_pos;
  lexbuf.lex_last_action <- magic action_23;
  match get_next_char lexbuf with
    `=` -> state_37 lexbuf
 |  _ -> backtrack lexbuf
and state_35 lexbuf =
  lexbuf.lex_last_pos <- lexbuf.lex_curr_pos;
  lexbuf.lex_last_action <- magic action_15;
  match get_next_char lexbuf with
    ` `|`\026`|`\013`|`\012`|`\n`|`\t` -> state_36 lexbuf
 |  _ -> backtrack lexbuf
and state_36 lexbuf =
  lexbuf.lex_last_pos <- lexbuf.lex_curr_pos;
  lexbuf.lex_last_action <- magic action_15;
  match get_next_char lexbuf with
    ` `|`\026`|`\013`|`\012`|`\n`|`\t` -> state_36 lexbuf
 |  _ -> backtrack lexbuf
and state_37 = action_24
and state_38 = action_21
and state_39 = action_33
and state_40 = action_35
and state_41 lexbuf =
  match get_next_char lexbuf with
    `9`|`8`|`7`|`6`|`5`|`4`|`3`|`2`|`1`|`0` -> state_48 lexbuf
 |  `-` -> state_49 lexbuf
 |  _ -> backtrack lexbuf
and state_42 lexbuf =
  lexbuf.lex_last_pos <- lexbuf.lex_curr_pos;
  lexbuf.lex_last_action <- magic action_18;
  match get_next_char lexbuf with
    `9`|`8`|`7`|`6`|`5`|`4`|`3`|`2`|`1`|`0` -> state_42 lexbuf
 |  `e`|`E` -> state_41 lexbuf
 |  `.` -> state_43 lexbuf
 |  _ -> backtrack lexbuf
and state_43 lexbuf =
  match get_next_char lexbuf with
    `9`|`8`|`7`|`6`|`5`|`4`|`3`|`2`|`1`|`0` -> state_44 lexbuf
 |  _ -> backtrack lexbuf
and state_44 lexbuf =
  lexbuf.lex_last_pos <- lexbuf.lex_curr_pos;
  lexbuf.lex_last_action <- magic action_17;
  match get_next_char lexbuf with
    `9`|`8`|`7`|`6`|`5`|`4`|`3`|`2`|`1`|`0` -> state_44 lexbuf
 |  `e`|`E` -> state_45 lexbuf
 |  _ -> backtrack lexbuf
and state_45 lexbuf =
  match get_next_char lexbuf with
    `9`|`8`|`7`|`6`|`5`|`4`|`3`|`2`|`1`|`0` -> state_46 lexbuf
 |  `-` -> state_47 lexbuf
 |  _ -> backtrack lexbuf
and state_46 lexbuf =
  lexbuf.lex_last_pos <- lexbuf.lex_curr_pos;
  lexbuf.lex_last_action <- magic action_17;
  match get_next_char lexbuf with
    `9`|`8`|`7`|`6`|`5`|`4`|`3`|`2`|`1`|`0` -> state_46 lexbuf
 |  _ -> backtrack lexbuf
and state_47 lexbuf =
  match get_next_char lexbuf with
    `9`|`8`|`7`|`6`|`5`|`4`|`3`|`2`|`1`|`0` -> state_46 lexbuf
 |  _ -> backtrack lexbuf
and state_48 lexbuf =
  lexbuf.lex_last_pos <- lexbuf.lex_curr_pos;
  lexbuf.lex_last_action <- magic action_17;
  match get_next_char lexbuf with
    `9`|`8`|`7`|`6`|`5`|`4`|`3`|`2`|`1`|`0` -> state_48 lexbuf
 |  _ -> backtrack lexbuf
and state_49 lexbuf =
  match get_next_char lexbuf with
    `9`|`8`|`7`|`6`|`5`|`4`|`3`|`2`|`1`|`0` -> state_48 lexbuf
 |  _ -> backtrack lexbuf
and state_50 = action_39
and state_51 = action_38
and state_52 = action_41
and state_53 = action_45
and state_54 = action_44
and state_55 = action_43
and state_56 = action_47
and state_57 = action_49
and state_58 = action_52
and state_59 lexbuf =
  lexbuf.lex_last_pos <- lexbuf.lex_curr_pos;
  lexbuf.lex_last_action <- magic action_16;
  match get_next_char lexbuf with
    `z`|`y`|`x`|`w`|`v`|`u`|`t`|`s`|`r`|`q`|`p`|`o`|`n`|`m`|`l`|`k`|`j`|`i`|`h`|`g`|`f`|`e`|`d`|`c`|`b`|`a`|`Z`|`Y`|`X`|`W`|`V`|`U`|`T`|`S`|`R`|`Q`|`P`|`O`|`N`|`M`|`L`|`K`|`J`|`I`|`H`|`G`|`F`|`E`|`D`|`C`|`B`|`A`|`9`|`8`|`7`|`6`|`5`|`4`|`3`|`2`|`1`|`0`|`'` -> state_59 lexbuf
 |  `_` -> state_61 lexbuf
 |  _ -> backtrack lexbuf
and state_60 = action_56
and state_61 lexbuf =
  match get_next_char lexbuf with
    `z`|`y`|`x`|`w`|`v`|`u`|`t`|`s`|`r`|`q`|`p`|`o`|`n`|`m`|`l`|`k`|`j`|`i`|`h`|`g`|`f`|`e`|`d`|`c`|`b`|`a`|`Z`|`Y`|`X`|`W`|`V`|`U`|`T`|`S`|`R`|`Q`|`P`|`O`|`N`|`M`|`L`|`K`|`J`|`I`|`H`|`G`|`F`|`E`|`D`|`C`|`B`|`A`|`9`|`8`|`7`|`6`|`5`|`4`|`3`|`2`|`1`|`0`|`'` -> state_59 lexbuf
 |  _ -> backtrack lexbuf
and state_62 = action_59
and state_63 = action_13
and state_64 = action_14
and state_65 lexbuf =
  lexbuf.lex_last_pos <- lexbuf.lex_curr_pos;
  lexbuf.lex_last_action <- magic action_14;
  match get_next_char lexbuf with
    `)` -> state_69 lexbuf
 |  _ -> backtrack lexbuf
and state_66 lexbuf =
  lexbuf.lex_last_pos <- lexbuf.lex_curr_pos;
  lexbuf.lex_last_action <- magic action_14;
  match get_next_char lexbuf with
    `*` -> state_68 lexbuf
 |  _ -> backtrack lexbuf
and state_67 = action_12
and state_68 = action_10
and state_69 = action_11
and state_70 lexbuf =
  lexbuf.lex_last_pos <- lexbuf.lex_curr_pos;
  lexbuf.lex_last_action <- magic action_9;
  match get_next_char lexbuf with
    `\`` -> state_80 lexbuf
 |  _ -> backtrack lexbuf
and state_71 lexbuf =
  lexbuf.lex_last_pos <- lexbuf.lex_curr_pos;
  lexbuf.lex_last_action <- magic action_9;
  match get_next_char lexbuf with
    `\`` -> state_79 lexbuf
 |  _ -> backtrack lexbuf
and state_72 lexbuf =
  lexbuf.lex_last_pos <- lexbuf.lex_curr_pos;
  lexbuf.lex_last_action <- magic action_9;
  match get_next_char lexbuf with
    `9`|`8`|`7`|`6`|`5`|`4`|`3`|`2`|`1`|`0` -> state_74 lexbuf
 |  `t`|`r`|`n`|`b`|`\``|`\\` -> state_73 lexbuf
 |  _ -> backtrack lexbuf
and state_73 lexbuf =
  match get_next_char lexbuf with
    `\`` -> state_78 lexbuf
 |  _ -> backtrack lexbuf
and state_74 lexbuf =
  match get_next_char lexbuf with
    `9`|`8`|`7`|`6`|`5`|`4`|`3`|`2`|`1`|`0` -> state_75 lexbuf
 |  _ -> backtrack lexbuf
and state_75 lexbuf =
  match get_next_char lexbuf with
    `9`|`8`|`7`|`6`|`5`|`4`|`3`|`2`|`1`|`0` -> state_76 lexbuf
 |  _ -> backtrack lexbuf
and state_76 lexbuf =
  match get_next_char lexbuf with
    `\`` -> state_77 lexbuf
 |  _ -> backtrack lexbuf
and state_77 = action_8
and state_78 = action_7
and state_79 = action_6
and state_80 = action_6
and state_81 = action_4
and state_82 = action_5
and state_83 lexbuf =
  lexbuf.lex_last_pos <- lexbuf.lex_curr_pos;
  lexbuf.lex_last_action <- magic action_5;
  match get_next_char lexbuf with
    `9`|`8`|`7`|`6`|`5`|`4`|`3`|`2`|`1`|`0` -> state_86 lexbuf
 |  `t`|`r`|`n`|`b`|`\\`|`"` -> state_85 lexbuf
 |  ` `|`\026`|`\013`|`\012`|`\n`|`\t` -> state_87 lexbuf
 |  _ -> backtrack lexbuf
and state_84 = action_0
and state_85 = action_2
and state_86 lexbuf =
  match get_next_char lexbuf with
    `9`|`8`|`7`|`6`|`5`|`4`|`3`|`2`|`1`|`0` -> state_88 lexbuf
 |  _ -> backtrack lexbuf
and state_87 lexbuf =
  lexbuf.lex_last_pos <- lexbuf.lex_curr_pos;
  lexbuf.lex_last_action <- magic action_1;
  match get_next_char lexbuf with
    ` `|`\026`|`\013`|`\012`|`\n`|`\t` -> state_87 lexbuf
 |  _ -> backtrack lexbuf
and state_88 lexbuf =
  match get_next_char lexbuf with
    `9`|`8`|`7`|`6`|`5`|`4`|`3`|`2`|`1`|`0` -> state_89 lexbuf
 |  _ -> backtrack lexbuf
and state_89 = action_3
and Main lexbuf =
  lexbuf.lex_start_pos <- lexbuf.lex_curr_pos;
  state_3 lexbuf

and Comment lexbuf =
  lexbuf.lex_start_pos <- lexbuf.lex_curr_pos;
  state_2 lexbuf

and Char lexbuf =
  lexbuf.lex_start_pos <- lexbuf.lex_curr_pos;
  state_1 lexbuf

and String lexbuf =
  lexbuf.lex_start_pos <- lexbuf.lex_curr_pos;
  state_0 lexbuf
;;
