(* The run-time library for lexers generated by mllex *)

#open "io";;
#open "obj";;

(* The following definitions are used by the generated scanners only.
   They are not intended to be used by user programs. *)

type lexbuf =
  { refill_buff : string -> int -> int;
    lex_buffer : string;
    mutable lex_abs_pos : int;
    mutable lex_start_pos : int;
    mutable lex_curr_pos : int;
    mutable lex_last_pos : int;
    mutable lex_last_action : lexbuf -> obj }
;;

value dummy_action : lexbuf -> obj
  and lex_refill : lexbuf -> unit
  and get_next_char : lexbuf -> char
  and get_lexeme : lexbuf -> string
  and get_lexeme_char : lexbuf -> int -> char
  and backtrack : lexbuf -> 'a
  and get_lexeme_start : lexbuf -> int
  and get_lexeme_end : lexbuf -> int
;;

(* The following two functions can be called by user code. *)

value create_lexer_channel : in_channel -> lexbuf
        (* Creates a lexer buffer on the given input channel.
           The ``lexer buffer'' is the argument passed to the scanning
           functions defined by the generated scanners. The lexer buffer
           holds the current state of the scanner, plus a function
           to read more characters from the input. create_lexer_channel
           returns a lexer buffer that reads form the given input channel,
           at the current position. *)
  and create_lexer : (string -> int -> int) -> lexbuf
        (* Creates a lexer buffer with the given function as reading method.
           When the scanner needs more characters, it will call the given
           function, giving it a character string s and a character count n.
           The given function should put at most n characters in s,
           starting at position 0, and return the number of characters
           provided. A return value of 0 means end of input.
           Example: to create a lexer buffer that reads from string s:
                create_lexer (let pos = ref 0 in
                  fun buffer req ->
                    let avail = string_length s - !pos in
                    let n = if avail < req then avail else req in
                      blit_string s !pos buffer 0 n;
                      pos := !pos + n;
                      n)
         *)
                
;;
