#open "unix";;
#open "misc";;
#open "instruct";;
#open "config";;
#open "opcodes";;
#open "symtable";;
#open "reloc";;
#open "emitdata";;

(* Production du code pour un fichier objet *)

let link_object outchan name =
  let truename = find_in_path (name ^ ".zo") !modules__path_interfaces in
  let inchan = open_in truename in
    try
      object_name := name;
      let n = input_int inchan in
      let buff = create_string n in
      input inchan buff 0 n;
      patch_object buff (input_value inchan : (reloc__info * int) list);
      output outchan buff 0 n;
      close_in inchan
    with x ->
      prerr_begline ">> Error while linking file ";
      prerr_string truename;
      prerr_endline ".";
      close_in inchan;
      raise x
;;

let write_symbols = ref false;;

(* Construction d'un executable bytecode *)

let link module_list exec_name =
  reset_linker_tables();
  let outchan =
    open_out_gen (mode_rdwr + mode_trunc + mode_creat)
                 (rights_everything * rights_everybody) exec_name in
  try
    output_string outchan path_runtime;
    let pos0 = pos_out outchan in
    let pos1 = pos0 + 16 in
(* On reserve de la place pour l'en-tete *)
    seek_out outchan pos1;
(* Le bytecode *)
    do_list (link_object outchan) module_list;
    output_char outchan (char_of_int STOP);
(* Les donnees initialisees *)
    let pos2 = pos_out outchan in
    emit_data_to_file outchan;
(* Les tables du linker *)
    let pos3 = pos_out outchan in
    if !write_symbols then save_linker_tables outchan;
(* Les infos de debug (inexistantes pour l'instant) *)
    let pos4 = pos_out outchan in
(* On met a jour l'en-tete avec les tailles *)
    seek_out outchan pos0;
    output_int outchan (pos2 - pos1);
    output_int outchan (pos3 - pos2);
    output_int outchan (pos4 - pos3);
    output_int outchan 0;
    close_out outchan
  with x ->
    unlink exec_name;
    close_out outchan;
    raise x
;;

