#open "const";;
#open "globals";;
#open "location";;
#open "syntax";;
#open "modules";;
#open "builtins";;
#open "errors";;

let make_expr desc = Expr(desc, get_current_location())
and make_pat desc = Pat(desc, get_current_location())
and make_typ desc = Typexp(desc, get_current_location())
and make_impl desc = Impl(desc, get_current_location())
and make_intf desc = Intf(desc, get_current_location())
;;

let make_apply = function
    Expr(Zconstruct(cstr1, []), _), [Expr(Zconstruct(cstr2, el), _) as e2] ->
      if is_constr_tuple cstr2 then
        make_expr(Zconstruct(cstr1, el))
      else
        make_expr(Zconstruct(cstr1, [e2]))
  | Expr(Zconstruct(cstr1, []), _), [e2] ->
      make_expr(Zconstruct(cstr1, [e2]))
  | e1, el ->
      make_expr(Zapply(e1,el))
;;

let make_unop op e =
  make_expr(Zapply(make_expr(Zident(ref (Zlocal op))), [e]))
and make_binop op e1 e2 =
  make_expr(Zapply(make_expr(Zident(ref (Zlocal op))), [e1;e2]))
and make_ternop op e1 e2 e3 =
  make_expr(Zapply(make_expr(Zident(ref (Zlocal op))), [e1;e2;e3]))
;;

let make_tuple_expr el =
  make_expr(Zconstruct(constr_tuple (list_length el), el))
and make_tuple_pat pl =
  make_pat(Zconstructpat(constr_tuple (list_length pl), pl))
;;

let make_list =
  makel (make_expr(Zconstruct(constr_nil, [])))
  where rec makel res = function
    [] -> res
  | e::l -> makel (make_expr(Zconstruct(constr_cons, [e;res]))) l
;;

let make_unary_minus = function
    Expr(Zconstant(SCatom(ACint i)), _) ->
      make_expr(Zconstant(SCatom(ACint(minus i))))
  | Expr(Zconstant(SCatom(ACfloat f)), _) ->
      make_expr(Zconstant(SCatom(ACfloat(minus_float f))))
  | e ->
      make_unop "minus" e
;;

let find_constructor gr =
  try
    find_constr_desc gr
  with failure "find_desc" ->
    unbound_err "Constructor" gr (get_current_location())
;;

let find_label gr =
  try
    find_label_desc gr
  with failure "find_desc" ->
    unbound_err "Label" gr (get_current_location())
;;

let gr_type_tuple = GRmodname {qual="builtin"; id="*"}
;;

let expr_constr_or_ident = function
    GRname s as gr ->
     (try
        make_expr(Zconstruct(find_constr_desc gr, []))
      with failure "find_desc" ->
        make_expr(Zident(ref(Zlocal s))))
  | GRmodname q as gr ->
     (try
        make_expr(Zconstruct(find_constr_desc gr, []))
      with failure "find_desc" ->
        make_expr(Zident(ref(Zqualid q))))
;;

let pat_constr_or_var s =
  try
    make_pat(Zconstructpat(find_constr_desc (GRname s), []))
  with failure "find_desc" ->
    make_pat(Zvarpat s)
;;
