#open "eq";;
#open "exc";;
#open "int";;
#open "fvect";;
#open "ref";;
#open "obj";;
#open "lexing";;
(***
#open "io";;
***)

let char = ref 0                    (* Le dernier token lu *)
and lval = ref (Repr 0)             (* et son attribut semantique *)
and symb_start = ref 0              (* position du debut du symbole courant *)
and symb_end = ref 0                (* Position de fin du symbole courant *)
and sp = ref 0                      (* Pointeur de pile de l'automate *)
and lhs_start = ref 0               (* position du debut du nonterm reduit *)
and lhs_end = ref 0                 (* Position de fin du nonterm reduit *)
;;

type parser_stacks =
  { mutable s : int vect;
    mutable v : obj vect;
    mutable symb_start : int vect;
    mutable symb_end : int vect }
;;

let stacksize = ref 100;;               (* Taille des piles *)

let stacks =
  { s = make_vect !stacksize 0;
    v = make_vect !stacksize (Repr 0);
    symb_start = make_vect !stacksize 0;
    symb_end = make_vect !stacksize 0 }
;;

let grow_stacks() =
  let oldsize = !stacksize in
  let newsize = oldsize * 2 in
  let new_s = make_vect newsize 0
  and new_v = make_vect newsize (Repr 0)
  and new_start = make_vect newsize 0
  and new_end = make_vect newsize 0 in
    blit_vect stacks.s 0 new_s 0 oldsize;
    stacks.s <- new_s;
    blit_vect stacks.v 0 new_v 0 oldsize;
    stacks.v <- new_v;
    blit_vect stacks.symb_start 0 new_start 0 oldsize;
    stacks.symb_start <- new_start;
    blit_vect stacks.symb_end 0 new_end 0 oldsize;
    stacks.symb_end <- new_end;
    stacksize := newsize;
    ()
;;

let clear_parser() =
  fill_vect stacks.v 0 !stacksize (Repr 0);
  lval := Repr 0;
  ()
;;

let yyparse tables start lexer_fct lexer_arg =

  let rec loop state =
(***
    print_string "Loop "; print_int state; 
    print_string " "; print_int !symb_start;
    print_string "-"; print_int !symb_end; print_newline();
***)
    let yyn = vect16_item tables.defred state in
      if yyn != 0 then
        reduce state yyn
      else begin
        if !char < 0 then begin
          let (token, token_start, token_end) = lexer_fct lexer_arg in
            char := tables.transl.(obj_tag(Repr token));
            lval := obj_field (Repr token) 0;
            symb_start := token_start;
            symb_end := token_end;
(***
            print_string "Token lu: code ";
            print_int !char;
            print_string " attribut ";
            print_int (magic !lval);
            print_string " "; print_int !symb_start;
            print_string "-"; print_int !symb_end; 
            print_newline();
***)
            ()
        end;
        let n1 = vect16_item tables.sindex state in
        let n2 = n1 + !char in
          if n1 != 0 & n2 >= 0 & n2 <= tables.tablesize &
             vect16_item tables.check n2 == !char
          then begin
            let sp2 = incr sp in
            if sp2 >= !stacksize then grow_stacks();
            let state2 = vect16_item tables.table n2 in
(***
              print_string "Shift ";
              print_int state2;
              print_newline();
***)
              stacks.s.(sp2) <- state2;
              stacks.v.(sp2) <- !lval;
              stacks.symb_start.(sp2) <- !symb_start;
              stacks.symb_end.(sp2) <- !symb_end;
              char := (-1);
              loop state2
          end else begin
            let n1 = vect16_item tables.rindex state in
            let n2 = n1 + !char in
              if n1 != 0 & n2 >= 0 & n2 <= tables.tablesize &
                 vect16_item tables.check n2 == !char
              then
                reduce state (vect16_item tables.table n2)
              else
                let c = !char in
                  raise (Parse_error
                           (fun tok -> tables.transl.(obj_tag tok) == c))
          end
      end

  and reduce state n =
    let sp2 = succ(!sp - vect16_item tables.len n) in
      lhs_start := stacks.symb_start.(sp2);
      lhs_end := stacks.symb_end.(!sp);
(***
      print_string "Reduce ";
      print_int n;
      print_string " "; print_int !lhs_start;
      print_string "-"; print_int !lhs_end; 
      print_newline();
***)
      let val = tables.actions.(n) () in
      let m = vect16_item tables.lhs n in
      let state1 =
        stacks.s.(pred sp2) in
      let state2 =
        let n1 = vect16_item tables.gindex m in
        let n2 = n1 + state1 in
          if n1 != 0 & n2 >= 0 & n2 <= tables.tablesize &
             vect16_item tables.check n2 == state1
          then
            vect16_item tables.table n2
          else
            vect16_item tables.dgoto m
      in
(***
        print_string "Goto ";
        print_int state2;
        print_newline();
***)
        if sp2 >= !stacksize then grow_stacks();
        stacks.s.(sp2) <- state2;
        stacks.v.(sp2) <- val;
        stacks.symb_end.(sp2) <- !lhs_end;
        sp := sp2;
        loop state2

  in
    char := start;
    sp := 0;
(***
    print_string "Token initial: ";
    print_int !char;
    print_newline();
***)
    try loop 0 with yyexit v -> magic_obj v
;;

let peek_val n =
  magic_obj stacks.v.(!sp - n)
;;

let symbol_start () =
  !lhs_start
and symbol_end () =
  !lhs_end
;;
