#open "misc";;

type 'a result =
    None
  | Result of 'a
  | Exception of exn
;;

type 'a t = {
  fct: unit -> 'a;                    (* fonction de production *)
  data : 'a result vect;              (* les resultats *)
  mutable producer : int;             (* position du producteur *)
  mutable consumer : int;             (* position du consommateur *)
  mutable count : int;                (* nombre d'objets disponibles *)
  mutable killed : bool;              (* on s'arrete de produire *)
  lock : thread__mutex;               (* pour l'acces exclusif *)
  nonempty : thread__condition;       (* quelque chose a ete mis *)
  nonfull : thread__condition;        (* quelque chose a ete lu *)
  mutable cnt_empty : int;      (* nbr. de fois ou le lecteur s'est bloque *)
  mutable cnt_full : int        (* nbr. de fois ou l'ecrivain s'est bloque *)
};;

let producer buffer () =
  while true do
    let res =
      begin try
        Result(buffer.fct ())
      with x ->
        Exception x
      end in
    thread__lock buffer.lock;
    while buffer.count >= vect_length buffer.data do
      buffer.cnt_full <- buffer.cnt_full + 1;
      thread__wait buffer.nonfull buffer.lock
    done;
    if buffer.killed then begin
      thread__unlock buffer.lock;
      thread__exit ()
    end;
    buffer.data.(buffer.producer) <- res;
    let n = succ buffer.producer in
      buffer.producer <- if n < vect_length buffer.data then n else 0;
    if buffer.count == 0 then
        thread__signal buffer.nonempty;
    buffer.count <- succ buffer.count;
    thread__unlock buffer.lock
  done
;;    

let new fct =
  let buff =
    { fct = fct;
      data = make_vect 100 None;
      producer = 0;
      consumer = 0;
      count = 0;
      killed = false;
      lock = thread__new_mutex();
      nonempty = thread__new_condition();
      nonfull = thread__new_condition();
      cnt_empty = 0;
      cnt_full = 0 } in
  thread__detach(thread__fork (producer buff));
  buff
;;

let get buffer =
  thread__lock buffer.lock;
  while buffer.count == 0 do
    buffer.cnt_empty <- buffer.cnt_empty + 1;
    thread__wait buffer.nonempty buffer.lock
  done;
  let res = buffer.data.(buffer.consumer) in
  buffer.data.(buffer.consumer) <- None;
  let n = succ buffer.consumer in
    buffer.consumer <- if n < vect_length buffer.data then n else 0;
  if buffer.count >= vect_length buffer.data then
    thread__signal buffer.nonfull;
  buffer.count <- pred buffer.count;
  thread__unlock buffer.lock;
  match res with
      Result x -> x
    | Exception e -> raise e
    | None -> fatal_error "producer__get"
;;

let kill buffer =
  thread__lock buffer.lock;
  prerr_string "empty: "; prerr_int buffer.cnt_empty;
  prerr_string " full: "; prerr_int buffer.cnt_full;
  prerr_endline "";
  buffer.killed <- true;
  buffer.count <- 0;
  thread__signal buffer.nonfull;
  thread__unlock buffer.lock
;;

