#open "const";;
#open "location";;
#open "globals";;

(* syntax.ml : la syntaxe abstraite du langage *)

type type_expression =
    Typexp of type_expression_desc * location
and type_expression_desc =
    Ztypevar of string
  | Ztypearrow of type_expression * type_expression
  | Ztypeconstr of global_reference * type_expression list
;;

type pattern =
    Pat of pattern_desc * location
and pattern_desc =
    Zwildpat
  | Zvarpat of string
  | Zaliaspat of pattern * string
  | Zconstantpat of atomic_constant
  | Zconstructpat of constr_desc global * pattern list
  | Zorpat of pattern * pattern
  | Zconstraintpat of pattern * type_expression
  | Zrecordpat of (label_desc global * pattern) list
;;

type expression =
    Expr of expression_desc * location
and expression_desc =
    Zident of expr_ident ref
  | Zconstant of struct_constant
  | Zconstruct of constr_desc global * expression list
  | Zapply of expression * expression list
  | Zlet of bool * (pattern * expression * int ref) list * expression
  | Zfunction of (pattern list * expression) list
  | Ztrywith of expression * (pattern * expression) list
  | Zsequence of expression * expression
  | Zcondition of expression * expression * expression
  | Zwhile of expression * expression
  | Zfor of string * expression * expression * bool * expression
  | Zsequand of expression * expression
  | Zsequor of expression * expression
  | Zconstraint of expression * type_expression
  | Zvector of expression list
  | Zassign of string * expression
  | Zrecord of (label_desc global * expression) list
  | Zrecord_access of expression * label_desc global
  | Zrecord_update of expression * label_desc global * expression

and expr_ident =
    Zlocal of string
  | Zglobal of value_desc global
  | Zqualid of qualified_ident
;;

type type_decl =
    Zabstract_type
  | Zvariant_type of
      (string * (type_expression * mutable_flag) list) list
  | Zrecord_type of
      (string * type_expression * mutable_flag) list
  | Zabbrev_type of type_expression
;;

type directiveu =
    Zdir of string * string
;;

type impl_phrase =
    Impl of impl_desc * location
and impl_desc =
    Zexpr of expression
  | Zletdef of bool * (pattern * expression * int ref) list
  | Ztypedef of (string * string list * type_decl) list
  | Zexcdef of (string * (type_expression * mutable_flag) list) list
  | Zimpldirective of directiveu
;;

type intf_phrase =
    Intf of intf_desc * location
and intf_desc =
    Zvaluedecl of (string * type_expression * prim_desc) list
  | Ztypedecl of (string * string list * type_decl) list
  | Zexcdecl of (string * (type_expression * mutable_flag) list) list
  | Zintfdirective of directiveu
;;

value free_vars_of_pat : pattern -> string list;;
