#open "misc";;
#open "const";;
#open "globals";;
#open "builtins";;
#open "location";;
#open "syntax";;
#open "modules";;
#open "types";;
#open "pr_type";;
#open "typing";;
#open "ty_error";;
(* Typage des phrases toplevel *)

let verbose = ref true;;

let enter_new_variant is_extensible loc (ty_desc, ty_res, constrs) =
  let nbr_constrs =
    list_length constrs in
  let rec make_constrs constr_idx = function
    [] -> []
  | (constr_name, args) :: rest ->
      if mem_assoc constr_name rest then
        duplicate_constr_in_type_decl_err "Constructor " loc constr_name;
      let ty_args =
        map (fun (texp, mut_flag) ->
              (type_of_type_expression true texp, mut_flag))
            args
      and purity_flag =
        for_all (fun (texp, Notmutable) -> true | (texp, Mutable) -> false)
                args
      and constr_tag =
        if is_extensible then
          ConstrExtensible {qual=compiled_module_name(); id=constr_name}
        else
          ConstrRegular(constr_idx, nbr_constrs) in
      let constr_glob =
        defined_global constr_name
          { cs_res = ty_res;
            cs_args = ty_args;
            cs_tag = constr_tag;
            cs_pure = purity_flag }
      in
        add_constr constr_glob;
        if (not purity_flag)
        or (exists (fun (ty, _) -> dangerous_vars ty != []) ty_args) then
          begin ty_desc.info.ty_dang <- true; () end;
        constr_glob :: make_constrs (succ constr_idx) rest
  in
    let constr_descs = make_constrs 0 constrs in
      pop_type_level();
      generalize_type_constr ty_res;
      do_list
        (fun cstr ->
          do_list (fun (ty,mut_flag) -> generalize_type_constr ty)
                  cstr.info.cs_args)
        constr_descs;
      Variant_type constr_descs
;;

let enter_new_record loc (ty_desc, ty_res, labels) =
  let rec make_labels i = function
    [] -> []
  | (name, typexp, mut_flag) :: rest ->
(*****
      if mem_assoc name rest then
        duplicate_constr_in_type_decl_err "Label " loc name;
******)
      let ty_arg = type_of_type_expression true typexp in
      let lbl_glob =
        defined_global name
          { lbl_res = ty_res; lbl_arg = ty_arg;
            lbl_mut = mut_flag; lbl_pos = i }
      in
        add_label lbl_glob;
        if mut_flag == Mutable or dangerous_vars ty_arg != [] then
          begin ty_desc.info.ty_dang <- true; () end;
        lbl_glob :: make_labels (succ i) rest in
  let label_descs = make_labels 0 labels in
    pop_type_level();
    generalize_type_constr ty_res;
    do_list
      (function lbl -> generalize_type_constr lbl.info.lbl_arg)
      label_descs;
    Record_type label_descs
;;
    
let enter_new_abbrev (ty_params, body) =
  let ty_body = type_of_type_expression true body in
    pop_type_level();
    generalize_type ty_body;
    map generalize_type ty_params;
    Abbrev_type (ty_params, ty_body)
;;

let enter_new_type (ty_name, params, def) =
  let ty_desc =
    { ty_stamp = new_type_stamp();
      ty_arity = list_length params;
      ty_dang  = false;
      ty_desc  = Abstract_type } in
  let ty_constr = defined_global ty_name ty_desc in
    add_type ty_constr;
    if !verbose then begin
      print_begline "Type "; print_string ty_name; print_endline " defined."
    end;
    (ty_constr, params, def)
;;

let define_new_type loc (ty_desc, params, def) =
  push_type_level();
  let ty_params =
    try
      reset_type_expression_vars params
    with failure "reset_type_expression_vars" ->
      duplicate_param_in_type_decl_err loc in
  let ty_res =
    Typ(Tconstr(ty_desc, ty_params), notgeneric) in
  let type_comp =
    match def with
      Zabstract_type ->
        pop_type_level(); Abstract_type
    | Zvariant_type constrs ->
        enter_new_variant false loc (ty_desc, ty_res, constrs)
    | Zrecord_type labels ->
        enter_new_record loc (ty_desc, ty_res, labels)
    | Zabbrev_type body ->
        enter_new_abbrev (ty_params, body)
  in
    ty_desc.info.ty_desc <- type_comp
;;

let type_typedecl loc decl =
  let new_decl =
    map enter_new_type decl
  in
    do_list (define_new_type loc) new_decl
;;

let type_excdecl loc decl =
  push_type_level();
  reset_type_expression_vars [];
  enter_new_variant true loc (constr_type_exn, type_exn, decl);
  if !verbose then begin
    do_list
      (fun (exc_name, args) ->
        print_begline "Exception "; print_string exc_name;
        print_endline " defined.")
      decl
  end
;;

let print_new_val (name, ty) =
  if !verbose then begin
    print_begline name; print_string " : ";
    print_one_type ty;
    print_endline ""
  end
;;

let fully_generalize_type loc ty =
  generalize_type ty;
  match free_type_vars ty with
     []  -> ()
  | vars -> cannot_generalize_err loc vars ty
;;

let type_valuedecl loc decl =
  let enter_val (name, typexp, prim) =
    push_type_level();
    reset_type_expression_vars [];
    let ty = type_of_type_expression false typexp in
      pop_type_level();
      fully_generalize_type loc ty;
      add_value (defined_global name { val_typ = ty; val_prim = prim });
      print_new_val (name, ty)
  in
    do_list enter_val decl
;;

let type_letdef loc rec_flag pat_expr_list =
  push_type_level();
  let (pat_ty_list, expr_ty_list) =
    list_it
     (fun (pat,expr,_) (pt, et) ->
        let alpha = new_type_var () in
          (pat,alpha,Notmutable)::pt, (expr,alpha)::et)
     pat_expr_list
     ([],[]) in
  let env =
    type_pattern_list2 pat_ty_list in
  let enter_val =
    do_list (fun (name,(ty,mut_flag)) ->
       add_value (defined_global name {val_typ=ty; val_prim=ValueNotPrim})) in
  if rec_flag then enter_val env;
  do_list (type_expr []) expr_ty_list;
  if rec_flag
  then do_list2 (fun (e, ty) (p, (Expr (d, loc)), size)
                      -> size := ty_max_size ty loc)
                expr_ty_list pat_expr_list;
  pop_type_level();
  do_list (fun (_, (ty, _)) -> fully_generalize_type loc ty) env;
  if not rec_flag then enter_val env;
  do_list (fun (name, (typ, mut_flag)) -> print_new_val (name,typ)) env;
  env
;;
  
let type_expression loc expr =
  push_type_level();
  let ty = new_type_var () in
    type_expr [] (expr,ty);
    pop_type_level();
    fully_generalize_type loc ty;
    if !verbose then begin
      print_begline "- : ";
      print_one_type ty;
      print_endline ""
    end;
    ty
;;
