/* Convert Unix-style end-of-line (LF) to MSDOS-style end-of-line (CRLF).
   With no arguments: run as a filter.
   With arguments: replace the files in-place. Wildcards accepted. */

#include <dir.h>
#include <dos.h>
#include <sys\stat.h>
#include <fcntl.h>
#include <io.h>

/* Wildcard expansion */

struct name_list {
  struct name_list * prev;
  char name[1];
};

char ** wildcard_expand(pattern, flags)
     char * pattern;
     int flags;
{
  int dir;
  int namecount;
  struct name_list * namelist, * n;
  struct ffblk ffblk;
  int done;
  int i;
  char ** res;
  
  dir = strlen(pattern);
  while (dir >= 0 && pattern[dir] != '\\' && pattern[dir] != ':')
    dir--;
  dir++;
  namecount = 0;
  namelist = 0;
  for (done = findfirst(pattern, &ffblk, flags);
       !done;
       done = findnext(&ffblk)) {
    n = (struct name_list *)
          malloc(sizeof(struct name_list) + dir + strlen(ffblk.ff_name));
    memcpy(n->name, pattern, dir);
    strcpy(n->name + dir, ffblk.ff_name);
    n->prev = namelist;
    namelist = n;
    namecount++;
  }
  res = (char **) malloc((namecount+1) * sizeof(char *));
  res[namecount] = 0;
  for(i = namecount - 1, n = namelist; i >= 0; i--, n = n->prev)
    res[i] = n->name;
  return res;
}

/* File copy */

int copy(in, out)
     int in, out;
{
  static char buffer[4096];
  int n;

  while(1) {
    n = read(in, buffer, sizeof(buffer));
    if (n <= 0) return n;
    n = write(out, buffer, n);
    if (n == -1) return n;
  }
}

/* Find and open a temporary file, in the same dir as the given argument */

char tempname[256];

int tempfile(master)
     char * master;
{
  int dir;
  struct stat stats;
  int j;
  int fd;

  dir = strlen(master);
  while (dir >= 0 && master[dir] != '\\' && master[dir] != ':')
    dir--;
  dir++;
  memcpy(tempname, master, dir);
  strcpy(tempname + dir, "tempaaa.dir");
  j = 0;
  while(1) {
    fd = open(tempname, O_CREAT | O_EXCL | O_TEXT, S_IREAD | S_IWRITE);
    if (fd != -1) return fd;
    j++;
    tempname[dir+4] = 'a' + ((j >> 8) & 15);
    tempname[dir+5] = 'a' + ((j >> 4) & 15);
    tempname[dir+6] = 'a' + ( j       & 15);
  }
}

/* The main program */

int main(argc, argv)
     int argc;
     char * argv [];
{
  int in, out;
  int retcode;
  int i;
  char ** f;

  retcode = 0;
  if (argc <= 1) {
    if (copy(0, 1) == -1)
      { perror("copying"); retcode = 1; }
  } else {
    for(i = 1; i < argc; i++) {
      for (f = wildcard_expand(argv[i], 0); *f; f++) {
        in = open(*f, O_RDONLY | O_BINARY);
        if (in == -1)
          { perror(*f); retcode = 1; continue; }
        out = tempfile(*f);
        if (copy(in, out) == -1)
          { perror("copying"); retcode = 1; continue; }
        close(in);
        close(out);
        if (unlink(*f) == -1)
          { perror("unlink"); retcode = 1; }
        if (rename(tempname, *f) == -1)
          { perror("rename"); retcode = 1; }
      }
    }
  }
  return retcode;
}

