#include <stdio.h>
#include <signal.h>
#include <setjmp.h>

#define LITTLE 0
#define BIG 1

int endianness()
{
  long n[2];
  char * p;

  n[0] = 0x41424344;
  n[1] = 0;
  p = (char *) n;
  if (strcmp(p, "ABCD") == 0) {
    return BIG;
  } else if (strcmp(p, "DCBA") == 0) {
    return LITTLE;
  } else {
    printf("Weird architecture -- neither big endian nor little endian.\n");
    exit(2);
  }
}

long foo;

void access16(p)
     short * p;
{
  foo = *p;
}

void access32(p)
     long * p;
{
  foo = *p;
}

jmp_buf failure;

void sig_handler()
{
  longjmp(failure, 1);
}

int test(fct, p)
     void (*fct)();
     char * p;
{
  int res;

  signal(SIGSEGV, sig_handler);
  signal(SIGBUS, sig_handler);
  if(setjmp(failure) == 0) {
    fct(p);
    res = 0;
  } else {
    res = 1;
  }
  signal(SIGSEGV, SIG_DFL);
  signal(SIGBUS, SIG_DFL);
  return res;
}

jmp_buf timer;

void alarm_handler()
{
  longjmp(timer, 1);
}

int speedtest(p)
     char * p;
{
  int * q;
  volatile int total;
  int i, sum;

  signal(SIGALRM, alarm_handler);
  sum = 0;
  if (setjmp(timer) == 0) {
    alarm(1);
    total = 0;
    while(1) {
      for (q = (int *) p, i = 1000; i > 0; q++, i--)
        sum += *q;
      total++;
    }
  }
  foo = sum;
  signal(SIGALRM, SIG_DFL);
  return total;
}


int alignment()
{
  long n[1001];
  int align;
  int aligned, unaligned;

  align = 1;
  if (test(access16, (char *) n + 1)) align = 2;
  if (test(access32, (char *) n + 1)) align = 2;
  if (test(access32, (char *) n + 2)) align = 4;
  if (align == 1) {
    /* Let's double check. Some OS (e.g. Ultrix on Decstations)
       trap and fix unaligned accesses in the kernel.
       These accesses are painfully slow. So we're going to check
       the speed of aligned accesses vs. the speed of unaligned accesses. */
    aligned = speedtest((char *) n);
    unaligned = speedtest((char *) n + 1);
    printf("FYI: accesses per sec.: aligned %d, unaligned %d\n",
           aligned * 1000, unaligned * 1000);
    if (aligned >= 5 * unaligned) {
      /* The performance penalty for unaligned accesses is too high.
         Don't do them. */
      align = 4;
    }
  }
  return align;
}

main()
{
  int endian = endianness();
  int align  = alignment();
  printf("Endianness/alignment options in config.h:\n");
  if (endian == BIG)
    printf("#define BIG_ENDIAN\n");
  else
    printf("#undef BIG_ENDIAN\n");
  if (align > 1)
    printf("#define ALIGNMENT\n#define INTERP_ALIGN\n");
  else
    printf("#undef ALIGNMENT\n#undef INTERP_ALIGN\n");
  return 0;
}
