(* fifo.ml *)

(* FIFO implementation of queues (interface queue.mli). *)

type 'a mulist =
    Nil
  | Cons of 'a * mutable 'a mulist
;;

type 'a t =
  { mutable entry : 'a mulist;
    mutable exit : 'a mulist }
;;

let new () = {entry = Nil; exit = Nil};;

let reset q =
  q.entry <- Nil;
  q.exit <- Nil;
  ()
;;

let put q val =
  match q.entry with
    Nil -> let elem = Cons (val, Nil)
           in q.entry <- elem;
              q.exit <- elem; ()
  | Cons (_, (Nil as tail)) -> tail <- Cons (val, Nil);
                             q.entry <- tail; ()
  | _ -> failwith "queue__put"
;;

let get q =
  match q.exit with
    Nil -> raise Empty
  | Cons (val, rest) -> q.exit <- rest;
                        (match rest with Nil -> q.entry <- Nil; () | _ -> ());
                        val
;;

let rec mulength n = function
    Nil -> n
  | Cons (_, rest) -> mulength (n+1) rest
;;

let length q = mulength 0 q.exit;;

let is_empty q = q.exit = Nil;;
