type 'a t =
   {lock : thread__mutex;
    cond : thread__condition;
    data : 'a queue__t;
    mutable closed : bool}
;;

exception Closed;;

let new () =
  {lock = thread__new_mutex ();
   cond = thread__new_condition ();
   data = queue__new ();
   closed = false}
;;

let close b =
  thread__lock b.lock;
  b.closed <- true;
  thread__unlock b.lock;
  thread__signal b.cond
;;

let put b x =
  thread__lock b.lock;
  queue__put b.data x;
  thread__unlock b.lock;
  thread__signal b.cond
;;

let rec get_loop b =
  if b.closed then raise Closed;
  try queue__get b.data
  with queue__Empty -> thread__wait b.cond b.lock;
                       get_loop b
;;

let get b =
  thread__lock b.lock;
  let x = get_loop b
  in thread__unlock b.lock;
     x
;;
