(* The Fibonacci function, parallelized. *)

let rec fib n =
  if n < 2 then 1 else fib(n-1) + fib(n-2)
;;

type 'a future = {
  mutable proc : thread__process;
  mutable v : 'a future_state
}
and 'a future_state =
   Not_yet
 | Done of 'a
 | Joined of 'a
;;

let new f arg =
  let result = {proc = thread__self (); v = Not_yet}
  in result.proc <- thread__fork (fun () -> result.v <- Done (f arg));
     result
;;

let get fut =
  match fut.v
  with Not_yet -> thread__join fut.proc;
                  begin match fut.v
                  with Done x -> fut.v <- Joined x; x
                     | _ -> failwith "future__get"
                  end
     | Done x -> thread__join fut.proc; fut.v <- Joined x; x
     | Joined x -> x
;;

let rec pfib m n =
  if m <= 0 then fib n else begin
    let fut = new (pfib (m-1)) (n-1)
    in let res2 = pfib (m-2) (n-2)
    in get fut + res2
  end
;;

if vect_length unix__command_line <> 2 then
  begin
    print_string "Usage: fib <number>\n";
    exit 1
  end
else
  begin
    let n =
      try
        int_of_string unix__command_line.(1)
      with failure "int_of_string" ->
        print_string "Bad integer constant\n";
        exit 1
    in
      print_int (pfib 6 n);
      print_newline();
      exit 0
  end
;;

