type 'a buffer = {
  data : 'a vect;
  mutable producer : int;             (* position du producteur *)
  mutable consumer : int;             (* position du consommateur *)
  mutable count : int;                (* nombre d'objets disponibles *)
  lock : thread__mutex;              (* pour l'acces exclusif *)
  nonempty : thread__condition;      (* quelque chose a ete mis *)
  nonfull : thread__condition        (* quelque chose a ete lu *)
};;

let new size init =
  { data = make_vect size init;
    producer = 0;
    consumer = 0;
    count = 0;
    lock = thread__new_mutex();
    nonempty = thread__new_condition();
    nonfull = thread__new_condition()
  }
;;

let put buffer data =
  thread__lock buffer.lock;
  while buffer.count == vect_length buffer.data do
    thread__wait buffer.nonfull buffer.lock
  done;
  buffer.data.(buffer.producer) <- data;
  let n = succ buffer.producer in
    buffer.producer <- if n < vect_length buffer.data then n else 0;
  buffer.count <- succ buffer.count;
  thread__signal buffer.nonempty;
  thread__unlock buffer.lock
;;

let get buffer =
  thread__lock buffer.lock;
  while buffer.count == 0 do
    thread__wait buffer.nonempty buffer.lock
  done;
  let res = buffer.data.(buffer.consumer) in
  let n = succ buffer.consumer in
    buffer.consumer <- if n < vect_length buffer.data then n else 0;
  buffer.count <- pred buffer.count;
  thread__signal buffer.nonfull;
  thread__unlock buffer.lock;
  res
;;

