let out c =
  output_char std_out c
;;

let rec display r = fun
   [] -> r
 | (c :: l) -> out c;
               display (c::r) l
;;

let display_rotate = fun
   [] -> failwith "display_rotate"
 | (c :: l) -> out c; (c :: display [] l)
;;

let rec display_back r = fun
   [] -> r
 | (c :: l) -> out `\008`;
               display_back (c::r) l
;;

let rec fib n = if n < 2 then 1 else fib (n-1) + fib (n-2);;

let rec show l =
  let l = display_rotate l
  in let _ = flush std_out
  in let l = display_back [] l
  in let _ = fib 16
  in show l
;;

let chars s =
  let result = ref []
  in for i = string_length s - 1 down to 0 do
     result := nth_char s i :: !result
     done;
     !result
;;

let proc_show () =
  show (chars "Caml Light, le langage qu'il vous fo !   ******   ")
;;

type r = {t1 : thing; t2 : thing; mutable t3 : thing; t4 : thing}
and thing =
   Nothing
 | Int of int
 | Float of float
 | Char of char
 | List of thing list
 | Triple of thing * thing * thing
 | Record of r
 | Vect of thing vect
;;

let random_new (seed : int) = ref seed;;
let random g max =
  g := 31415821 * !g + 1234567;
  let result = !g mod max
  in if result >= 0 then result else -result
;;

let random2 g max =
  g := 27182821 * !g + 56789;
  let result = !g mod max
  in if result >= 0 then result else -result
;;

let make_simple_thing g =
  match random g 4
  with 0 -> Nothing
     | 1 -> Int (random g 1000000)
     | 2 -> Float (float_of_int (random g 1000000))
     | 3 -> Char (char_of_int (random g 256))
     | _ -> failwith "make_simple_thing"
;;

let rec make_thing_list g n =
  if n <= 0 then []
  else make_simple_thing g :: make_thing_list g (n-1)
;;

let rec make_thing g m =
  if m <= 4 then make_simple_thing g
  else let n = random g 100000
       in if n < 20000
          then List (make_thing_list g m)
          else if n < 40000
          then let m1 = random g (m - 2)
               in let x1 = make_thing g m1
               in let m2 = random g (m - m1 - 2)
               in let x2 = make_thing g m2
               in let m3 = m - m1 - m2 - 2
               in let x3 = make_thing g m3
               in Triple (x1, x2, x3)
          else if n < 60000
          then let m1 = random g (m - 3)
               in let x1 = make_thing g m1
               in let m2 = random g (m - m1 - 3)
               in let x2 = make_thing g m2
               in let m3 = random g (m - m1 - m2 - 3)
               in let x3 = make_thing g m3
               in let m4 = m - m1 - m2 - m3 - 3
               in let x4 = make_thing g m4
               in Record {t1 = x1; t2 = x2; t3 = x3; t4 = x4}
     else let l = 35000 / (100001 - n)
          in let l = if l >= m then m - 1 else l
          in let v = make_vect l Nothing
          in let mrest = ref m
          in for i = 0 to (l - 1)
             do let mi = random g (!mrest - l)
                in (v.(i) <- make_thing g mi;
                    mrest := !mrest - mi)
             done;
             Vect v
;;

let make_one_thing seed =
  let g = random_new seed
  in match random g 20
  with 0 -> make_thing g 5
     | 1 -> make_thing g 10
     | 2 -> make_thing g 100
     | 3 -> make_thing g 1000
     | 4 -> make_thing g 10000
     | _ -> make_thing g 20
;;

type 'a queue = {
  lock : thread__mutex;
  cond : thread__condition;
  mutable contents : 'a list
};;

let new () = {
  lock = thread__new_mutex ();
  cond = thread__new_condition ();
  contents = []
};;

let put q x =
  thread__lock q.lock;
  begin try
    while list_length q.contents >= 10
    do thread__wait q.cond q.lock
    done;
    q.contents <- x :: q.contents
  with x -> thread__unlock q.lock; raise x
  end;
  thread__unlock q.lock;
  thread__signal q.cond
;;

let hd (x::t) = x
and tl (x::t) = t
;;

let get q =
  thread__lock q.lock;
  let result = begin try
                 while q.contents = []
                 do thread__wait q.cond q.lock
                 done;
                 let x = hd q.contents
                 in q.contents <- tl q.contents;
                    x
               with e -> thread__unlock q.lock; raise e
               end
  in thread__unlock q.lock;
     thread__signal q.cond;
     result
;;

type packet = {
  seed : int;
  th : thing ref;
  proc : thread__process
};;

let master_rng = random_new (int_of_string unix__command_line.(1));;

let rec prod q n =
  if n <= 0 then ()
  else let r = ref Nothing
       in let s = random2 master_rng 1000000
       in let p = { seed = s;
                    th = r;
                    proc = thread__fork (fun () -> r := make_one_thing s)
                  }
       in put q p;
          prod q (n - 1)
;;

let bug s =
  prerr_endline s;
  unix__exit 101
;;

let rec cons q n =
  if n <= 0 then ()
  else let p = get q
       in let _ = thread__join p.proc
       in let th2 = make_one_thing p.seed
       in if !(p.th) = th2
          then cons q (n - 1)
          else bug "not equal"
;;

let rec transmit q1 q2 n =
  if n <= 0 then ()
  else let p = get q1
       in put q2 p;
          transmit q1 q2 (n - 1)
;;

let n = int_of_string unix__command_line.(2);;

let qprod = (new () : packet queue);;
let qcons = (new () : packet queue);;

let pr = thread__fork (fun () -> prod qprod n);;
let tr = thread__fork (fun () -> transmit qprod qcons n);;
let co = thread__fork (fun () -> cons qcons n);;

let _ = thread__fork proc_show;;

thread__join pr;;
thread__join tr;;
thread__join co;;

print_endline "";;
flush std_out;;
unix__exit 0;;
